/* $Revision: 1.17 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGrid.C
 *
 * DESCRIPTION: Data structures to help with DEM creation
 *
 * EXPORTS:  
 *
 * NOTES:        
 *    Need to make this a template so float, double,
 *	or could have pointers for each type
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

// FIX: what to do if out of range for offsets?

/***************************************************************************
 * INCLUDES
 ***************************************************************************/
#include <cstdio>
#include <cmath>
#include <cfloat> // DBL_MAX

#include <iostream>
#include <string>

#include "XcGrid.H"

using namespace std;

/***************************************************************************
 * FUNCTION: Vector support code
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 *     PovRay uses a left handed coordinate frame.  Hence the -depth flip.
 ***************************************************************************/
// a[0] : x
// a[1] : b
// a[2] : z
// FIX:  may want to go back to floats
bool CrossProduct (const double a[3], const double b[3], double result[3]) {
  result[0] = a[1]*b[2]-a[2]*b[1];
  result[1] = a[2]*b[0]-a[0]*b[2];
  result[2] = a[0]*b[1]-a[1]*b[0];
  return (true);
}
double Distance (const double v[3]) {
  return (sqrt (v[0]*v[0]+v[1]*v[1]+v[2]*v[2]));
}
bool NormalizeVec (double v[3]) {
  double d = Distance (v);
  v[0] = v[0]/d;
  v[1] = v[1]/d;
  v[2] = v[2]/d;
  return (true);
}
bool GetNormal (const double a[3], const double b[3], double result[3]) {
  CrossProduct(a,b,result);
  NormalizeVec(result);
  return(true);
}
void PrintVector(float v[3]) {std::cout << v[0] << " " << v[1] << " " << v[2] << "\n";}


//////////////////////////////////////////////////////////////////////
// XcGrid -- X-Core Grid representation
//////////////////////////////////////////////////////////////////////


// Let's get this party started.
XcGrid::XcGrid () {
  regularGrid=true;
  w=h=0;
  dx=dy=-1.0;
  baseX=baseY=0.0;
  haveColorPalette=false;
}

// Let's get this party started.
XcGrid::XcGrid (const size_t width, const size_t height, bool reg) {
  regularGrid=reg;
  setSize(width,height);
  dx=dy=1.0;
  baseX=baseY=0.0;
  haveColorPalette=false;
}

// Destructor
//XcGrid::~XcGrid () {
//  if (cp) delete cp;
//}

// Not sure what will be needed for a resize.
void XcGrid::setSize(const size_t width, const size_t height) {
  assert (width < 100000); // We are going to get in trouble if this is huge
  assert (height < 100000);
  w = width; h = height;
  const size_t nelem=w*h;
  nan.assign(nelem,false); // default to no NaN values
  compressedIndex.resize(nelem);
  if (regularGrid) { x.resize(0);y.resize(0);
  } else {           x.resize(nelem);y.resize(nelem);}
  z.resize(nelem);
  u.resize(nelem);
  v.resize(nelem);
  nx.resize(nelem);
  ny.resize(nelem);
  nz.resize(nelem);
  triPerVert.resize(nelem);  // FIX: what to do if a 2nd resize happends?
  triangles.resize(0); triangles.reserve(2*nelem);
  return;
}

// FIX: these should probably be in the class def
void XcGrid::setDelXDelY(const float _dx, const float _dy) {dx=_dx;dy=_dy;}
void XcGrid::setOrigin(const float _x, const float _y) {baseX=_x; baseY=_y;}
void XcGrid::setRegular(bool reg) {regularGrid=reg;}

void XcGrid::setZ(const size_t x_offset, const size_t y_offset, float _z){
  assert (offsetsValid(x_offset,y_offset));
  size_t i = getIndex(x_offset,y_offset);
  z[i]=_z;
}

// val true means the point has no data (hole in the data)
// val false, the point has valid data
void XcGrid::setZToNaN(const size_t x_offset, const size_t y_offset, bool val){
  assert (offsetsValid(x_offset,y_offset));
  size_t i = getIndex(x_offset,y_offset);
  nan[i]=val;
}

void XcGrid::setUV(const size_t x_offset, const size_t y_offset, const float _u,const float _v) {
  assert (offsetsValid(x_offset,y_offset));
  size_t i = getIndex(x_offset,y_offset);
  u[i] = _u;
  v[i] = _v;
}

// FIX: allow aribrary functions.
// Maybe use something like python to do the text to math
void XcGrid::scaleZ(const float mult, const float offset) {
  {
    char buf[128];
    snprintf(buf, sizeof(buf),"zScale mult=%f offset=%f",mult,offset);
    addLogEntry (string(buf));
  }
  size_t i;
  for (i=0;i<w*h;i++) {
    if (isNaN(i)) continue;
    z[i] = z[i] * mult + offset;
  }
}

// Debugging dump of internal state.
void XcGrid::print() {
  std::cout << "  ** XcGrid::Print **\n"
	    << " w: " << w << "\n"
	    << " h: " << h << "\n"
	    << " r: " << regularGrid<< "\n"
	    << " bx: " << baseX << "\n"
	    << " by: " << baseY << "\n"
	    << " dx: " << dx << "\n"
	    << " dy: " << dy << "\n"
	    << " ns: " << nan.size() << "\n"
	    << " cs: " << compressedIndex.size() << "\n"
	    << " cl: " << compressedIndex.back() << "\n"
	    << " xs: " << x.size() << "\n"
	    << " ys: " << y.size() << "\n"
	    << " zs: " << z.size() << "\n"
	    << " ts: " << triangles.size() << "\n";
}


size_t XcGrid::getIndex(const size_t x_offset, const size_t y_offset) {
  assert (offsetsValid(x_offset,y_offset));
  return (x_offset + y_offset*w);
}

void XcGrid::getXYfromIndex(const size_t index, size_t &x_offset, size_t &y_offset) {
  y_offset = int(floor((index/(float(w)))));
  x_offset = index%w;
}

// How do we know that the index has been generated?
size_t XcGrid::getCompressedIndex(const size_t x_offset, const size_t y_offset) {
  assert (offsetsValid(x_offset,y_offset));
  return (compressedIndex[getIndex(x_offset,y_offset)]);
}
// FIX: This is slow.  Is there a better way?
// Need to do a divide an conquer
void XcGrid::getXYfromCompressedIndex(const size_t cindex, size_t &x_offset, size_t &y_offset) {
  assert (cindex<compressedIndex.size());
  size_t i;
  // FIX: use a STL find command!
  for (i=i;i<compressedIndex.size() && cindex != compressedIndex[i];i++);
  assert (i<compressedIndex.size());  // FIX: make a graceful failure
  assert (true!=nan[i]);
  getXYfromIndex(i,x_offset,y_offset);
}


// functions for regular grids
float XcGrid::getX(const size_t x_offset, const size_t y_offset) {
  if (regularGrid) {
    assert (0 < dx);
    return (x_offset * dx);
  }
  size_t i = getIndex(x_offset,y_offset);
  return (x[i]);
}

// functions for regular grids
float XcGrid::getY(const size_t x_offset, const size_t y_offset) {
  if (regularGrid) {
    //assert (0 < dy);  // FIX:  should actually be less than 0, right?
    return (y_offset * dy);
  }
  size_t i = getIndex(x_offset,y_offset);
  return (x[i]);
}

// functions for regular grids
float XcGrid::getZ(const size_t x_offset, const size_t y_offset) {
  size_t i = getIndex(x_offset,y_offset);
  return (z[i]);
}

void XcGrid::getXYZ(const size_t x_offset, const size_t y_offset, float &_x, float &_y, float &_z){
  assert (offsetsValid(x_offset,y_offset));
  size_t i = getIndex(x_offset,y_offset);
  _x=getX(x_offset,y_offset);
  _y=getY(x_offset,y_offset);
  _z = z[i];
}

void XcGrid::getXYZ(const size_t index, float &_x, float &_y, float &_z) {
  assert (index<z.size());
  size_t xoff,yoff;
  getXYfromIndex(index,xoff,yoff);
  _x=getX(xoff,yoff);
  _y=getY(xoff,yoff);
  _z = z[index];
}


void XcGrid::getNormal(const size_t x_offset, const size_t y_offset,
		       float &_nx, float &_ny, float &_nz) {
  assert (offsetsValid(x_offset,y_offset));
  size_t i = getIndex(x_offset,y_offset);
  _nx=nx[i];_ny=ny[i];_nz=nz[i];
}

void XcGrid::getUV(const size_t x_offset, const size_t y_offset, float &_u, float &_v){
  assert (offsetsValid(x_offset,y_offset));
  size_t i = getIndex(x_offset,y_offset);
  _u = u[i];
  _v = v[i];
}

// return true if the offsets are actually in the image
bool XcGrid::offsetsValid (const size_t x_offset, const size_t y_offset){
  if (x_offset >= w) return false;
  if (y_offset >= h) return false;
  return true;
}

void XcGrid::genCompressedIndex(){
  size_t i;
  int count=-1;
  for (i=0;i<nan.size();i++) {
    if (!nan[i]) count++;
    //std::cout << i << ": " << count << " " << nan[i] << "\n";
    compressedIndex[i]=count;
  }
}

// Calc the surface normal for a triangle
void XcGrid::triangleNormal(const size_t indexA, const size_t  indexB, const size_t indexC,
			    float &nx,float &ny,float &nz)
{
  float a[3],b[3],c[3];
  size_t ax,ay,bx,by,cx,cy;  // x and y offsets for each point
  getXYfromIndex(indexA,ax,ay);
  getXYfromIndex(indexB,bx,by);
  getXYfromIndex(indexC,cx,cy);
  getXYZ(ax,ay,a[0],a[1],a[2]);
  getXYZ(bx,by,b[0],b[1],b[2]);
  getXYZ(cx,cy,c[0],c[1],c[2]);

  double n[3];
  double v1[3],v2[3];
  v1[0]=a[0] - b[0]; v1[1]=a[1] - b[1]; v1[2] = a[2] - b[2];
  v2[0]=a[0] - c[0]; v2[1]=a[1] - c[1]; v2[2] = a[2] - c[2];
  GetNormal(v1,v2,n);
  nx=n[0];
  ny=n[1];
  nz=n[2];
}

float XcGrid::getXMin(bool checkNaNs) {
  assert (!checkNaNs);  // dealling with NaNs not yet implimented
  if (isRegular()) {
    float x,y,z;
    getXYZ(0,x,y,z);
    return(x);
  }
  assert(false); // Irregulars not implemented.
  return(-1.);
}
float XcGrid::getXMax(bool checkNaNs) {
  assert (!checkNaNs);  // dealling with NaNs not yet implimented
  if (isRegular()) {
    float x,y,z;
    getXYZ(w*h-1,x,y,z);
    return(x);
  }
  assert(false); // Irregulars not implemented.
  return(-1.);
}
float XcGrid::getYMin(bool checkNaNs) {
  assert (!checkNaNs);  // dealling with NaNs not yet implimented
  if (isRegular()) {
    float x,y,z;
    getXYZ(w*h-1,x,y,z);
    return(y);
  }
  assert(false); // Irregulars not implemented.
  return(-1.);
}
float XcGrid::getYMax(bool checkNaNs) {
  assert (!checkNaNs);  // dealling with NaNs not yet implimented
  if (isRegular()) {
    float x,y,z;
    getXYZ(0,x,y,z);
    return(y);
  }
  assert(false); // Irregulars not implemented.
  return(-1.);
}
float XcGrid::getZMin(bool checkNaNs) {
  assert (!checkNaNs);  // dealling with NaNs not yet implimented
  size_t i;
  double min = DBL_MAX;
  for (i=0;i<w*h;i++) {
    if (isNaN(i)) continue;
    if (z[i]<min) min = z[i];
  }
  // FIX: added error check for not found
  return (min);
}
float XcGrid::getZMax(bool checkNaNs) {
  assert (!checkNaNs);  // dealling with NaNs not yet implimented
  size_t i;
  double max = -DBL_MAX;
  for (i=0;i<w*h;i++) {
    if (isNaN(i)) continue;
    if (z[i]>max) max = z[i];
  }
  // FIX: added error check for not found
  return (max);
  
}

// Just return the center of the bounding box.
void XcGrid::getCenter (float &x, float &y, float &z) {
  x = (getXMin() + getXMax())/2;
  y = (getYMin() + getYMax())/2;
  z = (getZMin() + getZMax())/2;
}

// Averages all x,y,z for those that do not have NaN values.
// Might be go for centering on funny shaped terrain patches.
void XcGrid::getWeightedCenter (float &x, float &y, float &z) {
  assert(false); // NOT implemented
}



// fix up the data structures for triPerVert and triangles
// Calculate normal for triangle.
// Specify vertices in counter clockwise order when viewed from top
void XcGrid::addTriangle(const size_t indexA, const size_t indexB, const size_t indexC) {
  assert (!isNaN(indexA));
  if (isNaN(indexB)) {
    std::cerr << "b was NaN: " 
	      << indexA << " "
	      << indexB << " "
	      << indexC << "\n";
    assert(false);
  }
  assert (!isNaN(indexC));
  XcTriangle t;
  t.vertices[0] = indexA;
  t.vertices[1] = indexB;
  t.vertices[2] = indexC;

  triangleNormal(indexA, indexB, indexC, t.nx, t.ny, t.nz);

  //
  // Fix up the data structures so it is easy to go both ways
  // 
  triangles.push_back(t);  // Push the sucker on the stack of triangles
  // Don't anyone dare remove a triangle!
  size_t tri_index=triangles.size()-1; // Here is where it is.

  (triPerVert[indexA]).push_back(tri_index);
  (triPerVert[indexB]).push_back(tri_index);
  (triPerVert[indexC]).push_back(tri_index);
}

// Always use the uncompressed index!
void XcGrid::genTriangles() {
  size_t _x,_y;
  assert (w>0);
  assert (h>0);
  
  for (_y=0;_y<h-1;_y++) {
    for(_x=0;_x<w-1;_x++) {
      const int a = getIndex(_x,  _y);
      const int b = getIndex(_x+1,_y);
      const int c = getIndex(_x,  _y+1);
      const int d = getIndex(_x+1,_y+1);

      //std::cout << a << " " << b << " " << c << " " << d <<"\n";

      // Order matters.  Try to always have normal point up
      // right handed, counter clockwise is up
      // Vertices looking down on the DEM:
      //   a  b
      //   c  d
      if (!nan[b] && !nan[c]) {
	if (!nan[a]) {
	  addTriangle(a,c,b);	  // triangle a,c,b
	}
	if (!nan[d]) {
	  addTriangle(c,d,b);	  // triangle c,d,b
	}
      } else  if (!nan[a] && !nan[d]) {
	// Try making the triangle the other way to miss NaN point
	if (!nan[b]) {
	  addTriangle(a,d,b);	  // triangle a,d,b
	}
	if (!nan[c]) {
	  addTriangle(a,c,d);	  // triangle a,c,d
	}
      } else {
	//std::cout << "dropped square:" << a << " " << b << " " 
	//	  << c << " " << d <<"\n";
      }
    } // for x
  } // for y

}

// FIX:  This algorithm may be a bit boneheaded
// Just take the vector average.  Sum by component then div by # of triangles
void XcGrid::genNormals(){

  // Rebuild normal for each triangle first.  May have been through a scaling
  // FIX: Need to a way to track when this needs to be done.
  // e.g. Flag when  the normal data is stale.
  { 
    size_t tri;
    for (tri=0;tri<triangles.size();tri++) {
      triangleNormal(triangles[tri].vertices[0],
		     triangles[tri].vertices[1],
		     triangles[tri].vertices[2],
		     triangles[tri].nx,
		     triangles[tri].ny,
		     triangles[tri].nz
		     );
    }
  }

  size_t vert;
  for (vert=0;vert<nx.size();vert++) {
    if (nan[vert]) continue;	// Don't compute for missing points
    if (0==triPerVert[vert].size()) { // Isolated point.  Give a sane default
      nx[vert] = ny[vert] = 0.;
      nz[vert] = 1.0;
      continue;  
    }
    double nxsum=0,nysum=0,nzsum=0;
    size_t i;
    for (i=0;i<triPerVert[vert].size();i++) {
      const size_t t = (triPerVert[vert])[i];
      nxsum+=triangles[t].nx;
      nysum+=triangles[t].ny;
      nzsum+=triangles[t].nz;
    }
    nx[vert]=nxsum / triPerVert[vert].size();
    ny[vert]=nysum / triPerVert[vert].size();
    nz[vert]=nzsum / triPerVert[vert].size();
  }
}

// Simple 1:1 texture map  0..1 and 0..1
void XcGrid::genUV() {
  size_t xoff,yoff;
  float u,v;
  for (yoff=0;yoff<h;yoff++) {
    v = 1 - (yoff/float(h));
    for (xoff=0;xoff<w;xoff++) {
      u = xoff/float(w);
      setUV(xoff,yoff,u,v);
    }
  }
}


void XcGrid::addLogEntry(const std::string &l) {
  log.push_back(l);
}

// read one off of disk
bool XcGrid::setColorPalette(std::string s) {
  bool ok=cp.setColorsFromFile(s.c_str());
  if (ok) haveColorPalette=true;
  return (ok);
  
}
// Copy an existing color palette
void XcGrid::setColorPalette(const colorPalette &_cp) {
  cp=_cp;
  haveColorPalette=true;
}




//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////


#ifdef REGRESSION_TEST
int main (int argc, char *argv[]) {
  bool ok=true;
  {
    XcGrid g(2,2);
    g.setDelXDelY(1.0,1.0);
    g.setZ(0,0,.1);
    g.setZ(1,0,.11);
    g.setZ(0,1,.09);
    g.setZ(1,1,.095);
    g.setZToNaN(0,0);
    g.genCompressedIndex();
    g.genTriangles();
    g.genNormals();

    float nx,ny,nz;
    g.getNormal(1,0,nx,ny,nz);//  std::cout << "n(1,1):" << nx << " " << ny << " " << nz << "\n";
    g.getNormal(0,1,nx,ny,nz);//  std::cout << "n(1,1):" << nx << " " << ny << " " << nz << "\n";
    g.getNormal(1,1,nx,ny,nz);//  std::cout << "n(1,1):" << nx << " " << ny << " " << nz << "\n";
    // FIX: add tests to make sure it all worked!

    if (!g.setColorPalette("TestData/sample.cpt")) {
      std::cout << "   setColorPalette: FAILED!" << std::endl;
      ok=false;
    }
    if (!g.hasColorPalette()) {
      std::cout << "   hasColorPalette: FAILED!" << std::endl;
      ok=false;
    } else {
      float depth=200;
      float _r,_g,_b;
      colorPalette cp = g.getColorPalette();
      if (!cp.getColor(depth,_r,_g,_b)) {
	std::cout << "   getColor: FAILED!" << std::endl;
	ok=false;
      }
      if (255!=_r || 128 != _g || 0 != _b) {
	std::cout << "   getColor expected value: FAILED!" << std::endl;
	ok=false;
      }
    }
  }

  {
    std::cerr << "FIX: write test for compressed indices" << std::endl;
  }

  // FIX: add something that can actually be tested... e.g. check the normals

  std::cout << argv[0] << (ok?": ok.":": FAILED!") << std::endl;

  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}


#endif // REGRESSION_TEST
