/* $Revision: 1.17 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGrid.H
 *
 * DESCRIPTION: Data structures to help with DEM creation
 *
 * EXPORTS:  
 *
 * NOTES:        
 *
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

// Include Guard
#ifndef __XCGRID_H_
#define __XCGRID_H_

/***************************************************************************
 * INCLUDES
 ***************************************************************************/

#include <cassert>
#include <string>
#include <vector>

#include "colorPalette.H"

//////////////////////////////////////////////////////////////////////
// Kinda ugly cross dependency with XcGrid index into x,y,z,nan
class XcTriangle {
public:
  float nx,ny,nz;	// Normal
  size_t vertices[3];   // Indexes for each of the 3 vertices that make up the
			// triangle.  Uncompressed index
};

//////////////////////////////////////////////////////////////////////
class XcGrid {

public:
  XcGrid(const size_t width, const size_t height, bool reg=true);
  XcGrid();
  //~XcGrid();

  void setDelXDelY(const float _dx, const float _dy);
  void setSize(const size_t width, const size_t height); // FIX: Not resizeable 
  void getOrigin(float &_x,  float &_y) {_x=baseX; _y=baseY;} // Left,Top corner
  void setOrigin(const float _x, const float _y); // Left,Top corner
  void setRegular(bool reg);  // True for regular grids. False if x, y not perfect grid

  void setZ(const size_t x_offset, const size_t y_offset, const float _z);
  void setXY(const size_t x_offset, const size_t y_offset, const float _z);
  void setZToNaN(const size_t x_offset, const size_t y_offset, bool val=true);
  void setUV(const size_t x_offset, const size_t y_offset, const float _u,const float _v);

  // Z' = Z*mult + offset
  void scaleZ(const float mult, const float offset);

  void print(); // Mostly for debugging. // FIX: rename to print

  size_t getWidth() {return(w);}
  size_t getHeight() {return(h);}
  float  getDX() {return(dx);}
  float  getDY() {return(dx);}
  size_t getIndex(const size_t x_offset, const size_t y_offset);
  void   getXYfromIndex(const size_t index, size_t &x_offset, size_t &y_offset);
  // Compressed indices should be a sub class.
  size_t getCompressedIndex(const size_t x_offset, const size_t y_offset);
  size_t getCompressedIndex(const size_t i){/*FIX:check i*/return(compressedIndex[i]);}
  size_t getNumCompressedIndices() {return(1+compressedIndex.back());} // FIX: possible off by one error
  void   getXYfromCompressedIndex(const size_t cindex, size_t &x_offset, size_t &y_offset);
  float  getX(const size_t x_offset, const size_t y_offset);  // Y is unused for regular grid
  float  getY(const size_t x_offset, const size_t y_offset);  // X is "
  float  getZ(const size_t x_offset, const size_t y_offset);  // X is "
  void   getXYZ(const size_t x_offset, const size_t y_offset, float &_x, float &_y, float &_z);
  void   getXYZ(const size_t index, float &_x, float &_y, float &_z);  // regular index
  void   getUV(const size_t x_offset, const size_t y_offset, float &_u, float &_v);
  void   getNormal(const size_t x_offset,const size_t y_offset,float&_nx,float&_ny,float&_nz);
  bool   isNaN(const size_t index) {return nan[index];}
  bool   isNaN(const size_t x_offset, const size_t y_offset) {
    return nan[getIndex(x_offset,y_offset)];
  }
  void	 getRGB(const size_t x_offset, const size_t y_offset,
		float &r, float &g, float &b) {r=1.0;g=1.0;b=1.0;} // FIX:add real code

  void   triangleNormal(const size_t indexA, const size_t indexB, const size_t indexC,
			float &nx,float &ny,float &nz);

  size_t getNumTriangles () {return triangles.size();}
  XcTriangle getTriangle (const size_t index) {
    assert(index<triangles.size());
    return triangles[index];
  }

  // Bounding box queries...
  float getXMin(bool checkNaNs=false); // FIX: what does checkNaNs mean?
  float getXMax(bool checkNaNs=false);
  float getYMin(bool checkNaNs=false);
  float getYMax(bool checkNaNs=false);
  float getZMin(bool checkNaNs=false);
  float getZMax(bool checkNaNs=false);

  void getCenter (float &x, float &y, float &z);
  // Averages all x,y,z for those that do not have NaN values.
  void getWeightedCenter (float &x, float &y, float &z);


  void        setTextureFile(std::string filename) {textureFileName=filename;};
  std::string getTextureFile() {return(textureFileName);};
  std::string getTextureFileType() {return("FIX:I don't know");};

  bool setColorPalette(std::string s); // read one off of disk
  void setColorPalette(const colorPalette &cp); // Copy existing color palette
  colorPalette getColorPalette() { return cp; };
  bool hasColorPalette() {return haveColorPalette;}

  // return true if the offsets are actually in the image
  bool offsetsValid (const size_t x_offset, const size_t y_offset);
  // False means x and y are not computed but stored
  bool isRegular() {return(regularGrid);}


  // WARNING: once you do these, don't do any more changes to the DEM
  void genCompressedIndex();
  void genTriangles();  // Find all the triangles and their normals.  Fill in
			// datastructures
  void genNormals();	// Calculate all the vertex normals once the triangles
			// have been found.
  void genUV();		// make a default UV texture coord setup... regular 1:1 map

  void addLogEntry(const std::string &l);
  size_t getNumLogEntries() {return log.size();};
  std::string getLogEntry(size_t i) {assert(i<log.size());return(log[i]);};
private:
  void addTriangle(const size_t indexA, const size_t indexB, const size_t indexC);

  // FIX: add validation flags that the setSize, and gen functions have been run.

  size_t w,h;  		// Width and height of the DEM
  bool regularGrid;	// Set to false for irregular grid points
  float baseX,baseY;	// The location of the origin
  float dx,dy;		// Cell width
  std::vector<bool> nan;  	// True if the point at this index is a nan
  std::vector<size_t> compressedIndex;  // Index that leaves out NaN points
  std::vector<float> x,y;  	// The location of each point in the dem.
  std::vector<float> z;  	// The location of each point in the dem.
  std::vector<float> u,v;	// Texture map u,v cordinates
  std::vector<float> nx,ny,nz; // Normals for each vertex
  std::vector<std::vector<size_t> > triPerVert;  // per index list of triangles that
					     // use that vertex
  std::vector<XcTriangle> triangles;	// Array of all the triangles

  std::string textureFileName;	// Filename for the texture map

  std::vector<std::string> log; // Keep track of everything that we do to the grid here

  colorPalette cp; // delete in destructor if set.
  bool haveColorPalette; // true if the color palette has been set.
};




//////////////////////////////////////////////////////////////////////
// FIX:  These should probably be wrapped inside a XcGrid name space.
bool CrossProduct (const double a[3], const double b[3], double result[3]);
double Distance (const double v[3]);
bool NormalizeVec (double v[3]);
bool GetNormal (const double a[3], const double b[3], double result[3]);
void PrintVector(double v[3]);


#endif // ifdef __XCGRID_H_
