// Actual implementation of LwFile, Chunk, SubChunk

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 	// open
#include <sys/mman.h>	// mmap

#include <cstdio> 	// DebugPrintf

#include <cmath> // Why does it undefine isnan?
extern "C" int isnan (double);  // FIX: what is the right way to do this???

#include <iostream>
#include <fstream>
#include <vector>
#include <string>


#include "lw-tags.H"
#include "lw-io.H"

using namespace std;

/***************************************************************************
 * LOCAL MACROS and DEFINES
 ***************************************************************************/


/*** Debugging levels... the higher the # the more output you get. ***/
#define ALWAYS 0	/* Hint: if you use always, it won't work without -g */
#define TERSE 1
#define TRACE 4
#define VERBOSE 8
#define BOMBASTIC 20	/* Major spew... when life sucks. */

#ifdef NDEBUG
#  define DebugPrintf(n,s)		/* expand to nothing */
#else
   int debug_level = TERSE;//BOMBASTIC;		/* default debug level */
#  define DebugPrintf(n,s)  (((n) <= debug_level) ? printf s :0)
#endif

//////////////////////////////////////////////////////////////////////
// EVIL GLOBAL
//////////////////////////////////////////////////////////////////////

static LwTag lwTag; // FIX: only want one copy of this thing floating around!

//////////////////////////////////////////////////////////////////////
// static local Helper functions
//////////////////////////////////////////////////////////////////////

static void printBuf (const char *buf, const size_t len) {
  size_t i;
  for (i=0;i<len;i++)
    cout << i << ": " << (isprint(buf[i])?buf[i]:'*')
         << " Int: " << int((unsigned char)(buf[i])) << endl;
}

//////////////////////////////////////////////////////////////////////
// LwFile
//////////////////////////////////////////////////////////////////////

// Constructor
// INPUTS: filename - file to mmap off of the disk
//	   ok - true if loaded ok, false if filed to load
LwFile::LwFile(const string _filename, bool &ok) {
  ok = true;  // Assume it works till we hit an error
  filename = _filename;
  isMMapped=false; // Set to false until we have a confirmed good mmap
  start=data=0;
  dataSize=0;
  length=0;
  //
  // mmap data
  //
  {
    {
      struct stat sb;
      int r = stat (filename.c_str(), &sb);
      if (0 != r) { ok=false; perror("ERROR: stat failed"); return; }
      dataSize = sb.st_size;
    }

    int fd = open (filename.c_str(), O_RDONLY, 0);
    if (-1==fd) { ok=false; perror ("ERROR: open failed"); return; }

    data = (char *)mmap (0, dataSize, PROT_READ,  MAP_FILE, fd, 0);
    if ((char *)(-1)==data) { ok=false; perror ("ERROR: mmap failed"); return; }
    close(fd); // Close does not munmap, so close as soon as possible
  
  } // mmap section
  isMMapped=true;

  //
  // Verify the file just a little bit
  //

  // 4 byte FORM, 4 byte length, 4 byte LWO2
  if (dataSize<4+4+4) {ok=false;cerr<<"ERROR: file truncated "<<endl; return;}
  form = lwTag.getEnum(data,dataSize);
  if (FORM!=form) {ok=false; cerr <<"ERROR: no beginning FORM tag"<<endl;return;}
  length = getU4(data+4,dataSize-4);
  if ((dataSize-8) != length) {ok=false;cerr<<"ERROR: File truncated?"<<endl;return;}
  if (LWO2!=lwTag.getEnum(data+8,dataSize-8)) {ok=false;cerr <<"ERROR: not a LWO2 file"<<endl;return;}
  start = data+8;

  //
  // Parse all the Chunks
  //
  {
    char *cur=start+4; // Skip FORM, length, LWO2
    size_t curLen = length - 4; // ditto
    int count=0;
    while (curLen > 0) {
      //cout << "count=" << count << " cur=" << int(cur) << " curLen=" << curLen << endl;
      count ++;
      assert(cur);
      assert(cur<=data+dataSize);
      assert(curLen<999999999); // Watch for wrap around

      LwChunk *c = getChunk(cur,curLen);
      if (!c) {ok=false;cerr<<"ERROR: failed to getChunk"<<endl;return;}
      chunks.push_back(c);

      size_t cSize = c->getDataSize();
      assert (cSize>=4);

      cur += cSize; curLen -= cSize;
    }
  }

  // Last integrity check
  assert(isMMapped);
  assert(start);
  assert(dataSize>=8);
  assert((dataSize-8)==length);
  assert(data);
  assert(FORM==form);
} // LwFile::LwFile(const string _filename, bool &ok) - CONSTRUCTOR

//
// Destructor
//
LwFile::~LwFile() {
  size_t i;
  for(i=0;i<chunks.size();i++) delete chunks[i];
  chunks.clear();
  // Hang on to this mmap until everything else is cleaned up.
  if (isMMapped) {
    assert(data);
    int r = munmap(data,dataSize);
    if (-1==r) perror ("ERROR: munmap failed in destructor");
    isMMapped=false;
  }
} // LwFile DESTRUCTOR

// FIX: This is probably a disgusting memory hog
string
LwFile::setStr(bool useData) {
  string s;
  s = setStrID4(FORM); // or s = "FORM"
  s += setStrU4(0); // Set size to initialy 0
  s += setStrID4(LWO2); // or s = "FORM"

  size_t i;
  for (i=0;i<chunks.size();i++) {
    s += chunks[i]->setStr(useData);
  }
  string size=setStrU4(s.length()-8);
  assert(size.length()==4);
  for (i=0;i<4;i++) s[i+4] = size[i];
  return(s);
}

#ifdef REGRESSION_TEST
// Just for regression test
bool LwFile::checkSetStr() {
  if (!data) { cerr<<"LwFile checkSetStr ERR: data was not set!"<<endl;}
  bool ok=true;
  size_t i;
  for (i=0;i<chunks.size();i++) {
    if (!chunks[i]->checkSetStr()) {
      cout << "  checkSetStr failed on chunk " << i 
	   << " - " << lwTag.getString(chunks[i]->tag) <<endl;
      ok = false;
    }
  }
  cout << "  LwFile::checkSetStr: " << (ok?"ok":"FAILED") << endl;
  return(ok);
}
#endif // REGRESSION_TEST

void
LwFile::print() {
  cout << "filename:  " << filename << endl
       << "isMMapped: " << (isMMapped?"true":"false") << endl
       << "length:    " << length << endl
       << "dataSize:  " << dataSize << endl
       << "form:      " << lwTag.getString(form) << endl
;
}

void
LwFile::printTree() {
  cout << "LwFile tree - " << filename << endl;
  size_t i;
  for (i=0;i<chunks.size();i++) {
    chunks[i]->printTree( string ("  |") );
  }
}

void
LwFile::printTreeFull() {
  cout << "  LwFile printTreeFull - " << filename << endl;
  print();
  size_t i;
  for (i=0;i<chunks.size();i++) {
    chunks[i]->printTreeFull();
  }
}


////////////////////////////////////////
// getChunk
////////////////////////////////////////

// FIX: make this a static method of LwChunk
// returns 0 on failure
// buf points to the beginning of the TAG
// len is to the end of the file.
LwChunk *
LwFile::getChunk(const char *buf, const size_t len) {
  assert (buf);
  assert (len>=8); // 4 byte tag and 4 byte length
  LwChunk *c=0;

  const LwTagEnum t = lwTag.getEnum(buf,len);
  if (!lwTag.isGroupCHUNK(t)) {cerr<<"ERROR: expected CHUNK group tag"<<endl;return (0);}

  size_t tagLen = getU4(buf+4,len-4);
  //cout << "getChunk: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;

  // This is the main point of dependencies for all the CHUNKS...
  // Dispatch the right allocator

  switch (t) {
  case LAYR: c = new LwChunkLayr(buf,tagLen,true); assert(c); break;
  case PNTS: c = new LwChunkPnts(buf,tagLen,true); assert(c); break;
  case VMAP: c = new LwChunkVmap(buf,tagLen,true); assert(c); break;
  case POLS: c = new LwChunkPols(buf,tagLen,true); assert(c); break;
  case TAGS: c = new LwChunkTags(buf,tagLen,true); assert(c); break;
  case PTAG: c = new LwChunkPtag(buf,tagLen,true); assert(c); break;
  case VMAD: c = new LwChunkVmad(buf,tagLen,true); assert(c); break;
  case ENVL: cerr << "FIX: implement tag - ENVL" << endl;     break;
  case CLIP: c = new LwChunkClip(buf,tagLen,true); assert(c); break;
  case SURF: c = new LwChunkSurf(buf,tagLen,true); assert(c); break;
  case BBOX: c = new LwChunkBbox(buf,tagLen,true); assert(c); break;
  case DESC: cerr << "FIX: implement tag - DESC" << endl;     break;
  case TEXT: cerr << "FIX: implement tag - TEXT" << endl;     break;
  default:
    cerr << "ERROR: not a CHUNK tag - " << lwTag.getString(t) << endl;
    return (0);
  }

  if (!c) {
    cout << "getChunk: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;
    cout << "ERROR: did not set the Chunk, here is the first bit of the chunk..." << endl;
    printBuf(buf,16);
    cout << "I give up!" << endl;
    exit(EXIT_FAILURE);
  }

  assert (ERROR_LwTag!=c->tag);

  return c;
} // LwFile::getChunk


LwChunk *
LwFile::getChunkPtr(const LwTagEnum t) {
  if (!lwTag.isGroupCHUNK(t)) {
	cerr << "LwFile::getChunkPtr ERROR: this is not a chunk tag - "
	     << lwTag.getString(t) << endl;
	return (0);
  }
  size_t i;
  for (i=0;i<chunks.size();i++) {
    if (t==chunks[i]->tag) return (chunks[i]);
  }
  cerr << "LwFile::getChunkPtr ERROR: did not find chunk - "
       << lwTag.getString(t) << endl;
  return (0); // Not found
}



//////////////////////////////////////////////////////////////////////
// Chunk
//////////////////////////////////////////////////////////////////////
LwChunk::LwChunk() {
  //cout << "LwChunk::LwChunk() constructor" << endl;
  handleMemory();
}

LwChunk::LwChunk(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunk::LwChunk(const char *buf, const size_t len, bool mmapped)" << endl;
  
}

// If useData is true, then we are allowed use data to create the
// buffer for writing to disk
// FIX: this method probably only works if useData is true and data is set.

// FIX: need to return a vector <string *> so we are not pushing so much memory around
std::string LwChunk::setStr(bool useData) {
  if (useData && data) {
    string s(data,dataSize);
    if (s.length()%2 == 1) {
      cerr << "LwChunk::setStr WARNING: data block has odd length!!!  Need to add pad" << endl;
      // s = s + string()
      // FIX: need to insert new size U2 into data after tag
    }
    return (s);
  } 
  if (useData && !data) {
    cerr << "LwChunk::setStr WARNING: useData is true, but buf is 0" << endl;
  }
  cerr << "LwChunk::setStr WARNING: defaulting to no data" << endl;
  string s = setStrID4(tag);
  s += setStrU4(0); // Set length to be 0
  return (s);
}


#ifdef REGRESSION_TEST

static void dumpStrs(const string s1, const string s2) {
  size_t i;
  cout << "  DUMP STRING:"<< endl;
  cout << "  s1.length()=" << s1.length() << "  s2.length()="<<s2.length()<<endl;
  for (i=0;i<s1.length()&&i<s2.length();i++) {
    cout << "  "  << ((s1[i]==s2[i])?" ":"-")
	 << " "   << i << ": " << (isprint(s1[i])?s1[i]:'*') << " - " << int(s1[i]) 
	 << "   " << i << ": " << (isprint(s2[i])?s2[i]:'*') << " - " << int(s2[i]) << endl;
  }
}

// Just for regression test
bool LwChunk::checkSetStr() {
  bool ok = true;
  if (!data) { cerr<<"checkSetStr ERR: data was not set!"<<endl;}
  string s1(data,dataSize);
  string s2=setStr(false);
  string s3=setStr(true);
  if (0!=s1.compare(s2)){
    cerr<<"  1 checkSetStr for "<<lwTag.getString(tag)<<": FAILED!"<<endl;
    dumpStrs(s1,s2);
    ok=false;
  }
  if (0!=s1.compare(s2)){
    cerr<<"  2 checkSetStr FAILED for "<<lwTag.getString(tag)<<": FAILED!"<<endl;
    ok=false;
  }
  return(ok);
}

#endif // REGRESSION_TEST


void LwChunk::handleMemory() {
  tag = ERROR_LwTag;
  dataMMap = false;
  data=0;
  dataSize=0;
  start=0;
  length=0;
  pad = false;
}

// Helper for constructor
void LwChunk::handleMemory(const char *buf, const size_t len, bool mmapped) {
  assert(buf);
  assert(len>=8);
  dataMMap = mmapped;
  data = const_cast<char *>(buf);
  //dataSize=len;
  dataSize = getU4(buf+4,len-4) + 8;
  start = data+8;
  length=dataSize-8; // Length of tag
}

LwChunk::~LwChunk() {
  // Nothin to do.
  //size_t i; for (i=0;i<subChunks.size();i++) delete subChunks[i];
  //subChunks.clear();
}

void LwChunk::print() {
  cout << "  LwChunk::print()" << endl
       << "    TAG:      " << lwTag.getString(tag) << endl
       << "    dataSize: " << dataSize << "       total tag length" << endl
       << "    length:   " << length << "        payload in tag (no tag or len)" <<endl
       << "    pad:      " << (pad?"yes":"no") << endl
    ;
}

void LwChunk::printTree(const std::string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) << endl;
}

void LwChunk::printTreeFull() {
  print();
  // classes with SubChunks need to override this
}

// Chunks need to override this if they have subChunks
LwSubChunk *
LwChunk::getSubChunkPtr(const LwTagEnum t) {
  return (0); // Defaults is to have no subchunks.
}

//////////////////////////////////////////////////////////////////////
// CHUNK SUB-CLASSES
//////////////////////////////////////////////////////////////////////

LwChunkTags::LwChunkTags(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("Tags constructor"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;
  while (curLen>0 && curLen < 20000) {
    string s = getS0(cur,curLen);
    tag_strings.push_back(s);
    cur += getS0Len(s); curLen -= getS0Len(s);
    //cout << "  " << curLen << " " << getS0Len(s) << "  '" << s << "'"<<endl;
  }
  assert(0==curLen);
  tag = TAGS;
}

std::string LwChunkTags::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkTags WARN: useData true, buf is 0" << endl;
  //string s = setStrS0(tag_string);
  string s;
  size_t i; for (i=0;i<tag_strings.size();i++) s+=setStrS0(tag_strings[i]);
  s = setStrU4(s.length()) + s;
  s = setStrID4(tag) + s;
  return (s);
}

void LwChunkTags::print() {
  cout << "  LWChunkTags::print() " << endl;
  size_t i; for (i=0;i<tag_strings.size();i++) cout << "    " << i << " - " << tag_strings[i]<<endl;
}


void LwChunkTags::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << "  (" << tag_strings.size() << ")" << endl;
}

//////////
// LAYR
//////////
LwChunkLayr::LwChunkLayr(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkLayr constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;
  
  number = getU2(cur,curLen); cur += sizeof(U2); curLen -= sizeof(U2);
  flags  = getU2(cur,curLen); cur += sizeof(U2); curLen -= sizeof(U2);
  assert (12==sizeof(VEC12));
  {
    VEC12 p(cur,curLen); pivot = p; cur += sizeof(VEC12); curLen -= sizeof(VEC12);
  }
  name = getS0(cur,curLen); cur += getS0Len(name); curLen -= getS0Len(name);
  //cout << "Checking for parent: curLen = " << curLen << endl;
  // FIX: make sure the parent code works
  if (curLen>=2) {
    parent_given = true;
    parent = getU2(cur,curLen); cur +=- sizeof(U2); curLen -= sizeof(U2);
    //cout << "Found parent: " << parent << endl;
  } else {
    //cout << "NO parent" << endl;
    parent_given = false;
  }
  switch (curLen) {
  case 0: pad=false; break;
  case 1: pad=true;  break;
  default:
    cout << "ERROR: too many bytes at the end of LAYR:" << curLen << endl;
  }
  //cout << "pad:" << (pad?"yes":"no") << endl;

  // FIX: ?need to check for pad byte?
  tag = LAYR;
}


std::string LwChunkLayr::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkLayr WARN: useData true, buf is 0" << endl;
  string s = setStrU2(number) + setStrU2(flags) + pivot.setStr() + setStrS0(name);
  if (parent_given) s += setStrU2(parent);

  s = setStrID4(tag) + setStrU4(s.length()) + s;
  return (s);
}

void LwChunkLayr::print() {
  cout << "  LWChunkLayr::print() " 
       << "    number: " << number << endl
       << "    flags:  " << flags<< endl
       << "    pivot:  "; pivot.print();
  cout << "    name:" << name << endl
       << "    parent_given: " << (parent_given?"yes":"no") << endl;
  if (parent_given) cout << "    parent: " << parent << endl;
  cout << endl;
}

//////////
// PNTS
//////////

LwChunkPnts::LwChunkPnts(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkPnts constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;
  while (curLen >= sizeof(VEC12)) {
    point_location.push_back(VEC12(cur,curLen)); cur += sizeof(VEC12); curLen -= sizeof(VEC12);
  }
  cout << "  LwChunkPnts: num of point locations read: " << point_location.size() << endl;

  tag = PNTS;
} // LwChunkPnts constructor


std::string LwChunkPnts::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkPnts WARN: useData true, buf is 0" << endl;
  string s;
  size_t i;
  for (i=0;i<point_location.size();i++) {
    s += point_location[i].setStr();
  }
  s = setStrID4(tag) + setStrU4(s.length()) + s;
  return (s);
}

void
LwChunkPnts::print() {
  cout << "  LwChunkPnts::print()" << endl;
  size_t i;
  for (i=0; i<point_location.size(); i++) {
    cout << "    point_location("<<i<<"): "; point_location[i].print();
    if (i>5) { cout << "    more than 5... stopping early"<< endl;break;}
  }
  //LwChunk::print();
}

void LwChunkPnts::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << "  (" << point_location.size() << ")" << endl;
}


//////////
// VMAP
//////////

// VMAP { type[ID4], dimension[U2],  name[S0],
// ( vert[VX], value[F4] # dimension )* }
LwChunkVmap::LwChunkVmap(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkVmap constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;

  type = lwTag.getEnum(cur,curLen); cur+=4; curLen+=4;
  assert(lwTag.isGroupVMAP(type));

  dimension = getU2(cur,curLen); cur+=sizeof(U2); curLen-=sizeof(U2);
  name = getS0(cur,curLen);
  cur += getS0Len(name);  curLen -= getS0Len(name);

  size_t i;
  for (i=0;i<dimension;i++) {
    assert(curLen>=6);
    U4 vert = getVX(cur,curLen);{size_t vsize=getVXLen(cur,curLen); cur+=vsize;curLen+=vsize;}
    F4 value = getF4(cur,curLen); cur+=sizeof(F4); curLen-=sizeof(F4);
    verts.push_back(vert);
    values.push_back(value);
  }
  assert(verts.size()==values.size());
  cout << "Read VMAP: num values = " << verts.size() 
       << "  type = " << lwTag.getString(type) << " name = " << name << endl;
  tag = VMAP;
} // LwChunkVmap constructor

void
LwChunkVmap::print() {
  cout << "  LwChunkVmap::print()" << endl;

}

void LwChunkVmap::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << "  (" << values.size() << ")" << endl;
}

//////////
// VMAD
//////////

// DISCONTINUOUS VERTEX MAPPING

// VMAD { type[ID4], dimension[U2],  name[S0],
// ( vert[VX], poly[VX], value[F4]  # dimension )* }
LwChunkVmad::LwChunkVmad(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkVmad constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;

  type = lwTag.getEnum(cur,curLen); cur+=4; curLen+=4;
  assert(lwTag.isGroupVMAP(type));  // Tag type is the same as VMAP

  dimension = getU2(cur,curLen); cur+=sizeof(U2); curLen-=sizeof(U2);
  name = getS0(cur,curLen);
  cur += getS0Len(name);  curLen -= getS0Len(name);

  size_t i;
  for (i=0;i<dimension;i++) {
    assert(curLen>=6);
    U4 vert = getVX(cur,curLen);{size_t vsize=getVXLen(cur,curLen); cur+=vsize;curLen+=vsize;}
    U4 poly = getVX(cur,curLen);{size_t vsize=getVXLen(cur,curLen); cur+=vsize;curLen+=vsize;}
    F4 value = getF4(cur,curLen); cur+=sizeof(F4); curLen-=sizeof(F4);
    verts.push_back(vert);
    polys.push_back(poly);
    values.push_back(value);
  }
  assert(verts.size()==values.size());
  assert(verts.size()==polys.size());
  cout << "Read VMAD: dim="<<dimension<<" num values = " << verts.size() 
       << "  type = " << lwTag.getString(type) << " name = " << name << endl;
  tag = VMAD;
} // LwChunkVmad constructor

void
LwChunkVmad::print() {
  cout << "  LwChunkVmad::print()" << endl;

}

void LwChunkVmad::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << "  (" << values.size() << ")" << endl;
}





//////////
// BBOX
//////////

LwChunkBbox::LwChunkBbox(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkBbox constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;
  min = VEC12(cur,curLen); cur += sizeof(VEC12); curLen -= sizeof(VEC12);
  max = VEC12(cur,curLen); cur += sizeof(VEC12); curLen -= sizeof(VEC12);

  tag = BBOX;
} // LwChunkBbox::LwChunkBbox(const char *buf, const size_t len, bool mmapped) 

std::string LwChunkBbox::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkBbox WARN: useData true, buf is 0" << endl;
  string s = min.setStr() + max.setStr();
  s = setStrID4(tag) + setStrU4(s.length()) + s;
  return (s);
}

void
LwChunkBbox::print() {
  cout << "  LwChunkBbox::print()" << endl;
  cout << "    min: "; min.print();
  cout << "    max: "; max.print();
  //LwChunk::print();
}

//////////
// VertexSet
//////////

// ( numvert+flags[U2],  vert[VX] # numvert ) 
// Used in POLS

VertexSet::VertexSet(const char *buf, const size_t len) {
  //cout << "VertexSet constructor" << endl;
  assert(buf);
  assert(len>=2); // at least U2 count

  char *cur=const_cast<char *>(buf);
  size_t curLen = len;

  dataSize = 0;

  numvert = getU2(cur,curLen) & 0x3FF;
  flags = getU2(cur,curLen) & (~0x3FF);  // ~ is bitwise not... give the opposite mask
  cur += sizeof(U2); curLen -= sizeof(U2);
  dataSize += 2;

  size_t i;
  for (i=0;i<numvert;i++) {
    assert(curLen>=2);
    vert.push_back(getVX(cur,curLen));
    size_t size = getVXLen(cur,curLen);
    cur += size;  curLen += size;  dataSize +=size;
  }

  //cout << "num vertices read: " << vert.size() << "  dataSize: " << dataSize << endl;
  assert (numvert==vert.size());
} // VertexSet constructor


std::string VertexSet::setStr() {
  string s = setStrU2(U2(numvert)|flags);
  size_t i;
  for (i=0;i<vert.size();i++) {
    s += setStrVX(vert[i]);
  }
  return (s);
}

void VertexSet::print() {
  cout << "      VertexSet::print()" << endl
       << "        dataSize: " << dataSize << endl
       << "        numvert:  " << numvert << endl
       << "        flags:    " << flags << endl
       << "        Vertices: " << vert.size() << endl
    ;
  size_t i;
  for (i=0;i<vert.size();i++) {
    if (i>5) {cout << "      more than 5... stopping early" << endl;break;}
    cout << "  " << vert[i];
  }
  cout << endl;
}



//////////
// POLS
//////////

LwChunkPols::LwChunkPols(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkPols constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;

  type = lwTag.getEnum(start,curLen);
  cur+=4; curLen -=4;

  while (curLen>0) {
    VertexSet vs (cur,curLen);
    vertexSets.push_back(vs);
    cur += vs.getDataSize(); curLen -= vs.getDataSize();
  }
  cout << "num vertex sets read: " << vertexSets.size() << endl;

  tag = POLS;
} // LwChunkPols constructor


std::string LwChunkPols::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkPols WARN: useData true, buf is 0" << endl;
  string s = setStrID4(type);
  size_t i;
  for (i=0;i<vertexSets.size();i++) {
    s += vertexSets[i].setStr();
  }
  s = setStrID4(tag) + setStrU4(s.length()) + s;
  return (s);
}

void
LwChunkPols::print() {
  cout << "  LwChunkPols::print()" << endl
       << "    type: " << lwTag.getString(type) << endl
       << "    num vertex sets: " << vertexSets.size() << endl;
  size_t i;
  for (i=0;i<vertexSets.size();i++) {
    if (i>5) {cout << "    more than 5..."<<endl; break;}
    cout << "    ---- Set " << i << " ---"<< endl;
    vertexSets[i].print();
  }
  //LwChunk::print();
}

void LwChunkPols::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << "  (" << vertexSets.size() << ")" << endl;
}

//////////
// PolySet
//////////

// ( numvert+flags[U2],  vert[VX] # numvert ) 
// Used in POLS

PolySet::PolySet(const char *buf, const size_t len) {
  //cout << "PolySet constructor" << endl;
  assert(buf);
  assert(len>=2); // at least U2 count

  dataSize = 0;
  poly = getVX(buf,len);
  dataSize += getVXLen(buf,len);

  tag = getU2(buf+dataSize, len-dataSize);
  dataSize += 2;

  //cout << "  dataSize: " << dataSize << endl;
} // PolySet constructor


std::string PolySet::setStr() {
  string s = setStrVX(poly)+setStrU2(tag);
  return (s);
}

void PolySet::print() {
  cout << "      PolySet::print()" << endl
       << "        poly:  " << poly << endl
       << "        tag:  " << tag << endl
       << "        dataSize: " << dataSize << endl;
}

//////////
// PTAG
//////////

// Just says that each poly is a FACE

// { type[ID4], ( poly[VX], tag[U2] )* }
LwChunkPtag::LwChunkPtag(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkPtag constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;

  type = lwTag.getEnum(start,curLen);
  cur+=4; curLen -=4;

  while (curLen>0) {
    PolySet ps (cur,curLen);
    polySets.push_back(ps);
    cur += ps.getDataSize(); curLen -= ps.getDataSize();
  }
  cout << "num poly sets read: " << polySets.size() << endl;

  tag = PTAG;
} // LwChunkPtag constructor

std::string LwChunkPtag::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkPtag WARN: useData true, buf is 0" << endl;
  string s = setStrID4(type);
  size_t i;
  for (i=0;i<polySets.size();i++) {
    s += polySets[i].setStr();
  }
  s = setStrID4(tag) + setStrU4(s.length()) + s;
  return (s);
}

void
LwChunkPtag::print() {
  cout << "  LwChunkPtag::print()" << endl
       << "    type: " << lwTag.getString(type) << endl
       << "    num poly sets: " << polySets.size() << endl;
  size_t i;
  for (i=0;i<polySets.size();i++) {
    if (i>5) {cout << "      more than 5..." << endl; break;}
    cout << "    ---- Set " << i << " ---"<< endl;
    polySets[i].print();
  }
  //LwChunk::print();
}

void LwChunkPtag::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) << "  (" << polySets.size() << ")" << endl;
}


//////////
// SURF
//////////

//SURF { name[S0], source[S0], attributes[SUB-CHUNK] * }

LwChunkSurf::LwChunkSurf(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkSurf constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  //LwChunkSurf::print();

  char *cur=start;
  size_t curLen = length;

  name = getS0(cur,curLen);  //cout << "  name: " << name << " nameLen: " << getS0Len(name)<< endl;;
  cur += getS0Len(name);  curLen -= getS0Len(name);

  source = getS0(cur,curLen);  //cout << "  name: " << name << " nameLen: " << getS0Len(name)<< endl;;
  cur += getS0Len(source);  curLen -= getS0Len(source);
  //cout << "SOURCE= '"<<source<<"'  "<< getS0Len(source) <<endl;

  size_t subChunkTotSize=0;
  while (curLen>0) {
    LwSubChunk *sc = LwSubChunk::surfaceFactory(cur,curLen,mmapped);
    assert (sc);
    //cout << "LwChunkSurf: Adding this surface subchunk: " << endl;
    //sc->print();
    subChunks.push_back(sc);
    size_t size = sc->getDataSize();
    cur += size; curLen -= size;
    subChunkTotSize+=size;
  }
  assert (len==getS0Len(name)+2+subChunkTotSize);

  tag = SURF; // Danger: there are two things with the SURF tag (SURF has a SURF)
} // LwChunkSurf::LwChunkSurf(const char *buf, const size_t len, bool mmapped) 

LwChunkSurf::~LwChunkSurf() {
  //cout << "LwChunkSurf destructor" << endl;
  size_t i;
  for (i=0;i<subChunks.size();i++) delete subChunks[i];
}


std::string LwChunkSurf::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkSurf WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrS0(name);
  s += setStrS0(source);
  size_t i;
  for (i=0;i<subChunks.size();i++) {
    s += subChunks[i]->setStr(useData);
  }
  s = setStrID4(tag) + setStrU4(s.length()) + s;
  return (s);
}

void
LwChunkSurf::print() {
  cout << "  LwChunkSurf::print()" << endl
       << "    name: " << name << endl
       ;
#if 0
  size_t i;
  for (i=0;i<subChunks.size();i++) {
    cout << "    subChunk("<<i<<"):"<< endl;
    subChunks[i]->print();
  }
#endif
}

void LwChunkSurf::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) << endl;
  size_t i;
  for (i=0;i<subChunks.size();i++) {
    subChunks[i]->printTree( baseStr + string ("  |") );
  }
}

void
LwChunkSurf::printTreeFull() {
  print();
  size_t i;
  for (i=0;i<subChunks.size();i++) {
    cout << "    subChunk("<<i<<"):"<< endl;
    subChunks[i]->printTreeFull();
  }
}

LwSubChunk *
LwChunkSurf::getSubChunkPtr(const LwTagEnum t) {
  size_t i;
  for (i=0;i<subChunks.size();i++) if (t==subChunks[i]->tag) return(subChunks[i]);
  return (0); // Defaults is to have no subchunks.
}




//////////
// CLIP
//////////

// Image or Image Sequence
//   CLIP { index[U4], attributes[SUB-CHUNK] * }

LwChunkClip::LwChunkClip(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkClip constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;

  index = getU4(cur,curLen);
  cur += sizeof(U4);  curLen -= sizeof(U4);

  size_t subChunkTotSize=0;
  while (curLen>0) {
    LwSubChunk *sc = LwSubChunk::clipFactory(cur,curLen,mmapped);
    assert (sc);
    //cout << "LwChunkClip: Adding this clip subchunk: " << endl;
    //sc->print();
    subChunks.push_back(sc);
    size_t size = sc->getDataSize();
    cur += size; curLen -= size;
    subChunkTotSize+=size;
  }
#if 0
  cout << "subChunkTotSize="<< subChunkTotSize << "  len=" << len 
       << "  length="<< length<< " dataSize="<<dataSize<<endl
    ;
#endif

  tag = CLIP; 
} // LwChunkClip::LwChunkClip(const char *buf, const size_t len, bool mmapped) 

LwChunkClip::~LwChunkClip() {
  //cout << "LwChunkClip destructor" << endl;
  size_t i;
  for (i=0;i<subChunks.size();i++) delete subChunks[i];
}

std::string LwChunkClip::setStr(bool useData) {
  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkClip WARN: useData true, buf is 0" << endl;
  string s;
  //s = setStrID4(type);
  s += setStrU4(index);
  size_t i;
  for (i=0;i<subChunks.size();i++) s += subChunks[i]->setStr(useData);
  s = setStrID4(tag) + setStrU4(s.length()) + s;
  return (s);
}

void
LwChunkClip::print() {
  cout << "  LwChunkClip::print()" << endl
       << "    index: " << index << endl
       ;
  size_t i;
  for (i=0;i<subChunks.size();i++) {
    cout << "  subChunk("<<i<<"):"<< endl;
    subChunks[i]->print();
  }
  //LwChunk::print();
}

void LwChunkClip::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << " [" << subChunks.size() << "]"
       << endl;
  size_t i;
  for (i=0;i<subChunks.size();i++) {
    subChunks[i]->printTree( baseStr + string ("  |") );
  }
}

LwSubChunk *
LwChunkClip::getSubChunkPtr(const LwTagEnum t) {
  size_t i;
  for (i=0;i<subChunks.size();i++) if (t==subChunks[i]->tag) return(subChunks[i]);
  return (0); // Defaults is to have no subchunks.
}



//////////////////////////////////////////////////////////////////////
// SubChunk
//////////////////////////////////////////////////////////////////////

// Zeros everything out.
LwSubChunk::LwSubChunk() {
  //DebugPrintf(TRACE,("LwSubChunk constructor %s\n",lwTag.getString()));
  DebugPrintf(TRACE,("LwSubChunk constructor \n"));
  handleMemory();
}
LwSubChunk::~LwSubChunk() {
  //cout << "LwChunk::LwChunk() destructor" << endl;
}



// If useData is true, then we are allowed use data to create the
// buffer for writing to disk
// FIX: this method probably only works if useData is true and data is set.
std::string LwSubChunk::setStr(bool useData) {
  DebugPrintf(TRACE,("LwSubChunk::setStr %s.  SHOULD NOT BE CALLED\n",lwTag.getString(tag).c_str()));
  if (useData && data) {
    string s(data,dataSize);
    if (s.length()%2 == 1) {
      cerr << "LwSubChunk::setStr WARNING: data block has odd length!!!  Need to add pad" << endl;
      // s = s + string()
      // FIX: need to insert new size U2 into data after tag
    }
    return (s);
  } 
  if (useData && !data) {
    cerr << "LwSubChunk::setStr WARNING: useData is true, but buf is 0" << endl;
  }
  cerr << "LwSubChunk::setStr WARNING: defaulting to no data" << endl;
  string s = setStrID4(tag);
  s += setStrU2(0); // Set length to be 0
  return (s);
}


void
LwSubChunk::handleMemory() {
  tag = ERROR_LwTag;
  dataMMap = false;
  data=0;
  dataSize=0;
  start=0;
  length=0;
  pad=false;
}

// Header is tag[ID4] length[U2],  data[...], pad[U1] ?
// len does not include the 6 byte tag and len
void
LwSubChunk::handleMemory(const char *buf, const size_t len, bool mmapped) {
  assert(buf);
  //assert(len>=6);  len is tagLen which can be zero.
  dataMMap = mmapped;
  data = const_cast<char *>(buf);
  //dataSize = getU2(buf+4,len-4) + 6;
  dataSize = getU2(buf+4,2) + 6;
  start = data+6;
  length=dataSize-6; // Length of tag
  pad = false;
  // FIX: checkup on len:  should be that len==length
}

void
LwSubChunk::print() {
  cout << "    LwSubCunk::print()"<<endl
       << "      tag:" << lwTag.getString(tag) << endl
       << "      dataMMap: " << (dataMMap?"yes":"no")<< endl
       << "      dataSize: " << dataSize << endl
       << "      length:   " << length << endl
       << "      pad:      " << (pad?"yes":"no")<< endl;
}

void LwSubChunk::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) << endl;
}

void LwSubChunk::printTreeFull() {
  print();
}

LwSubChunk *
LwSubChunk::surfaceFactory(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::surfaceFactory \n"));
  assert(buf);
  assert(len>=6);
  LwSubChunk *sc=0;

  //printBuf(buf,20);
  const LwTagEnum t = lwTag.getEnum(buf,len);
  if (!lwTag.isGroupSURF(t)) {
    cerr<<"ERROR: expected SURF tag"<<endl
	<<"  got tag:" << lwTag.getString(t) << endl;
    return (0);
  }

  size_t tagLen = getU2(buf+4,len-4);
  //cout << "getSubChunk: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;

  // This is the main point of dependencies for all the CHUNKS...
  // Dispatch the right allocator

  switch (t) {
  case COLR: sc = new LwSubChunkColr(buf,tagLen,true); assert(sc); break;
  case DIFF: sc = new LwSubChunkBaseShadeVals(DIFF,buf,tagLen,true); assert(sc); break;
  case LUMI: sc = new LwSubChunkBaseShadeVals(LUMI,buf,tagLen,true); assert(sc); break;
  case SPEC: sc = new LwSubChunkBaseShadeVals(SPEC,buf,tagLen,true); assert(sc); break;
  case REFL: sc = new LwSubChunkBaseShadeVals(REFL,buf,tagLen,true); assert(sc); break;
  case TRAN: sc = new LwSubChunkBaseShadeVals(TRAN,buf,tagLen,true); assert(sc); break;
  case TRNL: sc = new LwSubChunkBaseShadeVals(TRNL,buf,tagLen,true); assert(sc); break;
  case GLOS: sc = new LwSubChunkGlos(buf,tagLen,true); assert(sc); break;
  case SHRP: cerr << "FIX: implement tag - SHRP" << endl; break;
  case BUMP: cerr << "FIX: implement tag - BUMP" << endl; break;
  case SIDE: sc = new LwSubChunkSide(buf,tagLen,true); assert(sc); break;
  case SMAN: sc = new LwSubChunkSman(buf,tagLen,true); assert(sc); break;
  case RFOP: sc = new LwSubChunkRfopTrop(RFOP,buf,tagLen,true); assert(sc); break;
  case RIMG: cerr << "FIX: implement tag - RIMG" << endl; break;
  case RSAN: cerr << "FIX: implement tag - RSAN" << endl; break;
  case RBLR: cerr << "FIX: implement tag - RBLR" << endl; break;
  case RIND: cerr << "FIX: implement tag - RIND" << endl; break;
  case TROP: sc = new LwSubChunkRfopTrop(TROP,buf,tagLen,true); assert(sc); break;
  case TIMG: cerr << "FIX: implement tag - TIMG" << endl; break;
  case TBLR: cerr << "FIX: implement tag - TBLR" << endl; break;
  case CLRH: cerr << "FIX: implement tag - CLRH" << endl; break;
  case CLRF: cerr << "FIX: implement tag - CLRF" << endl; break;
  case ADTR: cerr << "FIX: implement tag - ADTR" << endl; break;
  case GLOW: cerr << "FIX: implement tag - GLOW" << endl; break;
  case LINE: cerr << "FIX: implement tag - LINE" << endl; break;
  case ALPH: cerr << "FIX: implement tag - ALPH" << endl; break;
  case VCOL: cerr << "FIX: implement tag - VCOL" << endl; break;
  case BLOK: sc = new LwSubChunkBlok(buf,tagLen,true); assert(sc); break;
  default:
    cerr << "ERROR: not a SURF SubChunk tag - " << lwTag.getString(t) << endl;
    return (0);
  }

  if (!sc) {
    cout << "surfFactory: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;
    cout << "ERROR: did not set the Chunk, here is the first bit of the chunk..." << endl;
    printBuf(buf,16);
    cout << "I give up!" << endl;
    exit(EXIT_FAILURE);
  }
  assert (ERROR_LwTag!=sc->tag);
  return(sc);
} // LwSubChunk::surfaceFactory


//
// clipFactory
//

LwSubChunk *
LwSubChunk::clipFactory(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::clipFactory \n"));
  assert(buf);
  assert(len>=6);
  LwSubChunk *sc=0;

  //printBuf(buf,20);
  const LwTagEnum t = lwTag.getEnum(buf,len);
  if (!lwTag.isGroupCLIP(t)) {
    cerr<<"ERROR: expected CLIP tag"<<endl <<"  got tag:" << lwTag.getString(t) << endl;
    return (0);
  }

  size_t tagLen = getU2(buf+4,len-4);
  //cout << "getSubChunk: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;

  // This is the main point of dependencies for all the CHUNKS...
  // Dispatch the right allocator

  switch (t) {
  case STIL: sc = new LwSubChunkStil(buf,tagLen,true); assert(sc); break;
  case ISEQ: cerr << "FIX: implement tag - ISEQ" << endl; break;
  case ANIM: cerr << "FIX: implement tag - ANIM" << endl; break;
  case XREF: cerr << "FIX: implement tag - XREF" << endl; break;
  case STCC: cerr << "FIX: implement tag - STCC" << endl; break;
  default:
    cerr << "ERROR: not a CLIP SubChunk tag - " << lwTag.getString(t) << endl;
    return (0);
  }

  if (!sc) {
    cout << "clipFactory: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;
    cout << "ERROR: did not set the Chunk, here is the first bit of the chunk..." << endl;
    printBuf(buf,16);
    cout << "I give up!" << endl;
    exit(EXIT_FAILURE);
  }
  assert (ERROR_LwTag!=sc->tag);
  return(sc);
} // LwSubChunk::clipFactory

//
// factory -- Just lump the rest together, no real need to separate them out.
//

// Handles these groups: BLOK
LwSubChunk *
LwSubChunk::factory(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Factory\n"));
  assert(buf);
  assert(len>=6);
  LwSubChunk *sc=0;

  //printBuf(buf,5);
  const LwTagEnum t = lwTag.getEnum(buf,len);

  size_t tagLen = getU2(buf+4,len-4);
  //cout << "getSubChunk: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;

  // This is the main point of dependencies for all the CHUNKS...
  // Dispatch the right allocator

  switch (t) {
    //case STIL: 
    // BLOK HEADERS
  case IMAP: sc = new LwSubChunkBlokHeader(IMAP,buf,tagLen,true); 
    if (!sc) cerr << "Blokheader bailing...FIX" << endl;
    assert(sc); break;
  case PROC: cerr << "FIX: implement tag - PROC" << endl; break;
  case GRAD: cerr << "FIX: implement tag - GRAD" << endl; break;
  case SHDR: cerr << "FIX: implement tag - SHDR" << endl; break;

    // BLOK HEADER ATTRIBUTES
  case CHAN: sc = new LwSubChunkChan(buf,tagLen,true); assert(sc); break;
  case ENAB: sc = new LwSubChunkEnab(buf,tagLen,true); assert(sc); break;
  case OPAC: sc = new LwSubChunkOpac(buf,tagLen,true); assert(sc); break;
  case AXIS: sc = new LwSubChunkAxis(buf,tagLen,true); assert(sc); break;

    // NOT sure which group these really belong, CLIP or BLOK?
  case TIME: cerr << "FIX: implement tag - TIME" << endl; break;
  case CONT: cerr << "FIX: implement tag - CONT" << endl; break;
  case BRIT: cerr << "FIX: implement tag - BRIT" << endl; break;
  case SATR: cerr << "FIX: implement tag - SATR" << endl; break;
  case HUE: cerr << "FIX: implement tag - HUE" << endl; break;
  case GAMM: cerr << "FIX: implement tag - GAMM" << endl; break;
  case NEGA: sc = new LwSubChunkNega(buf,tagLen,true); assert(sc); break;
  case IFLT: cerr << "FIX: implement tag - IFLT" << endl; break;
  case PFLT: cerr << "FIX: implement tag - PFLT" << endl; break;

    // Maybe in BLOK?
  case TMAP: sc = new LwSubChunkTmap(buf,tagLen,true); assert(sc); break;

    // TMAP SubChunks
  case CNTR: sc = new LwSubChunkTmapPos(CNTR,buf,tagLen,true); assert(sc); break;
  case SIZE: sc = new LwSubChunkTmapPos(SIZE,buf,tagLen,true); assert(sc); break;
  case ROTA: sc = new LwSubChunkTmapPos(ROTA,buf,tagLen,true); assert(sc); break;
  case OREF: sc = new LwSubChunkOref(buf,tagLen,true); assert(sc); break;
  case FALL: sc = new LwSubChunkFall(buf,tagLen,true); assert(sc); break;
  case CSYS: sc = new LwSubChunkCsys(buf,tagLen,true); assert(sc); break;

    // IMAGE MAPS
  case PROJ: sc = new LwSubChunkProj(buf,tagLen,true); assert(sc); break;
  case WRAP: sc = new LwSubChunkWrap(buf,tagLen,true); assert(sc); break;
  case IMAG: sc = new LwSubChunkImag(buf,tagLen,true); assert(sc); break;
  case WRPW: sc = new LwSubChunkWrapAmount(WRPW,buf,tagLen,true); assert(sc); break;
  case WRPH: sc = new LwSubChunkWrapAmount(WRPH,buf,tagLen,true); assert(sc); break;
  case AAST: sc = new LwSubChunkAast(buf,tagLen,true); assert(sc); break;
  case PIXB: sc = new LwSubChunkPixb(buf,tagLen,true); assert(sc); break;

    // IMAGE MAPS ... there are two different VMAP tags... CHUNK and SUBCHUNK
  case VMAP:  sc = new LwSubChunkVmap(buf,tagLen,true); assert(sc); break;

  default:
    cerr << "ERROR: not a recognized SubChunk tag - " << lwTag.getString(t) << endl;
   return (0);
  }

  if (!sc) {
    cout << "factory: " << lwTag.getString(t) << "  tagLen=" << tagLen << endl;
    cout << "ERROR: did not set the Chunk, here is the first bit of the chunk..." << endl;
    printBuf(buf,16);
    cout << "I give up!" << endl;
    exit(EXIT_FAILURE);
  }
  assert (ERROR_LwTag!=sc->tag);
  return(sc);
} // LwSubChunk::factory




//////////
// COLR
//////////

// COLR { base-color[COL12], envelope[VX]  }
LwSubChunkColr::LwSubChunkColr(const char *buf, const size_t len, bool mmapped) {
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  base_color = COL12(start,length);
  assert (12==sizeof(COL12));
  envelope = getVX(start+sizeof(COL12), length-sizeof(COL12));
  tag = COLR; 
  DebugPrintf(VERBOSE,("LwSubChunk::COLR %s\n",lwTag.getString(tag).c_str()));
} // LwChunkColr const 

std::string LwSubChunkColr::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwSubChunkColr WARN: useData true, buf is 0" << endl;
  string s;
  s = base_color.setStr();
  s += setStrVX(envelope);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void
LwSubChunkColr::print() {
  cout << "    LwSubChunkColr::print()" << endl
       << "      base_color: "
       << "      ";  base_color.print();
  cout << "      envelope:   " << envelope << endl;
  //LwSubChunk::print();
}

/////////////////
// BASESHADEVALS
/////////////////

// Handle all these: DIFF, LUMI, SPEC, REFL, TRAN, TRNL
// WARNING: default diffuse value is 1.0, rest default to 0
LwSubChunkBaseShadeVals::LwSubChunkBaseShadeVals(const LwTagEnum t,
						 const char *buf, const size_t len, bool mmapped
						 )
{
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  intensity = getF4(start,length); // AKA FP4
  assert(0.<=intensity);
  assert(1.>=intensity);
  envelope = getVX(start+sizeof(F4), length-sizeof(F4));
  tag = t; 
  DebugPrintf(VERBOSE,("LwSubChunk:: baseshadevals constr %s\n",lwTag.getString(tag).c_str()));
} // LwChunkBaseShadeVals constructor

std::string LwSubChunkBaseShadeVals::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwSubChunkBaseShadeVals WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrFP4(intensity);
  s += setStrVX(envelope);
  s = setStrU2(s.length()) + s; // Set length to be 0
  s = setStrID4(tag) + s;
  return (s);
}

void LwSubChunkBaseShadeVals::print() {
  cout << "      LwSubChunkBaseShadeVals::print() - " << lwTag.getString(tag) << endl
       << "        intensity: " << intensity << endl
       << "        envelope:  " << envelope << endl;
  //LwSubChunk::print();
}

//////////
// STIL
//////////

// STIL { base-color[COL12], envelope[VX]  }
LwSubChunkStil::LwSubChunkStil(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkStil constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  name = getS0(start,length);
  tag = STIL; 
  DebugPrintf(VERBOSE,("LwSubChunk::Stil %s\n",lwTag.getString(tag).c_str()));

} // LwChunkStil::LwChunkStil(const char *buf, const size_t len, bool mmapped) 

LwSubChunkStil::~LwSubChunkStil() {
  //cout << "LwSubChunkStil destructor" << endl;
}

std::string LwSubChunkStil::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkStil WARN: useData true, buf is 0" << endl;
  string s = setStrS0(name);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void
LwSubChunkStil::print() {
  cout << "      LwSubChunkStil::print()" << endl
       << "        name:   " << name << endl;
}


//////////
// BLOK
//////////

// BLOK { base-color[COL12], envelope[VX]  }
LwSubChunkBlok::LwSubChunkBlok(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Blok\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;

  //cout << "BLOK FIX: what is this ordinal thing that does not seem to be there???" << endl;
  size_t subChunkTotSize=0;
  while (curLen>0) {
    LwSubChunk *sc = LwSubChunk::factory(cur,curLen,mmapped);
    assert (sc);
    //cout << "LwChunkBlok: Adding this surface subchunk: " << endl;
    //sc->print();
    if (!lwTag.isGroupBLOK(sc->tag)) {
      static bool warned=false;
      if (!warned) {
	warned = true;
	cout << " WARNING: tag is not in group BLOK!" << endl
	     << "   need to fully understand groupings." << endl
	     << "   TAG: " << lwTag.getString(sc->tag) << endl;
      }
    }
    attributes.push_back(sc);
    size_t size = sc->getDataSize();
    cur += size; curLen -= size;
    subChunkTotSize+=size;
  }

  tag = BLOK;
  DebugPrintf(VERBOSE,("LwSubChunk::Blok %s\n",lwTag.getString(tag).c_str()));
} // LwChunkBlok::LwChunkBlok(const char *buf, const size_t len, bool mmapped) 

LwSubChunkBlok::~LwSubChunkBlok() {
  //cout << "LwSubChunkBlok destructor" << endl;
  size_t i; for(i=0;i<attributes.size();i++) delete attributes[i];
  attributes.clear();
}

std::string LwSubChunkBlok::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkBlok WARN: useData true, buf is 0" << endl;
  string s;
  size_t i; for (i=0;i<attributes.size();i++) s += attributes[i]->setStr(useData);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void
LwSubChunkBlok::print() {
  cout << "      LwSubChunkBlok::print()" << endl
       << "        num block_attr: " << attributes.size() << endl;
  size_t i;
  for (i=0;i<attributes.size();i++) attributes[i]->print();

  //LwSubChunk::print();
}


void LwSubChunkBlok::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << " [" << attributes.size() << "]" << endl;
  size_t i;
  for (i=0;i<attributes.size();i++) {
    attributes[i]->printTree( baseStr + string ("  |") );
  }
}

//////////
// BlokHeader - IMAP, PROC, GRAD, SHDR
//////////


LwSubChunkBlokHeader::LwSubChunkBlokHeader(const LwTagEnum t,
					   const char *buf, const size_t len, const bool mmapped)
{
  DebugPrintf(VERBOSE,("LwSubChunk::BlockHeader\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  tag = t;

  char *cur=start;
  size_t curLen = length;
  ordinal = getS0(cur,curLen);
  cur += getS0Len(ordinal);  curLen -= getS0Len(ordinal);

  size_t subChunkTotSize=0;
  while (curLen>0) {
    LwSubChunk *sc = LwSubChunk::factory(cur,curLen,mmapped);
    assert (sc);
    if (!lwTag.isGroupBLOK(sc->tag) && !lwTag.isGroupMISC(sc->tag)) {
      cout << " WARNING: tag is not in group BLOK.  Probably ok." << endl;
      exit(EXIT_FAILURE);
    }
    block_attributes.push_back(sc);
    size_t size = sc->getDataSize();
    cur += size; curLen -= size;
    subChunkTotSize+=size;
  }
} // LwChunkBlokHeader constr

LwSubChunkBlokHeader::~LwSubChunkBlokHeader() {
  //cout << "LwSubChunkBlokHeader destructor" << endl;
  size_t i;
  for (i=0;i<block_attributes.size();i++) delete block_attributes[i];
}

std::string LwSubChunkBlokHeader::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkBlokHeader WARN: useData true, buf is 0" << endl;
  string s = setStrS0(ordinal);
  size_t i; for (i=0;i<block_attributes.size();i++) s += block_attributes[i]->setStr(useData);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void
LwSubChunkBlokHeader::print() {
  cout << "      LwSubChunkBlokHeader::print()" << endl
       << "        ordinal:        " << ordinal << endl
       << "        num block_attr: " << block_attributes.size() << endl;
  size_t i;
  for (i=0;i<block_attributes.size();i++) block_attributes[i]->print();

  //LwSubChunk::print();
}

void LwSubChunkBlokHeader::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << " [" << block_attributes.size() << "]"
       << endl;
  size_t i;
  for (i=0;i<block_attributes.size();i++) {
    block_attributes[i]->printTree( baseStr + string ("  |") );
  }
}



//////////
// CHAN
//////////

// CHAN { base-color[COL12], envelope[VX]  }
LwSubChunkChan::LwSubChunkChan(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Chan\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  texture_channel = lwTag.getEnum(start,length);
  if (!isValidChannel(texture_channel)) {
    cout << "LwSubChunkChan ERROR: texture_channel invalid" << endl;
  } else  tag = CHAN; 
} // LwChunkChan::LwChunkChan(const char *buf, const size_t len, bool mmapped) 

bool
LwSubChunkChan::isValidChannel(const LwTagEnum t) {
  switch (t) {
  case COLR: return (true);
  case DIFF: return (true);
  case LUMI: return (true);
  case SPEC: return (true);
  case GLOS: return (true);
  case REFL: return (true);
  case TRAN: return (true);
  case TRNL: return (true);
  case BUMP: return (true);
  default: cout << "isValidChannel false for " << lwTag.getString(t) << endl;
  }
  return(false);
}

std::string LwSubChunkChan::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkChan WARN: useData true, buf is 0" << endl;
  string s = lwTag.getString(texture_channel);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void
LwSubChunkChan::print() {
  cout << "      LwSubChunkChan::print()" << endl
       << "        texture_channel: " << lwTag.getString(texture_channel) << endl;
  //LwSubChunk::print();
}


//////////
// OPAC
//////////

// OPAC { base-color[COL12], envelope[VX]  }
LwSubChunkOpac::LwSubChunkOpac(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Opac\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  type = getU2(start,length);
  opacity = getF4(start+2,length-2);
  envelope = getVX(start+2+4, length-2-4);

  assert(0<=opacity);
  assert(1>=opacity);

  tag = OPAC; 
} // LwChunkOpac::LwChunkOpac(const char *buf, const size_t len, bool mmapped) 


std::string LwSubChunkOpac::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkOpac WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrU2(type);
  s += setStrFP4(opacity);
  s += setStrVX(envelope);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}


void
LwSubChunkOpac::print() {
  cout << "      LwSubChunkOpac::print()" << endl
       << "        type: " << type << endl
       << "        opacity: " << opacity << endl
       << "        envelope: " << envelope << endl
    ;
  //LwSubChunk::print();
}


//////////
// ENAB
//////////

// ENAB { base-color[COL12], envelope[VX]  }
LwSubChunkEnab::LwSubChunkEnab(const char *buf, const size_t len, bool mmapped) {
    DebugPrintf(VERBOSE,("LwSubChunk::Enab\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  U2 enableU2 = getU2(start,length);
  enable = (enableU2?true:false);
  //cout << "enable: " << enableU2 << endl;

  tag = ENAB; 
} // LwChunkEnab::LwChunkEnab(const char *buf, const size_t len, bool mmapped) 

std::string LwSubChunkEnab::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkEnab WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrU2(enable); // bool == U2
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkEnab::print() {
  cout << "      LwSubChunkEnab::print()" << endl
       << "        enable: " << (enable?"true":"false") << endl
    ;
  //LwSubChunk::print();
}



//////////
// NEGA
//////////

// NEGA { base-color[COL12], envelope[VX]  }
LwSubChunkNega::LwSubChunkNega(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Nega\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  U2 enableU2 = getU2(start,length);
  enable = (enableU2?true:false);

  tag = NEGA; 
  DebugPrintf(VERBOSE,("LwSubChunk::Nega done\n"));
} // LwChunkNega::LwChunkNega(const char *buf, const size_t len, bool mmapped) 

std::string LwSubChunkNega::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkNega WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrU2(enable); // bool == U2
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkNega::print() {
  cout << "      LwSubChunkNega::print()" << endl
       << "        enable: " << (enable?"true":"false") << endl
    ;
  //LwSubChunk::print();
}


//////////
// TMAP
//////////


LwSubChunkTmap::LwSubChunkTmap(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Tmap\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  char *cur=start;
  size_t curLen = length;

  //cout << "TMAP FIX: what is this ordinal thing that does not seem to be there???" << endl;
  size_t subChunkTotSize=0;
  while (curLen>0) {
    LwSubChunk *sc = LwSubChunk::factory(cur,curLen,mmapped);
    assert (sc);
    //cout << "LwChunkTmap: Adding this surface subchunk: " << endl;
    //sc->print();
    attributes.push_back(sc);
    size_t size = sc->getDataSize();
    cur += size; curLen -= size;
    subChunkTotSize+=size;
  }
  tag = TMAP;
} // LwChunkTmap constructor

LwSubChunkTmap::~LwSubChunkTmap() {
  size_t i; for (i=0;i<attributes.size();i++) delete attributes[i];
}

std::string LwSubChunkTmap::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkTmap WARN: useData true, buf is 0" << endl;
  string s;
  size_t i; for (i=0;i<attributes.size();i++) s += attributes[i]->setStr(useData);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}


void
LwSubChunkTmap::print() {
  cout << "      LwSubChunkTmap::print()" << endl;
  cout << "        num attributes: " << attributes.size() << endl;
  size_t i;
  for (i=0;i<attributes.size();i++) attributes[i]->print();
  //LwSubChunk::print();
}


void LwSubChunkTmap::printTree(const string baseStr) {
  cout << baseStr << "-" << lwTag.getString(tag) 
       << " [" << attributes.size() << "]"
       << endl;
  size_t i;
  for (i=0;i<attributes.size();i++) {
    attributes[i]->printTree( baseStr + string ("  |") );
  }
}

//////////
// TMAP-POS: CNTR,SIZE,ROTA
//////////


LwSubChunkTmapPos::LwSubChunkTmapPos(const LwTagEnum t,
				     const char *buf, const size_t len, bool mmapped)
{
  DebugPrintf(VERBOSE,("LwSubChunk::TmapPos %d\n",int(t)));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  tag = t;

  assert(sizeof(VEC12)==12);
  vector = VEC12(start,length);
  envelope = getVX(start+sizeof(VEC12),length-sizeof(VEC12));
} // LwChunkTmapPos constructor


std::string LwSubChunkTmapPos::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkTmapPos WARN: useData true, buf is 0" << endl;
  string s;
  s += vector.setStr();
  s += setStrVX(envelope);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void
LwSubChunkTmapPos::print() {
  cout << "      LwSubChunkTmapPos::print() - " << lwTag.getString(tag) << endl
       << "        vector:   "; vector.print();
  cout << "        envelope: " << envelope << endl;
  //LwSubChunk::print();
}



//////////
// FALL
//////////

// FALL { type[U2], vector[VEC12],  envelope[VX] }
LwSubChunkFall::LwSubChunkFall(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Fall\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers

  char *cur=start;
  size_t curLen = length;
  U2 type_tmp = getU2(cur, curLen); cur += sizeof(U2); curLen += sizeof(U2);
  vector = VEC12(cur,curLen); cur += sizeof(VEC12); curLen += sizeof(VEC12);
  envelope = getVX(cur,curLen);

  switch(type_tmp) {
  case 0: type=CUBIC;     break;
  case 1: type=SPHERICAL; break;
  case 2: type=LINEAR_X;  break;
  case 3: type=LINEAR_Y;  break;
  case 4: type=LINEAR_Z;  break;
  default:
    cerr << "FALL parse ERROR: type must be in range [0..4]" << endl
	 << "  got: " << type_tmp << endl;
    type = ERROR_LwTag;
  }
  tag = FALL; 
} // LwChunkFall::LwChunkFall(const char *buf, const size_t len, bool mmapped) 

//  FALL { type[U2], vector[VEC12],  envelope[VX] }
std::string LwSubChunkFall::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkFall WARN: useData true, buf is 0" << endl;
  string s;
  // FIX: This could easily be done by subtracting Cubic from the U2 of type;
  U2 typeU2;
  switch (type) {
  case CUBIC: typeU2 = 0; break;
  case SPHERICAL: typeU2 = 1; break;
  case LINEAR_X: typeU2 = 2; break;
  case LINEAR_Y: typeU2 = 3; break;
  case LINEAR_Z: typeU2 = 4; break;
  default:
    cerr << "FALL ERROR: type invalid"<<endl; assert(false);
  }
  s += setStrU2(typeU2);
  s += vector.setStr();
  s += setStrVX(envelope);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkFall::print() {
  cout << "      LwSubChunkFall::print()" << endl
       << "        type: " << lwTag.getString(type) << endl
       << "        vector: "; vector.print();
  cout << "        envelope: " << envelope << endl;
  //LwSubChunk::print();
}


//////////
// OREF
//////////

// OREF { type[U2], vector[VEC12],  envelope[VX] }
LwSubChunkOref::LwSubChunkOref(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::Oref\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  object_name = getS0(start,length);

  tag = OREF; 
} // LwChunkOref::LwChunkOref(const char *buf, const size_t len, bool mmapped) 

std::string LwSubChunkOref::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkOref WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrS0(object_name);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkOref::print() {
  cout << "      LwSubChunkOref::print()" << endl
       << "        object_name: " << object_name << endl;
  //LwSubChunk::print();
}

//////////
// CSYS
//////////

// CSYS { type[U2], vector[VEC12],  envelope[VX] }
LwSubChunkCsys::LwSubChunkCsys(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::csys\n"));

  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  U2 type_tmp = getU2(start, length);

  switch(type_tmp) {
  case 0: type=OBJ_COORD;     break;
  case 1: type=WORLD_COORD; break;
  default:
    cerr << "FALL parse ERROR: type must be in range [0..1]" << endl
	 << "  got: " << type_tmp << endl;
    type = ERROR_LwTag;
  }

  tag = CSYS; 
} // LwChunkCsys::LwChunkCsys(const char *buf, const size_t len, bool mmapped) 

std::string LwSubChunkCsys::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkCsys WARN: useData true, buf is 0" << endl;
  string s;
  U2 val = U2(type)-U2(OBJ_COORD);
  assert (0==val || 1==val);
  s += setStrU2 (val);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkCsys::print() {
  cout << "      LwSubChunkCsys::print()" << endl
       << "        type: " << lwTag.getString(type) << endl;
  //LwSubChunk::print();
}



//////////
// PROJ
//////////

// PROJ { type[U2], vector[VEC12],  envelope[VX] }
LwSubChunkProj::LwSubChunkProj(const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkProj constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  U2 type_tmp = getU2(start,length);

  switch(type_tmp) {
  case 0: projection_mode=PLANAR_PROJ;      break;
  case 1: projection_mode=CYLINDRICAL_PROJ; break;
  case 2: projection_mode=SPHERICAL_PROJ;   break;
  case 3: projection_mode=CUBIC_PROJ;       break;
  case 4: projection_mode=FRONT_PROJ;       break;
  case 5: projection_mode=UV_PROJ;          break;
  default:
    cerr << "PROJ parse ERROR: type must be in range [0..5]" << endl
	 << "  got: " << type_tmp << endl;
    projection_mode = ERROR_LwTag;
  }
  tag = PROJ; 
} // LwChunkProj::LwChunkProj(const char *buf, const size_t len, bool mmapped) 


std::string LwSubChunkProj::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkProj WARN: useData true, buf is 0" << endl;
  string s;
  U2 val = U2(projection_mode)-U2(PLANAR_PROJ);
  assert (U2(5)>=val); // U2(0)<=val is always true
  s += setStrU2 (val);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkProj::print() {
  cout << "      LwSubChunkProj::print()" << endl
       << "        type: " << lwTag.getString(projection_mode) << endl;
}




//////////
// AXIS
//////////

// AXIS  U2
LwSubChunkAxis::LwSubChunkAxis(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::axis\n"));
  //cout << "LwChunkAxis constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  U2 axis_tmp = getU2(start,length);

  switch(axis_tmp) {
  case 0: axis=X_AXIS;      break;
  case 1: axis=Y_AXIS; break;
  case 2: axis=Z_AXIS;   break;
  default:
    cerr << "AXIS parse ERROR: type must be in range [0..5]" << endl
	 << "  got: " << axis_tmp << endl;
    axis = ERROR_LwTag;
  }
  tag = AXIS; 
} // LwChunkAxis::LwChunkAxis(const char *buf, const size_t len, bool mmapped) 

std::string LwSubChunkAxis::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkAxis WARN: useData true, buf is 0" << endl;
  string s;
  U2 val = U2(axis)-U2(X_AXIS);
  assert (2>=val);  // U2(5)<=val is always true
  s += setStrU2 (val);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}


void 
LwSubChunkAxis::print() {
  cout << "      LwSubChunkAxis::print()" << endl
       << "        axis: " << lwTag.getString(axis) << endl;
}


//////////
// IMAG
//////////

// IMAG  U2
LwSubChunkImag::LwSubChunkImag(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::imag\n"));
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  texture_image = getVX(start,length);

  tag = IMAG; 
} // LwChunkImag constructor

std::string LwSubChunkImag::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkImag WARN: useData true, buf is 0" << endl;
  string s = setStrVX(texture_image);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkImag::print() {
  cout << "      LwSubChunkImag::print()" << endl
       << "        texture_image: " << texture_image << endl;
  //LwSubChunk::print();
}


//////////
// WRAP
//////////

LwTagEnum LwSubChunkWrap::getWrapType(U2 val) {
  switch (val) {
  case 0: return (RESET_WRAP);
  case 1: return (REPEAT_WRAP);
  case 2: return (MIRROR_WRAP);
  case 3: return (EDGE_WRAP);
  default:
    cerr << "Wrap parse ERROR: val not in range [0..3]" << endl;
  }
  return (ERROR_LwTag);
}

// WRAP  U2
LwSubChunkWrap::LwSubChunkWrap(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::wrap\n"));
  //cout << "LwChunkWrap constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  
  width_wrap  = getWrapType(getU2(start,length));
  height_wrap = getWrapType(getU2(start+2,length-2));

  tag = WRAP; 
} // LwChunkWrap::LwChunkWrap(const char *buf, const size_t len, bool mmapped) 

std::string LwSubChunkWrap::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkWrap WARN: useData true, buf is 0" << endl;
  string s;
  U2 w = U2(width_wrap)-U2(RESET_WRAP);
  assert (3>=w);  // U2(0)<=val is always true
  s += setStrU2 (w);
  U2 h = U2(height_wrap)-U2(RESET_WRAP);
  assert (3>=h);  // U2(0)<=val is always true
  s += setStrU2 (h);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkWrap::print() {
  cout << "      LwSubChunkWrap::print()" << endl
       << "        width_wrap:  " << lwTag.getString(width_wrap)  << endl
       << "        height_wrap: " << lwTag.getString(height_wrap) << endl;
}


// WRAP AMOUNT 
LwSubChunkWrapAmount::LwSubChunkWrapAmount(const LwTagEnum t,
					   const char *buf, const size_t len, bool mmapped) {
  //cout << "LwChunkWrap constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  
  cycles = getF4(start,length); // FIX: check FP4
  envelope = getVX(start+4, length-4);
  tag = t; 
} // LwChunkWrap::LwChunkWrap(const char *buf, const size_t len, bool mmapped) 


std::string LwSubChunkWrapAmount::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkWrap WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrFP4(cycles);
  s += setStrVX(envelope);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkWrapAmount::print() {
  cout << "      LwSubChunkWrapAmount::print() - " << lwTag.getString(tag) << endl
       << "        cycles:  " << envelope << endl
       << "        envelope:  " << envelope << endl;
}


// AAST 
LwSubChunkAast::LwSubChunkAast(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::aast\n"));
  //cout << "LwChunkAast constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  flags = getU2(start,length);
  strength = getF4(start+2,length-2);
  tag = AAST; 
} // LwChunkAast::LwChunkAast(const char *buf, const size_t len, bool mmapped) 

std::string LwSubChunkAast::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkAast WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrU2(flags);
  s += setStrFP4(strength);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkAast::print() {
  cout << "      LwSubChunkAast::print()" << endl
       << "        flags: " << flags << endl
       << "        strength: " << strength << endl;
}


// PIXB 
LwSubChunkPixb::LwSubChunkPixb(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::pixb\n"));
  //cout << "LwChunkPixb constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  flags = getU2(start,length);
  tag = PIXB; 
} // LwSubChunkPixb constr

std::string LwSubChunkPixb::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkPixb WARN: useData true, buf is 0" << endl;
  string s;
  s += setStrU2(flags);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkPixb::print() {
  cout << "      LwSubChunkPixb::print()" << endl
       << "        flags: " << flags << endl;
}

// VMAP 
LwSubChunkVmap::LwSubChunkVmap(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::vmap chunk\n"));
  //cout << "LwChunkVmap constructor" << endl;
  assert(buf);
  handleMemory(buf,len,mmapped);  // set all the pointers
  txuv_map_name = getS0(start,length);
  tag = VMAP;  // FIX: give it a different tag?  VMAP_SUB???
} // LwSubChunkVmap  constructor

std::string LwChunkVmap::setStr(bool useData) {
  cout << "FIX: vmap broken for Zareh/McKittrick_nm_crop2-Morrored-reduced2.lwo"<<endl;

  if (useData && data) return (LwChunk::setStr(useData));
  if (useData && !data) cerr << "LwChunkVmap WARN: useData true, buf is 0" << endl;
  string s;
  s = setStrID4(type);
  s += setStrU2(dimension);
  s += setStrS0(name);

  size_t i;
  for (i=0;i<verts.size();i++) s += setStrVX(verts[i]);
  for (i=0;i<values.size();i++) s += setStrF4(values[i]);

  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}


void 
LwSubChunkVmap::print() {
  cout << "      LwSubChunkVmap::print()" << lwTag.getString(tag) << endl
       << "        txuv_map_name: " << txuv_map_name << endl;
}

// RFOP 
LwSubChunkRfopTrop::LwSubChunkRfopTrop(const LwTagEnum t,
			       const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::rfop\n"));
  assert(buf);
  assert (RFOP==t || TROP==t);
  handleMemory(buf,len,mmapped);  // set all the pointers
  U2 optVal = getU2(start,length);
  switch(optVal) {
  case 0: options = BACKDROP_ONLY; break;
  case 1: options = RAYTRACE_BACKDROP; break;
  case 2: options = SPHERICAL_MAP; break;
  case 3: options = RAYTRACE_SPHERICAL_MAP; break;
  default: cerr<<"RFOP parse ERROR: option not in [0..3]"<<endl; options=ERROR_LwTag;
  }
  tag = t; 
} // LwChunkRfopTrop constructor


std::string LwSubChunkRfopTrop::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwSubChunkRfopTrop WARN: useData true, buf is 0" << endl;
  string s;
  //s = setStrID4(options);
  U2 val = options - U2(BACKDROP_ONLY);
  assert (val <= 3);
  s = setStrU2(val);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkRfopTrop::print() {
  cout << "      LwSubChunkRfop::print()" << lwTag.getString(tag) << endl
       << "        options: " << lwTag.getString(options) << endl;
  //LwSubChunk::print();
}

// SIDE 
LwSubChunkSide::LwSubChunkSide(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::side\n"));
  assert(buf); assert(len>=2);
  handleMemory(buf,len,mmapped);  // set all the pointers
  U2 sidednessVal = getU2(start,length);
  switch(sidednessVal) {
  case 1: sidedness = FRONT_ONLY; break;
  case 3: sidedness = FRONT_BACK; break;
  default: cerr<<"SIDE parse ERROR: option not in {1,3}"<<endl; sidedness=ERROR_LwTag;
  }
  tag = SIDE; 
} // LwChunkSide constructor

std::string LwSubChunkSide::setStr(bool useData) {
  if (useData && data) return (LwSubChunk::setStr(useData));
  if (useData && !data) cerr << "LwSubChunkSide WARN: useData true, buf is 0" << endl;
  string s;
  //s = setStrID4(options);
  U2 val = sidedness - U2(SIDE_0_NOT_VALID);
  assert (1==val || 3==val);
  s = setStrU2(val);
  s = setStrID4(tag) + setStrU2(s.length()) + s;
  return (s);
}

void 
LwSubChunkSide::print() {
  cout << "      LwSubChunk" << lwTag.getString(tag) << endl
       << "        sidedness: " << lwTag.getString(sidedness) << endl;
}


// SMAN 
LwSubChunkSman::LwSubChunkSman(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::sman\n"));
  assert(buf); assert(len>=2);
  handleMemory(buf,len,mmapped);  // set all the pointers
  max_smoothing_angle = getF4(start,length);
  tag = SMAN; 
} // LwChunkSman constructor

void 
LwSubChunkSman::print() {
  cout << "      LwSubChunk" << lwTag.getString(tag) << endl
       << "        max_smoothing_angle: " << max_smoothing_angle << endl;
}

// GLOS { glossiness[FP4], envelope[VX]  }
LwSubChunkGlos::LwSubChunkGlos(const char *buf, const size_t len, bool mmapped) {
  DebugPrintf(VERBOSE,("LwSubChunk::glos\n"));
  assert(buf); assert(len>=2);
  handleMemory(buf,len,mmapped);  // set all the pointers
  glossiness = getFP4(start,length);
  envelope = getVX(start+sizeof(FP4), length-sizeof(FP4));
  tag = GLOS; 
} // LwChunkGlos constructor

void 
LwSubChunkGlos::print() {
  cout << "      LwSubChunk" << lwTag.getString(tag) << endl
       << "        glossiness: " << glossiness << endl
       << "        envelope:   " << envelope << endl;
}

//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////


#ifdef REGRESSION_TEST
int main (int argc, char *argv[]) {
  bool ok=true;
  bool r; // results from an individual test

  // VertexSet
  {
    unsigned char ubuf[8] = {
      0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02
    };
    char *buf = (char *)ubuf;
    VertexSet vs(buf,8);
    string s1(buf,8);
    string s2=vs.setStr();
    if (0!=s1.compare(s2)) {ok=false;cerr<<"  VertexSet static test:  FAILED"<<endl;}
  }

  // Test POLS and VertexSets
  {
    unsigned char ubuf[20] = {
      0x50, 0x4f, 0x4c, 0x53, 0x00, 0x00, 0x00, 0x0c, 0x46, 0x41, 
      0x43, 0x45, 0x00, 0x03, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02
    };
    char *buf = (char *)ubuf;
    LwChunkPols p(buf,20);
    string s = p.setStr();
    LwChunkPols p2(s.c_str(),s.length());
    VertexSet v = p2.vertexSets[0];
    if (0 != v.vert[0]) {ok=false;cerr<<"  ERROR: failed POLS test - v0"<<endl;}
    if (1 != v.vert[1]) {ok=false;cerr<<"  ERROR: failed POLS test - v0"<<endl;}
    if (2 != v.vert[2]) {ok=false;cerr<<"  ERROR: failed POLS test - v0"<<endl;}
    if (!p.checkSetStr())  cout << "  POLS static check: FAILED!" << endl;
  }

  // Test writing
  {
    cout << "  Write test section" << endl;
    //string inFilename("../TestData/Triangle.lwo");
    string inFilename("../TestData/TriangleWithTexture.lwo");
    LwFile lwFile(inFilename,r);
    if (!r) {
      cerr << "ERROR: LwFile constructor failed loading file: " << inFilename << endl;
      ok=false;
    }

    // Use checkSetStr to verify that we will write out exactly the same thing
    if (!lwFile.checkSetStr()) cout << "  lwFile.checkSetStr: FAILED" << endl;

    if (false) {
      string outFilename("test-w1.lwo");
      std::ofstream o(outFilename.c_str(), std::ios::out);
      if (!o.is_open()) {
	std::cerr << "ERROR: output file open failed for " << outFilename << endl;
	return (false);
      }
      //
      // This is going to be really hard on memory management!
      //
      string s = lwFile.setStr();
      o << s;

      o.close(); // FIX: error check... which one is right??? fail or bad?
      if(o.fail()){ok=false;cerr<<"ERROR(fail): close was not a happy camper - "<<outFilename<<endl;}
      if(o.bad() ){ok=false;cerr<<"ERROR(bad ): close was not a happy camper - "<<outFilename<<endl;}
    }

    {
      string outFilename("test-w2.lwo");
      std::ofstream o(outFilename.c_str(), std::ios::out);
      if (!o.is_open()) {
	std::cerr << "ERROR: output file open failed for " << outFilename << endl;
	return (false);
      }

      bool useData=false;  string s = lwFile.setStr(useData);  o << s;

      o.close(); // FIX: error check... which one is right??? fail or bad?
      if(o.fail()){ok=false;cerr<<"ERROR(fail): close was not a happy camper - "<<outFilename<<endl;}
      if(o.bad() ){ok=false;cerr<<"ERROR(bad ): close was not a happy camper - "<<outFilename<<endl;}
    }
  }

  vector<string> filenames;
  filenames.push_back(string("California.lwo"));
#if 1
  filenames.push_back(string("Zareh/McKittrick_nm_crop2-Morrored-reduced2.lwo"));
  filenames.push_back(string("../TestData/Triangle.lwo"));
  filenames.push_back(string("../TestData/TriangleWithTexture.lwo"));
  filenames.push_back(string("../TestData/4x4GridOfTriangles.lwo"));
  filenames.push_back(string("../TestData/4x4GridOfTrianglesWith1CornerBentUp.lwo"));
  filenames.push_back(string("../TestData/SquareWith2TriangelesTextureMapped.lwo"));
  filenames.push_back(string("../TestData/SquareWith2Triangles.lwo"));
#endif
  size_t i;
  for (i=0;i<filenames.size();i++) {
    cout << endl << "Loading LWO: " << filenames[i] << endl << endl;
    r=false;
    LwFile lwFile(filenames[i],r);
    if (!r) {
      cerr << "ERROR: LwFile constructor failed loading file: " << filenames[i] << endl;
      ok=false;
    }

    // Use checkSetStr to verify that we would write out exactly the same thing
    if (!lwFile.checkSetStr()) cout << "  lwFile.checkSetStr: FAILED" << endl;

    //cout << endl << "   TREE" << endl << endl;
    cout << endl << endl;
    lwFile.printTree();
  }
  std::cout << argv[0] << (ok?": ok.":": FAILED!") << std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif // REGRESSION_TEST
