
#include "lw-io.H"
#include "lw-tags.H"

using namespace std;

//////////////////////////////////////////////////////////////////////
// GLOBAL VARIABLES
//////////////////////////////////////////////////////////////////////
static LwTag lwTag; // FIX: only want one copy of this thing floating around!

//////////////////////////////////////////////////////////////////////
// static helper functions
//////////////////////////////////////////////////////////////////////
#if 0
static void printBuf (const char *buf, const size_t len) {
  size_t i;
  for (i=0;i<len;i++)
    cout << i << ": " << (isprint(buf[i])?buf[i]:'*')
         << " Int: " << int((unsigned char)(buf[i])) << endl;
}
#endif

static LwTagEnum getHostEndian(void) {
  int short s=1;
  char *b = (char *)&s;
  if (b[0]) return (LW_LITTLE_ENDIAN);
  return (LW_BIG_ENDIAN);
}

// True if we need to byte swap on this host when reading lightwave
static bool LwNeedsSwap() { return ((LW_LITTLE_ENDIAN == getHostEndian()?true:false)); }


template <class T>
void fixEndian(T &i) {
  if (!LwNeedsSwap()) return;
  char *b = (char *)&i;
  char tmp[sizeof(T)];
  unsigned int x;
  for (x=0;x<sizeof(T);x++) tmp[x]=b[x];
  for (x=0;x<sizeof(T);x++) b[x]=tmp[(sizeof(T)-1)-x];
}


//////////////////////////////////////////////////////////////////////
// WRITE PRIMITIVE TYPES
//////////////////////////////////////////////////////////////////////
std::string setStrU4 (const U4 u4) {
  assert (4==sizeof(U4));
  U4 tmpU4 = u4;
  fixEndian (tmpU4);
  const char *tmpchar = (const char *)&tmpU4;
  string s(tmpchar,sizeof(U4));
  return (s);
}
std::string setStrU2 (const U2 u2) {
  assert (2==sizeof(U2));
  U2 tmpU2 = u2;
  fixEndian (tmpU2);
  const char *tmpchar = (const char *)&tmpU2;
  string s(tmpchar,sizeof(U2));
  return (s);
}
std::string setStrU1 (const U1 u1) {
  assert (1==sizeof(U1));
  const char *tmpchar = (const char *)&u1;
  string s(tmpchar,sizeof(U1));
  return (s);
}

std::string setStrF4 (const F4 f4) {
  assert (4==sizeof(F4));
  F4 tmpF4 = f4;
  fixEndian(f4);
  const char *tmpchar = (const char *)&tmpF4;
  const string s(tmpchar,sizeof(F4));
  return(s);
}
std::string setStrFP4 (const FP4 fp4) {
  assert (4==sizeof(FP4));
  assert (0.<=fp4);
  assert (1.>=fp4);
  const string s = setStrF4(fp4);
  return (s);
}
std::string setStrANG4 (const ANG4 a4) {
  assert (4==sizeof(ANG4));
  assert (-7.<=a4);
  assert (+7.>=a4);
  const string s = setStrF4(a4);
  return (s);
}

std::string setStrID4(const LwTagEnum t) {
  string s = lwTag.getString(t);
  if (s.length() != 4) {
    cerr << "setStrID4 ERROR:  all tags used in write ops should be 4 bytes!" << endl
	 << "  got: " << t << " - '" << s << "'" << endl;;
  }
  assert (s.length()==4);
  return (s);
}

// Make sure the string is null terminated and even aligned
std::string setStrS0 (const std::string _s) {
  string s = _s;
  if ('\0'!= s[s.length()-1])
    s += '\0';
  if (s.length()%2!=0) 
    s += '\0';
  return (s);
}


std::string setStrVX (const U4 u4) {
  if (u4<0xFF00) return (setStrU2(u4));
  return (setStrU4(u4));
}

//////////////////////////////////////////////////////////////////////
// READ PRIMITIVE TYPES
//////////////////////////////////////////////////////////////////////

// FIX: can we make this a template?  Probably not, since no type in args

U4 getU4 (const char *buf, const size_t len) {
  assert (buf);
  assert (len>=sizeof(U4));
  U4 *tmp=(U4 *)buf;
  U4 tmp2 = *tmp;
  fixEndian(tmp2);
  //cout << "getU4() = " << tmp2 << endl; // FIX: remove debugging code
  //printBuf(buf,4);
  return(tmp2);
}

U2 getU2 (const char *buf, const size_t len) {
  assert (buf);
  assert (len>=sizeof(U2));
  U2 *tmp=(U2 *)buf;
  U2 tmp2 = *tmp;
  fixEndian(tmp2);
  return(tmp2);
}

U1 getU1 (const char *buf, const size_t len) {
  assert (buf);
  assert (len>=sizeof(U1));
  U1 *tmp=(U1 *)buf;
  U1 tmp2 = *tmp;
  //fixEndian(tmp2); // Not needed
  return(tmp2);
}

#if 0
I4 getI4 (const char *buf, const size_t len) {
  assert (buf);
  assert (len>=sizeof(I4));
  I4 *tmp=(I4 *)buf;
  I4 tmp2 = *tmp;
  fixEndian(tmp2);
  return(tmp2);
}
#endif

#if 0
I2 getI2 (const char *buf, const size_t len) {
  assert (buf);
  assert (len>=sizeof(I2));
  I2 *tmp=(I2 *)buf;
  I2 tmp2 = *tmp;
  fixEndian(tmp2);
  return(tmp2);
}
#endif

#if 0
I1 getI1 (const char *buf, const size_t len) {
  assert (buf);
  assert (len>=sizeof(U1));
  I1 *tmp=(I1 *)buf;
  I1 tmp2 = *tmp;
  //fixEndian(tmp2);  // Not needed
  return(tmp2);
}
#endif

F4 getF4 (const char *buf, const size_t len) {
  assert(buf);
  assert(len>=4); 
  F4 *tmp=(F4 *)buf;
  F4 tmp2 = *tmp;
  fixEndian(tmp2);
  return(tmp2);
}

FP4 getFP4 (const char *buf, const size_t len) {
  assert(buf);
  assert(len>=4); 
  const F4 tmp = getF4(buf,len);
  assert (0.<=tmp);
  assert (1.>=tmp);
  return(tmp);
}


// FIX: what range are angles in?  Is it 0..2PI or -PI..PI?
ANG4 getANG4 (const char *buf, const size_t len) {
  assert(buf);
  assert(len>=4); 
  const F4 tmp = getF4(buf,len);
  assert (-100.<=tmp);
  assert (+100.>=tmp);
  return(tmp);
}

//////////////////////////////////////////////////////////////////////
// On to slightly more complicated tags
//////////////////////////////////////////////////////////////////////

LwTagEnum getID4 (const char *buf, const size_t len) {
  assert(buf);
  assert(len>=4);
  string s(buf,4);
  LwTagEnum e = lwTag.getEnum(s);
  return (e);
}

string getS0(const char *buf, const size_t len) {
  assert(buf);
  if (0==len) {
    cerr << "WARNING: getS0 len is ZERO!!" << endl;
    return (string("")); // FIX: danger!
  }
  size_t offset;
  //printBuf (buf,(30<len)?30:len);
  for (offset=0;offset < len && buf[offset]!='\0'; offset++);
  if ('\0'!=buf[offset]) {
    cerr << "getS0 ERROR: failed to find end of string in buffer" << endl;
    return (string ("getS0 ERROR"));
  }
  string s(buf,offset);
  //cout << "  getS0() size: " << s.size() << endl; // FIX: remove
  return (s);
}

// Figure out how many bytes this string takes up.
// Make sure we handle padding right.
size_t getS0Len(string s) {
  size_t l = s.size();
  if (s[s.size()-1] != '\0') l++;
  if (1==l%2) l++; // nul pads to make it fall on an even boundary;
  //cout << "FIX: getS0Len() algorithm may be off by 2" << endl;
  return (l);
}


U4 getVX(const char *buf, const size_t len) {
  assert(buf);
  assert(len>=2);
  U2 u2 = getU2(buf,len);
  if (u2<0xFF00) return u2;
  assert(len>=4);
  char tmpbuf[4];
  memcpy(tmpbuf,buf,4);
  tmpbuf[0]=0;  // Mask out 0xFF
  U4 u4 = getU4(tmpbuf,len);
  return (u4);
}
size_t getVXLen(const char *buf, const size_t len) {
  assert (sizeof(U2)==2);
  assert (sizeof(U4)==4); // make sure we can port to 64 bit arch ok.
  assert(buf);
  assert(len>=2);
  U2 u2 = getU2(buf,len);
  if (u2<0xFF00) return (sizeof(U2));
  return (sizeof(U4));
}

//////////////////////////////////////////////////////////////////////
// AGGREGATE TYPES (COLOR, VECTOR, etc)
// AKA COMPOSITE DATA TYPES
//////////////////////////////////////////////////////////////////////

VEC12::VEC12(const char *buf, const size_t len) {
  assert (buf);
  assert (12==sizeof(VEC12));
  assert (len>=sizeof(VEC12));
  X = getF4(buf,len);
  Y = getF4(buf+4,len-4);
  Z = getF4(buf+8,len-8);
}

// Buffer for writing to disk
std::string VEC12::setStr() {
  const string sx = setStrF4(X);
  const string sy = setStrF4(Y);
  const string sz = setStrF4(Z);
  const string tot = sx + sy + sz;
  assert (tot.length() == size());
  return (tot);
}

void
VEC12::print() {
  cout << "VEC12: " << X << " " << Y << " " << Z << endl;
}

//////////////////////////////
// Colors - 12 bytes: red[F4], green[F4], blue[F4] 
//////////////////////////////

// colors are usually [0..1], but can be outside of this range?!?!

COL12::COL12(const char *buf, const size_t len) {
  assert (buf);
  assert (12==sizeof(VEC12));
  assert (len>=sizeof(VEC12));
  red   = getF4(buf,len);
  green = getF4(buf+4,len-4);
  blue  = getF4(buf+8,len-8);
}

// Buffer for writing to disk
std::string COL12::setStr() {
  const string sr = setStrF4(red);
  const string sg = setStrF4(green);
  const string sb = setStrF4(blue);
  const string tot = sr + sg + sb;
  assert (tot.length() == size());
  return (tot);
}

void
COL12::print() {
  cout << "COL12: " << red << " " << green << " " << blue << endl;
}



//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST
int main (int argc, char *argv[]) {
  bool ok=true;

  //LwTag lwTag; // defined at the top

  cout << "  Host endian: " << lwTag.getString(getHostEndian()) << endl;
  if (LW_LITTLE_ENDIAN == getHostEndian()) {
    cout << "    Need to swap endian for this host." << endl;
  }

  // Be mindful of 64-bit vrs 32-bit machines
  if (4 != sizeof(U4) ) { ok=false; cerr << "  ERROR: U4 bad size" << endl; }
  if (2 != sizeof(U2) ) { ok=false; cerr << "  ERROR: U2 bad size" << endl; }
  if (1 != sizeof(U1) ) { ok=false; cerr << "  ERROR: U1 bad size" << endl; }

#if 0
  if (4 != sizeof(I4) ) { ok=false; cerr << "  ERROR: I4 bad size" << endl; }
  if (2 != sizeof(I2) ) { ok=false; cerr << "  ERROR: I2 bad size" << endl; }
  if (1 != sizeof(I1) ) { ok=false; cerr << "  ERROR: I1 bad size" << endl; }
#endif

  if (4 != sizeof(F4) ) { ok=false; cerr << "  ERROR: F4 bad size" << endl; }

  // Test U4
  {
    unsigned char buf[4]={0,0,1,238};
    if (494 != getU4((char *)buf,4)) {
      ok=false;  cerr << "ERROR: U4 failed to get endian right." << endl;
    }
  }

  // Test U2
  {
    unsigned char buf[2]={0,2};
    if (2 != getU2((char *)buf,2)) {
      ok=false;  cerr << "  ERROR: U2 failed to get endian right." << endl;
    }
  }

  // Test I1
#if 0
  {
    char buf[1]={-2};
    if (-2 != getI1(buf,1)) {
      ok=false;  cerr << "  ERROR: I1 failed to get endian right." << endl;
    }
  }
#endif

  if (XREF!=getID4("XREFblahblah",4)) {
      ok=false;  cerr << "  ERROR: getID4 failed to get XREF right." << endl;
  }

  // Test setStrU1
  {
    const U1 u1_4 = 4;
    const string s = setStrU1(u1_4);
    const U1 u1_4b = getU1(s.c_str(),s.length());
    if (u1_4 != u1_4b) {ok=false; cerr << "  ERROR: setStrU1 failed." << endl;}

    const U1 u1_201 = 201;
    const string sb = setStrU1(u1_201);
    const U1 u1_201b = getU1(sb.c_str(),sb.length());
    if (u1_201!=u1_201b) {ok=false; cerr << "  ERROR: setStrU1 failed."<<endl;}
  }
  // Test setStrU2
  {
    const U2 u2a = 4;
    const string s = setStrU2(u2a);
    const U2 u2aa = getU2(s.c_str(),s.length());
    if (u2a != u2aa) {ok=false; cerr << "  ERROR: setStrU2 failed." << endl;}

    const U2 u2b = 2002;
    const string sb = setStrU2(u2b);
    const U2 u2bb = getU2(sb.c_str(),sb.length());
    if (u2b!=u2bb) {ok=false; cerr << "  ERROR: setStrU2 failed."<<endl;}
  }
  // Test setStrU4
  {
    const U4 u4a = 4;
    const string s = setStrU4(u4a);
    const U4 u4aa = getU4(s.c_str(),s.length());
    if (u4a != u4aa) {ok=false; cerr << "  ERROR: setStrU4 failed." << endl;}

    const U4 u4b = 456789;
    const string sb = setStrU4(u4b);
    const U4 u4bb = getU4(sb.c_str(),sb.length());
    if (u4b!=u4bb) {ok=false; cerr << "  ERROR: setStrU4 failed."<<endl;}
  }

  // Test setStrF4
  {
    const F4 f4a = 4;
    const string s = setStrF4(f4a);
    const F4 f4aa = getF4(s.c_str(),s.length());
    if (f4a != f4aa) {ok=false; cerr << "  ERROR: setStrF4 failed." << endl;}

    const F4 f4b = 8765432.;
    const string sb = setStrF4(f4b);
    const F4 f4bb = getF4(sb.c_str(),sb.length());
    if (f4b!=f4bb) {ok=false; cerr << "  ERROR: setStrF4 failed."<<endl;}

    const F4 f4c = -0.12345;
    const string sc = setStrF4(f4c);
    const F4 f4cc = getF4(sc.c_str(),sc.length());
    if (f4c!=f4cc) {ok=false; cerr << "  ERROR: setStrF4 failed."<<endl;}
  }

  // Test setStrFP4
  {
    const FP4 f4a = 0.123;
    const string s = setStrFP4(f4a);
    const FP4 f4aa = getFP4(s.c_str(),s.length());
    if (f4a != f4aa) {ok=false; cerr << "  ERROR: setStrF4 failed." << endl;}
  }

  // Test setStrANG4
  {
    const ANG4 f4a = 0.123;
    const string s = setStrANG4(f4a);
    const ANG4 f4aa = getANG4(s.c_str(),s.length());
    if (f4a != f4aa) {ok=false; cerr << "  ERROR: setStrF4 FAILED!" << endl;}
  }

  // Test setStrID4
  {
    const LwTagEnum t=TAGS;
    string s = setStrID4(t);
    if (s.length()!=4) {ok=false; cerr <<" ERROR: setStrID4 was not 4 bytes"<<endl;}
    const LwTagEnum tb=getID4(s.c_str(),4);
    if (t!=tb) {ok=false; cerr <<" ERROR: setStrID4 FAILED!"<<endl;}
  }

  // Test VX
  {
    U4 vx;
    const char vx1_4c[4] = {0xFF, 0x00, 0x00, 0x01};
    const char vx1_2c[4] = {0x00, 0x01};

    if (2 != getVXLen(vx1_2c,4)) {ok=false; cerr <<" VX ERROR: 0x0001 len FAILED! " << endl;}
    vx = getVX (vx1_2c,4);
    if (1 != vx)  {ok=false; cerr <<" VX ERROR: 0x0001 FAILED!  "<< vx << endl;}

    if (4 != getVXLen(vx1_4c,4)) {ok=false; cerr <<" VX ERROR: 0xFF000001 len FAILED! " << endl;}
    vx = getVX (vx1_4c,4);
    if (1 != vx)  {ok=false; cerr <<" VX ERROR: 0xFF000001 FAILED!  "<< vx << endl;}
  }

  // Test COL12
  {
    const F4 r=0.1, g=0.2, b=0.3;
    COL12 c1(r,g,b);
    string s1 = c1.setStr();
    if (c1.size() != s1.length())  {ok=false; cerr <<" COL12 ERROR: setStr FAILED!  "<< endl;}
    COL12 c2(s1.c_str(), s1.length());
    F4 r2,g2,b2;
    c2.getRGB(r2,g2,b2);
    if (r != r2)  {ok=false; cerr <<" COL12 ERROR: r2 FAILED!  "<< r << " " << r2 << endl;}
    if (g != g2)  {ok=false; cerr <<" COL12 ERROR: g2 FAILED!  "<< g << " " << g2 << endl;}
    if (b != b2)  {ok=false; cerr <<" COL12 ERROR: b2 FAILED!  "<< b << " " << b2 << endl;}
  }

  // Test VEC12
  {
    const F4 x=0.1, y=-0.2, z=1234.5678;
    VEC12 v1(x,y,z);
    string s1 = v1.setStr();
    if (v1.size() != s1.length())  {ok=false; cerr <<" VEC12 ERROR: setStr FAILED!  "<< endl;}
    VEC12 c2(s1.c_str(), s1.length());
    F4 x2,y2,z2;
    c2.getXYZ(x2,y2,z2);
    if (x != x2)  {ok=false; cerr <<" VEC12 ERROR: x2 FAILED!  "<< x << " " << x2 << endl;}
    if (y != y2)  {ok=false; cerr <<" VEC12 ERROR: g2 FAILED!  "<< y << " " << y2 << endl;}
    if (z != z2)  {ok=false; cerr <<" VEC12 ERROR: b2 FAILED!  "<< z << " " << z2 << endl;}
  }


  {
    // PNTS VEC12's from Triangle.lwo
    //cout << "  Length = " << 0x24 << endl;
    unsigned char ubuf[0x24] = {
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // a
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, // b
      0x3f, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // c
    };
    char *buf = (char *)ubuf;
    VEC12 a(buf,12);
    VEC12 b(buf+12,12);
    VEC12 c(buf+12+12,12);

    if (2.  != b.Z) {ok=false; cerr << "  ERROR: b.Z != 2." << endl;}
    if (1.5 != c.X) {ok=false; cerr << "  ERROR: c.X != 1.5" << endl;}

    string sb = b.setStr();
    if (12!=sb.length()) {ok=false; cerr << "  ERROR: VEC12 setStr not 12 bytes"<<endl;}
    VEC12 bb(sb.c_str(), sb.length());
    

    string sc = c.setStr();
    if (12!=sc.length()) {ok=false; cerr << "  ERROR: VEC12 setStr not 12 bytes"<<endl;}
    VEC12 cc(sc.c_str(), sc.length());

    if (2.  != bb.Z) {ok=false; cerr << "  ERROR: b.Z != 2." << endl;}
    if (1.5 != cc.X) {ok=false; cerr << "  ERROR: c.X != 1.5" << endl;}
  }


  std::cout << argv[0] << (ok?": ok.":": FAILED!")<<std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif // REGRESSION_TEST
