
#include <sys/types.h>
#include <sys/stat.h>

#include <vector>
#include <string>
#include <iostream>

#include "lw-tags.H"

#include <stdint.h> // Linux uses this for uint32_t
typedef uint32_t U4;
typedef uint16_t U2;
typedef uint8_t  U1;

typedef int32_t I4;
typedef int16_t I2;
typedef int8_t	I1;

typedef float F4;
typedef float FP4; // F4 but only in range 0..1
typedef float ANG4; // F4 - angle in radians
typedef std::string S0;

// len is the size of the input buffer, not necessarily
// the size of the data item that you are trying to get

U4 getU4 (const char *buf, const size_t len);
U2 getU2 (const char *buf, const size_t len);
U1 getU1 (const char *buf, const size_t len);

std::string setStrU4 (const U4 u4);
std::string setStrU2 (const U2 u2);
std::string setStrU1 (const U1 u1);

// Does not seem to be used in the spec.
#if 0
I4 getI4 (const char *buf, const size_t len);
I2 getI2 (const char *buf, const size_t len);
I1 getI1 (const char *buf, const size_t len);
std::string setStrI4 (I4 i4);
std::string setStrI2 (I2 i2);
std::string setStrI1 (I1 i1);
#endif

F4 getF4 (const char *buf, const size_t len);
FP4 getFP4 (const char *buf, const size_t len);
ANG4 getANG4 (const char *buf, const size_t len); // radians..

std::string setStrF4 (const F4 f4);
std::string setStrFP4 (const FP4 fp4);
std::string setStrANG4 (const ANG4 a4);

//std::string getID4 (const char *buf, const size_t len);
LwTagEnum getID4 (const char *buf, const size_t len);
std::string setStrID4(const LwTagEnum t);

std::string getS0 (const char *buf, const size_t len);
size_t getS0Len(const std::string s);
std::string setStrS0(const std::string s);


//////////////////////////////////////////////////////////////////////
// Composite data types
//////////////////////////////////////////////////////////////////////

U4 getVX(const char *buf, const size_t len);
size_t getVXLen(const char *buf, const size_t len);  // FIX: return 4?

std::string setStrVX(const U4 vx);  // For writing to disk

class COL12 {
public:
  COL12 (F4 r, F4 g, F4 b) {red=r;green=g;blue=b;};
  COL12 () {red=green=blue=0.;};
  COL12 (const char *buf, const size_t len);
  std::string setStr();  // Buffer for writing eto disk
  void getRGB(F4 &r, F4 &g, F4 &b) {r=red;g=green;b=blue;};
  size_t size() {return (12);};
  void print();
private:
  F4 red, green, blue;
};

class VEC12 {
public:
  VEC12 (F4 x, F4 y, F4 z) {X=x;Y=y;Z=z;};
  VEC12 () {X=Y=Z=0.;};
  VEC12 (const char *buf, const size_t len);
  std::string setStr();  // Buffer for writing to disk
  void getXYZ (F4 &x, F4 &y, F4 &z) {x=X;y=Y;z=Z;};
  void print();
  size_t size() {return (12);};
  F4 X,Y,Z;
};


//////////////////////////////////////////////////////////////////////
// Types generated in reading a file
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
// CHUNK
//////////////////////////////////////////////////

class LwSubChunk {
  //protected:
public:
  LwTagEnum tag; // ID4
  bool dataMMap; // true if data is mmapped, otherwise delete in destructor
  char *data;	// Pointer to beginning of subchunk (start of the tag)
  size_t dataSize; // size of data - including tag and length

  char *start; // Pointer to beginning of the chunk data (data+6_
  U2 length;   // size - 6
  bool pad;

public:
  LwSubChunk();
  virtual ~LwSubChunk();

  // Return vector of unsigned char bufs...
  //virtual std::vector<std::string *> getBufs();
  virtual std::string setStr(bool useData=true);


  void handleMemory();  // set it all to empty
  void handleMemory(const char *buf, const size_t len, bool mmapped);

  virtual void print();
  virtual void printTree(const std::string baseStr);
  virtual void printTreeFull();
  size_t getDataSize(){return(dataSize);};

  // Factories - returns ptr to some subclass of LwSubChunk

  // Surface Sub-Chunks (There are a ton of these)
  static LwSubChunk *surfaceFactory(const char *buf, const size_t len, bool mmapped);
  static LwSubChunk *   clipFactory(const char *buf, const size_t len, bool mmapped);
  // Everything else
  static LwSubChunk *   factory(const char *buf, const size_t len, bool mmapped);
};

// COLR { base-color[COL12], envelope[VX]  }
class LwSubChunkColr : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkColr(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true);  // Buffer for writing eto disk
  void print();
private:
  COL12 base_color;
  U4 envelope;
};


// Handle all these: DIFF, LUMI, SPEC, REFL, TRAN, TRNL
// WARNING: default diffuse value is 1.0, rest default to 0
class LwSubChunkBaseShadeVals : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkBaseShadeVals(const LwTagEnum t,
			  const char *buf, const size_t len, const bool mmapped=false
			  );
  std::string setStr(bool useData=true);  // Buffer for writing eto disk
  void print();
private:
  F4 intensity; // FP4 runs [0..1]
  U4 envelope; // VX
};

// BLOK { header[SUB-CHUNK],  attributes[SUB-CHUNK] * }
// block-header { ordinal[S0], block-attributes[SUB-CHUNK] * }
// SUB-CHUNK ::= tag[ID4], length[U2],  data[...], pad[U1] ?
class LwSubChunkBlok : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkBlok(const char *buf, const size_t len, bool mmapped=false);
  ~LwSubChunkBlok();
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
private:
  std::vector<LwSubChunk *> attributes;
};


// BLOK Header - IMAP, PROC, GRAD, SHDR
class LwSubChunkBlokHeader : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkBlokHeader(const LwTagEnum t,
		 const char *buf, const size_t len, const bool mmapped=false);
  ~LwSubChunkBlokHeader();
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
private:
  std::string ordinal;
  std::vector<LwSubChunk *> block_attributes;
};

class LwSubChunkChan : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkChan(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  bool isValidChannel(const LwTagEnum t);
private:
  LwTagEnum texture_channel; // COLR, DIFF, LUMI, SPEC, GLOS,REFL, TRAN, TRNL, or BUMP
};

// OPAC { type[U2], opacity[FP4],  envelope[VX] }
class LwSubChunkOpac : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkOpac(const char *buf, const size_t len, bool mmapped=false);
  //~LwSubChunkOpac();
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  U2 type;
  F4 opacity; // FP4 [0..1]
  U4 envelope; // VX
};

// ENAB { enable[U2] } - if missing, blok defaults to ENABLED
class LwSubChunkEnab : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkEnab(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  bool enable; // U2
};


// NEGA
class LwSubChunkNega : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkNega(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  bool enable;
};


// TMAP
class LwSubChunkTmap : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkTmap(const char *buf, const size_t len, bool mmapped=false);
  ~LwSubChunkTmap();
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
private:
  std::vector<LwSubChunk*> attributes;
};

// TMAP POS: Position, Orientation, Size
// CNTR, SIZE, ROTA { vector[VEC12], envelope[VX] }
class LwSubChunkTmapPos : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkTmapPos(const LwTagEnum t,
		    const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  VEC12 vector;
  U4 envelope;
};

// FALL
class LwSubChunkFall : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkFall(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  LwTagEnum type;
  VEC12 vector;
  U4 envelope; // VX
};

// OREF
class LwSubChunkOref : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkOref(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  std::string object_name;
};

// CSYS { type[U2] }
class LwSubChunkCsys : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkCsys(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  LwTagEnum type;  // OBJ_COORD or WORLD_COORD
};


// PROJ { projection-mode[U2] }
class LwSubChunkProj : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkProj(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  LwTagEnum projection_mode;  // [0..5]
};

// AXIS
class LwSubChunkAxis : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkAxis(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  LwTagEnum axis;  // [0..2]  U2
};

// IMAG  texture-image[VX]
class LwSubChunkImag : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkImag(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  U4 texture_image;
};


// WRAP  { width-wrap[U2], height-wrap[U2]  }
class LwSubChunkWrap : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkWrap(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  static LwTagEnum getWrapType(U2 val);
private:
  LwTagEnum width_wrap, height_wrap;
};

// WRAP AMOUNT 
class LwSubChunkWrapAmount : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkWrapAmount(const LwTagEnum t,
		       const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  F4 cycles; // FP4
  U4 envelope; // VX
};

// Anti-Alias strength
// AAST  
class LwSubChunkAast : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkAast(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  U2 flags;
  F4 strength; // FP4
};

// Pixel Blending
// PIXB flags[U2]
class LwSubChunkPixb : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkPixb(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  U2 flags;
};

// WARNING:  There are two different VMAPS!  Chunk and SubChunk... this is SUB-CHUNK
// VMAP flags[U2]
class LwSubChunkVmap : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkVmap(const char *buf, const size_t len, bool mmapped=false);
  void print();
private:
  std::string txuv_map_name; // S0
};

// Reflection options / Transparency options
// RFOP flags[U2]
class LwSubChunkRfopTrop : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkRfopTrop(const LwTagEnum t,
		     const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  LwTagEnum options;
};

// SIDE flags[U2]
class LwSubChunkSide : public LwSubChunk {
public:
  LwSubChunkSide(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
private:
  LwTagEnum sidedness; // U2
};

// SMAN { max-smoothing-angle [ANG4] }
class LwSubChunkSman : public LwSubChunk {
public:
  LwSubChunkSman(const char *buf, const size_t len, bool mmapped=false);
  void print();
private:
  F4 max_smoothing_angle; // ANG4 - randians
};

// GLOS { glossiness[FP4], envelope[VX]  }
class LwSubChunkGlos : public LwSubChunk {
public:
  LwSubChunkGlos(const char *buf, const size_t len, bool mmapped=false);
  void print();
private:
  FP4 glossiness; // F4 in [0..1]
  U4 envelope; // VX
};

////////////////////
// CLIP SUB-CHUNKS
////////////////////

// 
class LwSubChunkStil : public LwSubChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwSubChunkStil(const char *buf, const size_t len, bool mmapped=false);
  ~LwSubChunkStil();
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
//private:
  std::string name; // FNAM0, which is an S0
};


//////////////////////////////////////////////////
// CHUNK
//////////////////////////////////////////////////

class LwChunk {
public:
  LwChunk();
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunk(const char *buf, const size_t len, bool mmapped);
  virtual ~LwChunk();

  // FIX: need something that handles memory better
  virtual std::string setStr(bool useData=true); // emit buf to write to disk

#ifdef REGRESSION_TEST
  bool checkSetStr(); // compare data with setStr to make sure it all works
#endif
  // Sub classes can call this to setup the LwChunk memory section
  void handleMemory();  // set it all to empty
  void handleMemory(const char *buf, const size_t len, bool mmapped);

  size_t getDataSize(){return(dataSize);}; // includes tag and length field
  virtual void print();
  virtual void printTree(const std::string baseStr);
  virtual void printTreeFull();

  // Get the first one we hit
  virtual LwSubChunk *getSubChunkPtr(const LwTagEnum t);

  LwTagEnum tag; // ID4
  //std::vector<LwSubChunk *> subChunks;  // Chunks must define this themselves

  //
  // memory
  //
  bool dataMMap; // true if data is mmapped, otherwise delete in destructor
  char *data;  // Pointer to beginning of the chunk
  size_t dataSize; // size of start - including tag and length

  char *start; // Pointer to data inside chunk
  size_t length; // U4 -- length of just the tag
  //U4 length;

  bool pad;  // is there A U1 pad nul byte at the end to make it even aligned?
};

//////////////////////////////////////////////////////////////////////
// CHUNCK SUB-CLASSES - one for each tag
//////////////////////////////////////////////////////////////////////

// FIX: make tags handle more than one string.
// TAGS { tag-string[S0] * }
class LwChunkTags : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkTags(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
  std::vector<std::string> tag_strings;
};

// LAYR { number[U2], flags[U2], pivot[VEC12], name[S0], parent[U2] ?  }
// "parent is -1 or missing to indicate no parent."
// Arr... but it is UNsigned.  So you mean 0xFFFF??
class LwChunkLayr : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkLayr(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();

  U2 number;
  U2 flags;
  VEC12 pivot;
  std::string name;
  bool parent_given;  // true if parent listed
  U2 parent;  // Optional
};


class LwChunkPnts : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkPnts(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
  std::vector <VEC12> point_location;  // No destructor needed
};

// TXUV is a UV Texture map of dimension 2
class LwChunkVmap : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkVmap(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
  LwTagEnum type; // PICK, WGHT, TXUV, RGBA, etc.
  size_t dimension; // U2... length of values
  std::string name; // S0
  std::vector <U4> verts;  // VX // as long as dimension
  std::vector <F4> values;  // as long as dimension
};

// DISCONTINUOUS VERTEX MAPPING
// TXUV is a UV Texture map of dimension 2
// VMAD { type[ID4], dimension[U2],  name[S0],
// ( vert[VX], poly[VX], value[F4]  # dimension )* }
class LwChunkVmad : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkVmad(const char *buf, const size_t len, bool mmapped=false);
  void print();
  void printTree(const std::string baseStr);
  LwTagEnum type; // PICK, WGHT, TXUV, RGBA, etc.
  size_t dimension; // U2... length of values
  std::string name; // S0
  std::vector <U4> verts;  // VX // as long as dimension
  std::vector <U4> polys;  // VX // as long as dimension
  std::vector <F4> values;  // as long as dimension
};

class LwChunkBbox : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkBbox(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  VEC12 min,max;
};

// Inside of LwChunkPols - POLS
// ( numvert+flags[U2],  vert[VX] # numvert )
class VertexSet {
public:
  VertexSet(const char *buf, const size_t len);
  void print();
  size_t getDataSize() {return(dataSize);};
  std::string setStr(); // emit buf to write to disk

  size_t dataSize; // total bytes that make up this object in file.
  size_t numvert;
  U2 flags;
  std::vector<U4> vert;
};

class LwChunkPols : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkPols(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
  LwTagEnum type; // ID4
  std::vector<VertexSet> vertexSets; // Is there ever more than one?
};

// Inside of PTAG
// ( poly[VX], tag[U2] )
class PolySet {
public:
  PolySet(const char *buf, const size_t len);
  std::string setStr(); // emit buf to write to disk
  void print();
  size_t getDataSize() {return(dataSize);};

  size_t dataSize; // total bytes that make up this object in file.
  U4 poly;
  U2 tag;
};

class LwChunkPtag : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkPtag(const char *buf, const size_t len, bool mmapped=false);
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
  LwTagEnum type; // ID4
  std::vector<PolySet> polySets; // Is there ever more than one?
};

// SURF { name[S0], source[S0], attributes[SUB-CHUNK] * }
class LwChunkSurf : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkSurf(const char *buf, const size_t len, bool mmapped=false);
  ~LwChunkSurf();
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
  void printTreeFull();
  LwSubChunk *getSubChunkPtr(const LwTagEnum t);
  // LwTagEnum type; // ID4
  std::string name;
  std::string source;
  std::vector<LwSubChunk *> subChunks;
};

class LwChunkClip : public LwChunk {
public:
  // buf points to the beginning of the tag.
  // len is num of bytes including the tag and size field
  //   just the tag, not the entire file.
  LwChunkClip(const char *buf, const size_t len, bool mmapped=false);
  ~LwChunkClip();
  std::string setStr(bool useData=true); // emit buf to write to disk
  void print();
  void printTree(const std::string baseStr);
  LwSubChunk *getSubChunkPtr(const LwTagEnum t);
  //LwTagEnum type; // ID4
  U4 index;
  std::vector<LwSubChunk *> subChunks; 
};




//////////////////////////////////////////////////////////////////////
// LWFILE
//////////////////////////////////////////////////////////////////////

// FIX: move getChunk to be a static member of LwChunk

class LwFile {
public:
  LwFile(); // do not load from a file, build it yourself
  LwFile(const std::string _filename, bool &ok); // load from disk, mmap
  ~LwFile();

  std::string setStr(bool useData=true); // emit the buffer to write to disk
#ifdef REGRESSION_TEST
  bool checkSetStr(); // compare data with setStr to make sure it all works
#endif

  void print(); // cout a structural represenation of the tags
  void printTree(); // cout a structural represenation of the tags
  void printTreeFull(); // cout a structural represenation of the tags

  // FIX: really need to move this into LwChunk as a static:
  LwChunk *getChunk(const char *buf, const size_t len);
  
  LwChunk *getChunkPtr(const LwTagEnum t);

//private:
  std::string filename;
  bool isMMapped;	// true if data is mmap'ed from disk
  char *start; // Pointer to beginning of the chunks
  U4 length;  // length of start.  File size (- 8 for the form, and size) // FIX: size_t

  // file info - directly about the file
  size_t dataSize; // size of data - including tag and length (mmapped size)
  char *data;  // mmapped data

  LwTagEnum form; // ID4 Should be just "FORM";
  std::vector<LwChunk *> chunks;
};

