/// Strip off the pdf header from a PDF/VIC file to produce a vicar file

/*

  There are several ways to do this.  The most brain dead is to scan
  for LBLSIZE= and start the new file with that.  I think the
  technically correct way is to read the size from the IMG pds header
  and then jump to the LBLSIZE, but that sounds like more work.

  RECORD_TYPE                       = FIXED_LENGTH
  RECORD_BYTES                      = 128
  FILE_RECORDS                      = 330
  LABEL_RECORDS                     = 169

  // POINTERS TO DATA OBJECTS 

  ^IMAGE_HEADER                     = 170
  ^IMAGE                            = 267

*/

#include <unistd.h>	// close()

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 	// open
#include <sys/mman.h>	// mmap

#include <cassert>
#include <iostream>

using namespace std;


size_t badValue() {return(std::numeric_limits<size_t>::max());}

void PrintHelp (const string name) {
    cout << endl
	 << "Description: strip the PDS text header off of a PDS/VICAR file to" << endl
	 << "             give you straight vicar file." << endl << endl;
    cout << "USAGE:" << endl
	 << "  " << name << " <in PDS/IMG file> <out vicar file>" << endl;
}

bool MmapFile (const string filename, char *&data,size_t &dataSize) {
  struct stat sb;
  int r = stat (filename.c_str(), &sb);
  if (0 != r) { perror("ERROR: stat failed"); return (false); }
  dataSize = sb.st_size;


  int fd = open (filename.c_str(), O_RDONLY, 0);
  if (-1==fd) { perror ("ERROR: open failed"); return (false); }

  int flags =
#ifdef linux
    MAP_PRIVATE
#elif __APPLE__
    MAP_FILE
#else
#error What flags to give to mmap on this os?
#endif
    ;

  data = (char *)mmap (0, dataSize, PROT_READ,  flags, fd, 0);
  //data = (char *)mmap (0, dataSize, PROT_READ,  MAP_FILE, fd, 0);
  if ((char *)(-1)==data) { close(fd); perror ("ERROR: mmap failed"); return (false); }
  close(fd); // Close does not munmap, so close as soon as possible

  return (true);
}

/// \brief Scan for the LBLSIZE= string that is at the beginning of the vicar file
/// \param data pointer to beginning of file buffer
/// \param dataSize how long is this file buffer
/// \return the offset into the file for the beginning of LBLSIZE= or badValue() if not found

size_t GetLabelStart(const char *data, const size_t dataSize) {
  assert(data);
  const string lbl("LBLSIZE=");
  for (const char *cur=data; cur < data+dataSize-lbl.size(); cur++)
    if ('L'==*cur) {
      string testStr(cur,lbl.size());
      if (0==lbl.compare(testStr)) return (cur-data);
    }
  return (badValue());
}

//////////////////////////////////////////////////////////////////////
/// \brief HERE WE ARE
int main (int argc, char *argv[]) {
  if (3!=argc) {
    PrintHelp(string(argv[0]));
    return (EXIT_FAILURE);
  }

  bool ok=true;

  char *data;
  size_t dataSize;
  if (!MmapFile(string(argv[1]),data,dataSize)) return (EXIT_FAILURE);
  
  const size_t start = GetLabelStart(data,dataSize);
  if (badValue() == start) {
    cerr << "Failed to find the Vicar Label" << endl;
    if (-1== munmap(data,dataSize)) {perror ("ERROR: munmap failed");}
    return (EXIT_FAILURE);
  }

  FILE *outF = fopen(argv[2],"wb");
  if (!outF) {perror("failed to open output file"); return(EXIT_FAILURE);}


  const size_t len = dataSize-start;
  if (len != fwrite (data+start,sizeof(char),len,outF)) {
    perror ("not able to write the whole vicar file");
    ok=false;
  }

  if (outF) if (0!=fclose(outF)) { perror("Failed to close output file"); ok=false; }

  if (data) if (-1== munmap(data,dataSize)) {perror ("ERROR: munmap failed"); ok=false;}

  return (ok?EXIT_SUCCESS:EXIT_FAILURE);

}
