
#include <string>
#include <vector>
#include <iostream>
#include <fstream>

#include <map>

using namespace std;

bool isStart(const string s, string &name) {
  size_t start = s.find("START");
  if (start!=0) return (false);
  start = s.rfind(" ");
  if (0==start || string::npos==start) {
    cerr << "isStart parse error" << endl;
  }
  name = s.substr(start+1,s.length());
  //cout << "BEGIN: " << name << endl;
  return (true);
}

bool isEnd(const string s, string &name) {
  size_t start = s.find("END");
  if (start!=0) return (false);
  start = s.rfind(" ");
  if (0==start || string::npos==start) {
    cerr << "isEnd parse error" << endl;
  }
  name = s.substr(start+1,s.length());
  //cout << "END:   " << name << endl;
  return (true);
}

//////////////////////////////////////////////////////////////////////

int
main (int argc, char *argv[]) {
  if (4 != argc) {
    cerr << "ERROR: incorrect num of args:" << endl
	 << "  " <<  argv[0] << " tagfile basename" << endl;
    return (EXIT_FAILURE);
  }
  const string filename(argv[1]);
  const string outFilename(argv[2]);
  const string basename(argv[3]);

  //
  // Load the file into a vector of strings.
  //
  vector<string> lines;
  {
    std::ifstream in(filename.c_str(), ios::in);
    if (!in.is_open()) {
      cerr << "Failed to open input file" << endl;
      exit (EXIT_FAILURE);
    }

    while (in.good() && !in.eof()) {
      char buf[256];
      in.getline(buf,256);
      string s(buf);
      if (0==s.length()) continue; // Skip blank lines
      int start = s.find_first_not_of(" \t");
      int comment = s.find("#");
      if (start == comment) { 
	cout << "skipping comment: " << s << endl;
	continue; 
      }
      // FIX: skip trailing comments too.
      lines.push_back(s.substr(start,s.length()));
    }
  }

  lines.push_back(string("ERROR_"+basename)); // give ourselves an error value;


  vector<string> finalNames; // replace "START " with "START_" and "END " with "END_"
  {
    size_t i;
    for (i=0;i<lines.size();i++) {
      string newName=lines[i], name;
      if (isStart(lines[i],name)) newName = string("START_")+name;
      if (isEnd  (lines[i],name)) newName = string("END_")  +name;
      finalNames.push_back(newName);
    }
  }

  vector<string> groups; // check for all the starts
  vector<size_t> groupStart, groupEnd;  // indices for where groups begin and end
  {
    bool inGroup=false;
    string curGroup;
    size_t i;
    for (i=0;i<lines.size();i++) {
      string name;
      if (isStart(lines[i],name)) {
	if (inGroup) {cerr << "ERROR: Found start in a group in Line "<<i<<endl;exit(EXIT_FAILURE);}
	inGroup=true;
	curGroup = name;
	groups.push_back(name);
	groupStart.push_back(i);
      }
      if (isEnd(lines[i],name)) {
	if (!inGroup) {
	  cerr << "ERROR: Found END not in group. Line "<<i<<endl
	       << "  Line value: " << lines[i] << endl;
	  exit(EXIT_FAILURE);
	}
	if (0!=name.compare(curGroup)) {
	  cerr << "ERROR: group name missmatch on line " <<i<<endl;exit(EXIT_FAILURE);
	}
	curGroup=string("-*- none -*-");
	inGroup=false;
	groupEnd.push_back(i);
      }
    }
    if (inGroup) {cerr<<"ERROR: no closing END at end of file"<<endl; exit(EXIT_FAILURE);}
  }

  cout << argv[0] << ":" << endl
       << "  Number of lines  read: " << lines.size() << endl
       << "  Number of groups read: " << groups.size() << endl;


  string outFileC = outFilename + string(".C");
  std::ofstream oC(outFileC.c_str(),std::ios::out);
  string outFileH = outFilename + string(".H");
  std::ofstream oH(outFileH.c_str(),std::ios::out);

  // FIX: check is_open() on both

  // Include guard
  oH << "#ifndef _LW_TAGS_H_" << endl
     << "#define _LW_TAGS_H_" << endl << endl;


  oH << "#include <map> " << endl
     << "#include <string> " << endl
     << "#include <iostream> " << endl
     << endl;
  oC << endl
     << "#include \""<<outFileH<<"\"" << endl 
     << "using namespace std;" << endl << endl;

  //
  // Write out the enum
  //
  oH << "enum " << basename << "Enum {" << endl;
  {
    size_t i;
    for (i=0;i<lines.size();i++) {
      string gname;
      if (i==lines.size()-1)
	oH << endl << " //// EXTRA TAG to use in case of an error" << endl;
      if (isStart(lines[i],gname)) 
	oH << endl << " //// GROUP: " << gname << endl;
      if (i<lines.size()-1 && !isStart(lines[i],gname) && !isEnd(lines[i],gname)) oH << "  ";
      oH << " " << finalNames[i] << "="<< i <<"," << endl;
    }
  }
  oH << "}; // End of enum " << basename << endl << endl;


  const string stringMapType(basename + string("StringMap"));
  const string enumMapType  (basename + string("EnumMap"));
  const string stringPairType(basename + string("StringPair"));
  const string enumPairType  (basename + string("EnumPair"));

  oH << "typedef std::map <std::string,"<<basename<<"Enum> "<<stringMapType<<";" << endl
     << "typedef std::map <"<<basename<<"Enum,std::string> "<<enumMapType<<";" << endl
     << "typedef std::pair<std::string,"<<basename<<"Enum> "<<stringPairType<<";" << endl
     << "typedef std::pair<"<<basename<<"Enum,std::string> "<<enumPairType<<";" << endl
     << endl;
 
  //
  //  Build the class
  // 
  oH << "class "<<basename<<" {"<<endl
     << "public:" << endl
     << "  "<<basename<<"();  // constructor"<<endl;

  // Lookups:
  oH << "  "<<basename<<"Enum getEnum(const std::string s);" << endl;
  oH << "  "<<basename<<"Enum getEnum(const char *s, const size_t len);" << endl;
  oH << "  std::string getString(const "<<basename<<"Enum);" << endl;

  // Memberships
  {
    size_t i;
    for (i=0;i<groups.size();i++) {
      oH << "  bool isGroup"<<groups[i]<<"(const "<<basename<<"Enum e);" << endl;
    }
  }

  //// PRIVATE

  oH << "private:" << endl
     << "  "<<basename<<"StringMap stringMap;"<<endl
     << "  "<<basename<<"EnumMap enumMap;"<<endl;
  oH << "}; // End class " << basename << endl
     << endl;



  //
  //  Initialize the maps
  // 
  // CONSTRUCTOR
  oC << basename <<"::" << basename << "() {"<<endl;
  {
    size_t i;
    for (i=0;i<finalNames.size();i++) {
      const string b = basename;
      const string n = finalNames[i];
      oC << "  stringMap.insert("
	 << b <<"StringPair(\"" << n << "\", "<< n << " ));" << endl;
      //oC << "  enumMap.insert  ("
      // << b <<"EnumPair  ( " << n << " ,\""<< n << "\"));" << endl;
    }
  }
  // Build the reverse lookup table.
  oC << "  {" << endl
     << "    "<<stringMapType<<"::iterator iter;" << endl
     << "    for (iter=stringMap.begin(); iter!= stringMap.end(); iter++) {" << endl
     << "      enumMap.insert("<< enumPairType << "(iter->second,iter->first));" << endl
     << "    }" << endl
     << "  }" << endl
    ;
  oC << "} // end " << basename << " constructor"<<endl<<endl;

  //
  // Lookups
  //
  oC << basename<<"Enum "<<basename<<"::getEnum(const std::string s) {" << endl
     << "  " << stringMapType << "::iterator iter;" << endl
     << "  iter=stringMap.find(s);" << endl
     << "  if (iter!=stringMap.end()) return(iter->second);" << endl
     << "  cout << \"Did not find tag string in stringMap\"<<endl;" << endl
     << "  return (ERROR_"<<basename<<");" << endl
     << "} // getEnum(string) "<< endl << endl;

  oC << basename<<"Enum "<<basename<<"::getEnum(const char *s, const size_t len) {" << endl
     << "  assert(s);" << endl
     << "  assert(len>=4);" << endl
     << "  std::string str(s,4);" << endl
     << "  return(getEnum(str));" << endl
     << "} // getEnum (char *..." << endl
     << endl;

  oC << "std::string "<<basename<<"::getString(const "<<basename<<"Enum e) {" << endl
     << "  " << enumMapType << "::iterator iter;" << endl
     << "  iter=enumMap.find(e);" << endl
     << "  if (iter!=enumMap.end()) return(iter->second);" << endl
     << "  cout << \"Did not find tag enum in enumMap\"<<endl;" << endl
     << "  return (string(\"ERROR_"<<basename<<"\"));" << endl
     << "} // getString " << endl << endl;


  //
  // Membership methods
  //
  // Don't include the start/end tags.  These are our fence posts
  {
    oC << "//" << endl << "// Membership methods"<<endl << "//" << endl;
    size_t i;
    for (i=0;i<groups.size();i++) {
      oC << "bool "<<basename<<"::isGroup"<<groups[i]<<"(const "<<basename<<"Enum e) {" << endl
	 << "  const int val = int(e);" << endl
	 << "  if (val<=" << groupStart[i] << ") return(false);" << endl
	 << "  if (val>=" << groupEnd  [i] << ") return(false);" << endl
	 << "  return (true);" << endl
	 << "};" << endl << endl;
    }
  }


  //
  // Generate switch block of all tags
  //
  {
    string filename ("switch-tmp.C");
    std::ofstream o(filename.c_str(),std::ios::out);
    if (!o.is_open()) {
      cerr << "ERROR: failed to open for writing... " << filename << endl;
      exit(EXIT_FAILURE);
    }

    o << "  switch (LwTagEnum) {" << endl;
    size_t i;
    for (i=0;i<lines.size();i++) {
      string groupName;
      if (isStart(lines[i],groupName)) o << endl << "  //// Group: " << groupName << endl;
      o << "  case " << finalNames[i] << ":";
      if (isStart(lines[i],groupName) || isEnd(lines[i],groupName)) {
	o << " cerr << \"Should never get group tag: "<< finalNames[i] <<"\" << endl;"
	  << " break;" << endl;
      } else {
	o << " cerr << \"FI" << "X: implement tag - "<<finalNames[i]<<"\" << endl;"
	  << " break;" << endl;
      }
    }
    o << "  default: " << endl
      << "    cerr << \"switch ERROR: should never his default!\" << endl;" << endl
      << "  }" << endl;
      ;
  }

  // Close the header include guard
  oH << "#endif // _LW_TAGS_H_" << endl;

  return (EXIT_SUCCESS);
}
