/* $Revision: 1.3 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGridGdal
 *
 * DESCRIPTION: Use gdal to read/write lots of formats.
 *
 * EXPORTS:  
 *
 * NOTES:        
 *
 * 
 * SEE ALSO:
 * 
 *	http://www.remotesensing.org/gdal/
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

/***************************************************************************
 * INCLUDES
 ***************************************************************************/

//
// C headers
//
#include <cassert>
#include <cstdlib>

//
// C++ headers
//
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

//
// Library headers
//
#include <gdal.h>
#include <gdal_priv.h>
//
// Local headers
//
#include "XcGrid.H"
#include "XcGridDrv.H"
#include "XcGridFile.H"
#include "xc-gdal.H"

/***************************************************************************
 * GLOBALS
 ***************************************************************************/

bool init=false;
bool gdalInitialized=false;

/***************************************************************************
 * FUNCTIONS AND DEFINITIONS
 ***************************************************************************/
// http://www.remotesensing.org/gdal/formats_list.html
// Fix: switch to a vector shared between these two.
bool XcGridDrvGdal::canReadType(FileType t) {
  if (JPG==t) return (true);
  if (GIF==t) return (true);
  if (PNG==t) return (true);
  if (TIF==t) return (true);
  // FIX: There are lots more
  return (false);
}

bool XcGridDrvGdal::canWriteType(FileType t) {
  // Defaults to be same as write
  return (canReadType(t));
  // FIX: There are lots more
}

/***************************************************************************
 * FUNCTION: XcGridGdalWrite
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/

bool XcGridGdalWrite (std::string filename, XcGrid &grd) {
  if (!init) {
    init=true;
    GDALAllRegister();
  }
  assert (false);
} // XcGridGdalWrite

/***************************************************************************
 * FUNCTION: XcGridGdalRead
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/

bool XcGridGdalRead (std::string filename, XcGrid &grd) {
  if (!init) {
    init=true;
    GDALAllRegister();
  }
  GDALDataset *poDataset;
  
  poDataset = (GDALDataset *)GDALOpen(filename.c_str(), GA_ReadOnly );
  if (!poDataset) {
    std::cerr << "Failed to load dataset\n";
    return (EXIT_FAILURE);
  }

  std::cout << "Driver: "
	    << poDataset->GetDriver()->GetDescription() << " "
	    << poDataset->GetDriver()->GetMetadataItem(GDAL_DMD_LONGNAME) 
	    << "\n";

  std::cout << "Size is "
	    << poDataset->GetRasterXSize() << "x"
	    << poDataset->GetRasterYSize() << "x"
	    << poDataset->GetRasterCount() << "\n";


  if (poDataset->GetProjectionRef()) {
    if (0==strlen(poDataset->GetProjectionRef()))
	std::cout << "empty projection\n";
    else 
      std::cout << "Projection is "
		<< poDataset->GetProjectionRef() << "\n";
  }
  else
    std::cout << "No projection info\n";

  {
    double adfGeoTransform[6];
    if (CE_None ==poDataset->GetGeoTransform(adfGeoTransform)) {
      std::cout << "Origin = (" << adfGeoTransform[0] << ","
		<< adfGeoTransform[3] << ")\n";
      std::cout << "Pixel Size = (" << adfGeoTransform[1] << ","
		<< adfGeoTransform[5] << ")\n";
    } else std::cout << "No geo transform info\n";
      
  } 

  //
  // Now try to read some info about band 1	
  //

  {
    GDALRasterBand *poBand;
    int blockXSize, blockYSize;
    int bmin, bmax;
    double adfMinMax[2];

    poBand = poDataset->GetRasterBand(1);
    assert(poBand);
    poBand->GetBlockSize (&blockXSize, &blockYSize);
    std::cout << "block=" << blockXSize << "x" << blockYSize
	      << " Type = " << GDALGetDataTypeName(poBand->GetRasterDataType())
	      << ", ColorInterp=" 
	      << GDALGetColorInterpretationName(poBand->GetColorInterpretation())
	      << "\n";

    adfMinMax[0] = poBand->GetMinimum(&bmin);
    adfMinMax[1] = poBand->GetMaximum(&bmax);
    if (! (bmin && bmax) ) {
      std::cout << "Failed to get minmax.  Computing...\n";
      GDALComputeRasterMinMax((GDALRasterBandH)poBand, true, adfMinMax);
    }
    std::cout << "MinMax: " << adfMinMax[0] << " " <<  adfMinMax[1] << "\n";

    std::cout << "Band has " << poBand->GetOverviewCount() << " overviews\n";
    if (poBand->GetColorTable()) 
      std::cout << "Band has " << poBand->GetColorTable()->GetColorEntryCount() << " color tables\n";
    else
      std::cout << "no colortables\n";
  } // END READ INFO

  //
  // READ THE DATA.  Scan line at a time.
  //
  grd.setSize (poDataset->GetRasterXSize(),poDataset->GetRasterYSize());
  grd.setOrigin (0.,0.);  // FIX: get this from GDAL poDataset
  grd.setDelXDelY (1.,1.); // FIX: get from GDAL poDataset

  {
    float *line;
    GDALRasterBand *band;
    band = poDataset->GetRasterBand(1);
    assert(band);
    int nXSize = band->GetXSize();
    assert(band->GetXSize() == poDataset->GetRasterXSize());
    assert(band->GetYSize() == poDataset->GetRasterYSize());
    size_t x,y;
    line = new float[nXSize];
    for (y=0;y<size_t(band->GetYSize());y++) {
      band->RasterIO( GF_Read, 0,y, nXSize, 1,
		      line, nXSize, 1, GDT_Float32,
		      0,0);
      for (x=0;x<size_t(nXSize);x++) {
	grd.setZ(x,y,line[x]);
      }
    }
    delete [] line;
  }

  return (true);
} // XcGridGdalRead

//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////
std::string XcGridDrvGdal::getName() { return("GDAL"); }
//bool XcGridDrvGdal::canReadType(FileType t) {return (false);}
//bool XcGridDrvGdal::canWriteType(FileType t) {return(false);}
std::vector<FileType> XcGridDrvGdal::getReadTypes() {
  std::vector<FileType> types;
  types.push_back(GIF);
  types.push_back(PNG);
  types.push_back(TIF);
  types.push_back(JPG);
  return (types);
}
std::vector<FileType> XcGridDrvGdal::getWriteTypes() {
  return (getReadTypes());
}

bool XcGridDrvGdal::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  if (!canReadType(t)) {
    std::cerr << "unable to read selected type!\n";
    return (false);
  }
  bool r = XcGridGdalRead(filename, grd);
  return (r);
}
bool XcGridDrvGdal::WriteGrid(std::string filename, FileType t, XcGrid &grd, gengetopt_args_info &opts) {
  if (!canReadType(t)) {
    std::cerr << "unable to read selected type!\n";
    return (false);
  }
  bool r = XcGridGdalWrite (filename, grd);
  return (r);
}

//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST

// Have to provide out own register command so we don't depend on
// other drivers... Only register ourself.
XcG::VecDriverPtr XcG::XcGridDrvRegister() {
  VecDriverPtr drivers;
  drivers.push_back(new XcGridDrvGdal);
  return(drivers);
}


int main (int argc, char *argv[]) {
  bool ok=true;

  //
  // READ TESTS
  //
  {
    size_t i=0;
    std::vector<std::string> files;
    files.push_back("TestData/gray-4x4.gif");
    files.push_back("TestData/gray-4x4.jpg");
    files.push_back("TestData/gray-4x4.png");
    files.push_back("TestData/gray-4x4-mac-jpeg.tif");
    files.push_back("TestData/gray-4x4-mac-lzw.tif");
    files.push_back("TestData/gray-4x4-mac-none.tif");
    files.push_back("TestData/gray-4x4-xv.tif");

    for (i=0;i<files.size();i++) {
      bool r;
      XcGrid g;
      r=XcGridGdalRead(files[i],g);
      if (!r) ok=false;
      std::cout << "    Read " << files[i] << ": " << (r?"ok.":"FAILED!") << std::endl;
    }
  } // READ TESTS


  // Test the register interface.
  {
    VecDriverPtr drivers = XcGridDrvRegister();
  }

  std::cout << argv[0] << (ok?": ok.":": FAILED!")<<std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
} // MAIN
#endif // REGRESSION_TEST

