/* $Revision: 1.15 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGmtGrid.C
 *
 * DESCRIPTION: Fill in an XcGrid from a GMT grd netcdf file.
 *
 * EXPORTS:  
 *
 * NOTES:        
 *
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

/***************************************************************************
 * INCLUDES
 ***************************************************************************/

//
// C headers
//
#include <cassert>
#include <cstdlib>
#include <cmath> // GMT nan code uses isnan() from here
extern "C" { // GMT does not know about C++
#include <gmt.h>  // Generic Mapping Tools
} // extern "C"

//
// C++ headers
//
#include <iostream>
#include <string>
//
// Local headers
//
//#include "XcGrid.H" // Already included from XcGmtGrid.H
#include "XcGridDrv.H"
#include "xc-gmt.H"

/***************************************************************************
 * FUNCTION: PrintGmtHeader
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/

// See gmt_grd.h
void PrintGmtHeader(struct GRD_HEADER &h) {
  std::cout << "  ** GRD_HEADER **\n"
	    << "nx: " << h.nx << "\n"
	    << "ny: " << h.ny << "\n"
	    << "node_offset: " << h.node_offset << "\n"
	    << "x_min: " << h.x_min << "\n"
	    << "x_max: " << h.x_max << "\n"
	    << "y_min: " << h.y_min << "\n"
	    << "y_max: " << h.y_max << "\n"
	    << "z_min: " << h.z_min << "\n"
	    << "z_max: " << h.z_max << "\n"
	    << "x_inc: " << h.x_inc << "\n"
	    << "y_inc: " << h.y_inc << "\n"
	    << "z_scale_factor: " << h.z_scale_factor << "\n"
	    << "z_add_offset: " << h.z_add_offset << "\n"
	    << "x_units: " << h.x_units << "\n"
	    << "y_units: " << h.y_units << "\n"
	    << "z_units: " << h.z_units << "\n"
	    << "title: " << h.title << "\n"
	    << "command: " << h.command << "\n"
	    << "remark: " << h.remark << "\n";
  return;
};

/***************************************************************************
 * FUNCTION: XcGridGmtLoad
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/
bool
XcGridGmtLoad (std::string filename,XcGrid &grd) {
  char *file=strdup(filename.c_str());  // Don't forget to free

  struct GRD_HEADER hdr;
  int argc=1; char *argv[]={"XcGmtGridLoad",0};
  argc=GMT_begin(argc,argv); // Sets crazy globals
  GMT_grd_init (&hdr, argc, argv, false); // Initialize grd header structure 
  if (0!=access(filename.c_str(),R_OK)) {
    std::cerr << "ERROR: Unable to read file:" << filename << "\n";
    return(false);
  }
  GMT_read_grd_info (file,&hdr);

  grd.setRegular(true);
  grd.setSize(hdr.nx, hdr.ny);
  grd.setDelXDelY(hdr.x_inc,-hdr.y_inc); // -y_inc cause going south
  grd.setOrigin(hdr.x_min,hdr.y_max);  // FIX: really y max?  Upper left corner

  // See gmt_grdio.c
  // Formats listed in
  // http://gmt.soest.hawaii.edu/doc/html/GMT_Docs/node32.html
  // 0  is GMT netCDF 4-byte float format (default)
  // 11 is netCDF 8-byte double format
  {
    char fname[BUFSIZ];
    double scale = GMT_d_NaN, offset = 0.0;
    int format = GMT_grd_get_i_format (file,fname,&scale,&offset);
    //std::cout << "GMT format: " << format << " " << fname 
    //	      << " " << scale << " " << offset << "\n";
    assert (0==format); // Would be nice to handle others, but not yet
  }

  const size_t nm = hdr.nx * hdr.ny;
  float *a;
  a = (float *) GMT_memory (VNULL, (size_t)nm, sizeof (float), GMT_program);

  // Read the entire grd image
  if (GMT_read_grd (file, &hdr, a, 0.0, 0.0, 0.0, 0.0, GMT_pad, FALSE)) {
    //std::cout << "read return true.  This may be BAD?!?!\n";
  } else {
     // This is when it read okay....???
    //std::cout << "read return false.  This is okay?\n";
  }

  size_t i;
  for (i=0;i<nm;i++) {
    size_t x, y;
    grd.getXYfromIndex(i,x,y);
    if (GMT_is_fnan(a[i])) {
      grd.setZToNaN(x,y);
      continue;
    }
    grd.setZ(x,y,a[i]);
  }
 
  free(file);
  return(true);
}

/***************************************************************************
 * FUNCTION: XcGridGmtWrite
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/

bool
XcGridGmtWrite (std::string filename,XcGrid &grd) {
  bool ok=true;
  struct GRD_HEADER hdr;
  int argc=1; char *argv[]={"XcGmtGridWrite",0};
  argc=GMT_begin(argc,argv); // Sets crazy globals
  // Set header

  hdr.nx = grd.getWidth();
  hdr.ny = grd.getHeight();
  hdr.node_offset = 0; // 1 is pixel format, 0 is grid format(Normal  FIX: describe.
  hdr.x_min = grd.getXMin();
  hdr.x_max = grd.getXMax();
  hdr.y_min = grd.getYMin();
  hdr.y_max = grd.getYMax();
  hdr.z_min = grd.getZMin();
  hdr.z_max = grd.getZMax();
  hdr.x_inc = grd.getDX();
  hdr.y_inc = grd.getDY(); // due to grid shape
  //std::cout << "yinc: " << hdr.y_inc << std::endl;
  hdr.z_scale_factor = 1.0;
  hdr.z_add_offset = 0.0;
  strcpy (hdr.x_units, "xunits");//hdr.x_units = "xunits"; // GRD_UNIT_LEN
  strcpy (hdr.y_units,"yunits");
  strcpy (hdr.z_units,"zunits");
  strcpy (hdr.title,"Title"); // GRD_TITLE_LEN
  strcpy (hdr.command,"Some command"); // GRD_COMMAND_LEN
  strcpy (hdr.remark,"Remark"); // GRD_REMARK_LEN

  //PrintGmtHeader(hdr);

  // Set data block
  float *data=new float[grd.getWidth()*grd.getHeight()];
  assert (data);
  {
    size_t x,y;
    for (y=0;y<grd.getHeight();y++) {
      for (x=0;x<grd.getWidth();x++) {
	data[x+y*grd.getWidth()] = float(grd.getZ(x,y));
	//std::cout <<x<<" "<<y<<" "<<grd.getZ(x,y)<<std::endl;
      }
    }
  }

  // GMT_write_grd
  //GMT_pad[0]=GMT_pad[1]=GMT_pad[2]=GMT_pad[3];
  int pad[4]={0,0,0,0}; // Don't use evil gmt global GMT_pad in gmt_init.h
  double w=0,e=0,n=0,s=0; // Writes out the whole grid
  char *name=strdup(filename.c_str());
  if (GMT_write_grd(name, &hdr, data, w,e,s,n,pad, false)) {
    std::cerr << "GMT_write_grd failed\n";
    ok=false;
  }
  free (name);
  // Free up everything
  delete  [] data;

  // Go home
  return (ok ? true : false);
}

//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////

std::string XcGridDrvGmt::getName() { return("GMT");}
//bool XcGridDrvGmt::canReadType(FileType t) {return (false);}
//bool XcGridDrvGmt::canWriteType(FileType t) {return(false);}
std::vector<FileType> XcGridDrvGmt::getReadTypes() {
  std::vector<FileType> types;
  types.push_back(GMT0);
  return (types);
}
std::vector<FileType> XcGridDrvGmt::getWriteTypes() {
  std::vector<FileType> types;
  types.push_back(GMT0);
  return (types);
}


bool XcGridDrvGmt::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  assert (GMT0==t);
  bool r=XcGridGmtLoad(filename, grd);
  return(r);
}

bool XcGridDrvGmt::WriteGrid(std::string filename, FileType t, XcGrid &grd, gengetopt_args_info &opts) {
  assert (GMT0==t);
  bool r=XcGridGmtWrite(filename,grd);
  return (r);
}



//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////


#ifdef REGRESSION_TEST

#include "XcGridFile.H"
#include "XcGridDrv.H"

using namespace std;

// Have to provide out own register command so we don't depend on
// other drivers... Only register ourself.
XcG::VecDriverPtr XcG::XcGridDrvRegister() {
  VecDriverPtr drivers;
  drivers.push_back(new XcGridDrvGmt);	// Generic Mapping Tool
  return(drivers);
}



int main (int argc, char *argv[]) {
  bool ok=true;
  {
    XcGrid g(3,3);
    g.setDelXDelY(1,-1);
    g.setOrigin(0,0);
    g.setZ(0,0,0);
    g.setZ(1,0,1);
    g.setZ(2,0,2);

    g.setZ(0,1,3);
    g.setZ(1,1,4);
    g.setZ(2,1,5);

    g.setZ(0,2,6);
    g.setZ(1,2,7);
    g.setZ(2,2,8);

    std::string filename="test1.grd";
    bool r=true;
    if (!XcGridGmtWrite(filename,g))  ok = r = false;
    std::cout << "  XcGridGmtWrite " << (ok?": ok.":": FAILED!") << std::endl;
  }

  std::vector<std::string> files;
  files.push_back("TestData/eel-em1000-subsample5.grd");

  size_t i=0;
  for (i=0;i < files.size();i++) {
    bool r=true;
    //std::string s = files[i];
    XcGrid grd;
    if (!XcGridGmtLoad (files[i],grd)) r=ok=false;
    std::cout << "  " << files[i] << (r?": ok.":": FAILED!") << std::endl;
  }

  {
    VecDriverPtr drivers = XcGridDrvRegister();  // vector of drivers
    bool r = (1==drivers.size()); // Should get at least 3 drivers (gmt, pds, pgm)
    if (!r) ok=false;
    std::cout << "  XcGridDrvRegister test: " << (ok?"ok.":": FAILED!") << std::endl;

    // FIX: what is the not found for vector, it's NOT npos??
    if (r) {
      size_t drv;
      drv = XcGridDrvGetReader(drivers,GMT0);
      if (string::npos==drv || !drivers[drv]->canReadType(GMT0)) {
	std::cout << "  ERROR:  gmt driver failed to respond correctly" << std::endl;
	ok=false;
      }
      drv = XcGridDrvGetWriter(drivers,GMT0);
      if (string::npos==drv || !drivers[drv]->canWriteType(GMT0)) {
	std::cout << "  ERROR:  gmt driver failed to respond correctly" << std::endl;
	ok=false;
      }
    }

    // Test reading.  Btw, we know we have the driver at position 0.
    {
      //XcGridOpts opts;
      gengetopt_args_info opts;
      gengetopt_args_info_INIT(opts);

      XcGrid g;
      std::string filename("TestData/eel-em1000-subsample5.grd");
      r = drivers[0]->ReadGrid(filename,GMT0,g);
      if (!r) ok=false;
      std::cout << "  XcGridDrvRead  test: " << (r?"ok.":"FAILED!") << std::endl;
    }

    // Test write
    {
      //XcGridOpts opts;
      gengetopt_args_info opts;
      gengetopt_args_info_INIT(opts);

      cerr << "FIX: What to do about y range being negative?" << endl;

      XcGrid g(2,2);
      g.setDelXDelY(1.0,1.0);
      g.setZ(0,0,0);
      g.setZ(1,0,1.1);
      g.setZ(0,1,2.2);
      g.setZ(1,1,3.3);

      std::string filename("test-gmt-write.grd");
      r = drivers[0]->WriteGrid(filename,GMT0,g,opts);
      if (!r) ok=false;
      std::cout << "  XcGridDrvWrite test: " << (r?"ok.":"FAILED!") << std::endl;
    }

    std::cout << "  XcGridDrv->getName test: ";
    std::cout.flush();
    cout << drivers[0]->getName() << endl;

    XcGridDrvDelete(drivers);
    r = (0==drivers.size());
    if (!r) ok=false;
    std::cout << "  XcGridDrvDelete test: " << (r?"ok.":"FAILED!") << std::endl;

  } // Registry test



  std::cout << argv[0] << (ok?": ok.":": FAILED.") << std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif // REGRESSION_TEST
