/* $Revision: 1.22 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGridWriteIV.C
 *
 * DESCRIPTION: Use XcGrid to write an OpenInventor mesh
 *
 * EXPORTS:  
 *
 * NOTES:        
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

//#include <cmath>
#include <iostream>
#include <fstream>
//using namespace std;

#include "XcGrid.H"
#include "xc-inventor.H"

/***************************************************************************
 * FUNCTION: XcGridWriteIV
 * DESCRIPTION: Create a basic IndexedFaceSet in OpenInventor ascii format
 * INPUTS:  texture    - true to write texture map
 *	    normals    - true to use XcGrid normals
 *	    colortable - true to use Colortable   FIX: implement
 *	    textureFile - If texture is true, the file name in here will be the texture
 * OUTPUTS: 
 * RETURNS: true on success, false on failure
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 ***************************************************************************/

// FIX: need to start using the compressed index!! Reduce the output size

bool
XcGridWriteIV (const std::string filename, XcGrid &grd,
	       bool texture, bool normals, bool colortable, bool normalSpikes) {
  std::ofstream o(filename.c_str(),std::ios::out);
  if (!o.is_open()) {
    std::cerr << " output file open failed for "<< filename << std::endl;
    return (false);
  }

  o << "#Inventor V2.1 ascii\n"
    << "\n"
    << "# Written using XcGridWriteIV from Kurt Schwehr's X-Core package\n"
    << "# XcGridWrite $Revision: 1.22 $ \n"
    << "# http://schwehr.org/xcore\n"
    << "\n";

  // Write all log entries as comments.
  {
    o << " ###############"
      << " # LOG ENTRIES #"
      << " ###############\n\n";

    o << " # NumLogEntries = " << grd.getNumLogEntries() << "\n";
    size_t i;
    for (i=0;i<grd.getNumLogEntries();i++) {
      o << " # "
	<< grd.getLogEntry(i)
	<< "\n";
    }
    o << "\n # END LOG ENTRIES\n\n";
  }


  o << "Separator { # Start " << filename << "\n";

  // creaseAngle causes coin/Inventor to generate normals for us.
  // This makes the model look tons better!
  o << "  ShapeHints {\n"
    << "     vertexOrdering COUNTERCLOCKWISE\n"
    //<< "     vertexOrdering CLOCKWISE\n" //FIX: expected COUNTERCLOCKWISE!!!
    //<< "     faceType CONVEX\n"
    //<< "     faceType UNKNOWN_FACE_TYPE\n"
    //<< "     shapeType SOLID\n"
    //<< "     shapeType UNKNOWN_SHAPE_TYPE\n"
    ;
  if (!normals)  o << "     creaseAngle 1.57\n";
  o << "  }\n";

  //
  // Setup output stream floating point writing.  Want best precision.
  // 
  o.setf (std::ios_base::fixed,std::ios_base::floatfield);
  const int precision=12;  // FIX: what is really a good number for this?
  o.precision (precision);

  ////////////////////////////////////////
  // TEXTURE
  ////////////////////////////////////////

  // Per vertex texture map
  // Would be more efficient to loop over x and y instead of the index
  if (texture) {
    o << "  Texture2 { filename \"" << grd.getTextureFile() << "\" }\n"
      << "  TextureCoordinate2 { point [ \n";

    const size_t nelemts=grd.getWidth()*grd.getHeight();
    size_t i,xoff,yoff;
    float u,v;
    for (i=0;i<nelemts;i++) {
      if (grd.isNaN(i)) continue;
      grd.getXYfromIndex(i,xoff,yoff);
      grd.getUV(xoff,yoff,u,v);
      o << "    " << u << " " << v << ",\n";
    }

    o << "    ] # texture point\n"
      << "  } # TextureCoordinate2\n";
  } // texture

  ////////////////////////////////////////
  // NORMALS
  ////////////////////////////////////////

  if (normals) {
    o.precision (precision); // Reset to higher precision
    //o.precision (4); // FIX: Make sure this is not too small
    o << "  Normal { vector [\n";
    size_t i,xoff,yoff,nelemts=grd.getWidth()*grd.getHeight();
    float nx,ny,nz;
    for (i=0;i<nelemts;i++) {
      if (grd.isNaN(i)) continue;
      grd.getXYfromIndex(i,xoff,yoff);
      grd.getNormal(xoff,yoff,nx,ny,nz);
      o << "      " << nx << " " << ny << " " << nz << ",\n";
    }
    o << "    ] # normal vectors\n"
      << "  } # End Normals\n\n";
    o.precision (precision); // Reset to higher precision
  } // normals


  ////////////////////////////////////////
  // COLOR MAP/TABLE
  ////////////////////////////////////////

  if (colortable) {
    o.precision (3); // Really can't see more colors than this!
    assert (!texture); // What happens if we have both?
    o << "  MaterialBinding { value PER_VERTEX_INDEXED }\n"
      << "  BaseColor { rgb [ \n";
    size_t i,xoff,yoff,nelemts=grd.getWidth()*grd.getHeight();
    float r,g,b;
    for (i=0;i<nelemts;i++) {
      if (grd.isNaN(i)) continue;
      grd.getXYfromIndex(i,xoff,yoff);
      grd.getRGB(xoff,yoff,r,g,b);
      o << "      " << r << " " << g << " " << b << ",\n";
    }
    o << "    ] # rgb \n"
      << "  } # BaseColor\n\n";
    o.precision (precision); // Reset to higher precission
  } // colortable

  ////////////////////////////////////////
  // COORDINATES
  ////////////////////////////////////////
  {
    o.precision (precision); // Reset to higher precission
    o << "  Coordinate3 { point [\n";
    size_t i,xoff,yoff,nelemts=grd.getWidth()*grd.getHeight();
    float x,y,z;
    for (i=0;i<nelemts;i++) {
      if (grd.isNaN(i)) continue;
      grd.getXYfromIndex(i,xoff,yoff);
      grd.getXYZ(xoff,yoff,x,y,z);
      o << "      " << x << " " << y << " " << z << ",\n";
    }
    o << "    ] # point\n"
      << "  } # Coord3\n\n";
  } // Coords

  // FIX: make sure the coordinates actually match up ok!!

  ////////////////////////////////////////
  // INDEX TO COORDS
  ////////////////////////////////////////
  {
    o << "  IndexedFaceSet { coordIndex [\n";
    size_t i;
    const size_t numTri = grd.getNumTriangles();
    for (i=0; i<numTri; i++) {
      XcTriangle t = grd.getTriangle(i);
      //if (grd.isNaN(i)) continue; // This i has nothing to do with NaN's!
      o << "      " 
	<< grd.getCompressedIndex(t.vertices[0]) << ","
	<< grd.getCompressedIndex(t.vertices[1]) << ","
	<< grd.getCompressedIndex(t.vertices[2]) << ",-1,\n";
    }
    o << "    ] # coordIndex\n"
      << "  } # IndexedFaceSet\n";
  } // Coord indexes

  ////////////////////////////////////////
  // DEBUGGING OUTLINES
  ////////////////////////////////////////
  if (false) {
    o << "\n  ##### DEBUGGING OUTLINES #####\n\n"
      << "  BaseColor { rgb [1 0 1] }\n\n";
    o.precision (precision); // Reset to higher precission
    size_t i;
    const size_t numTri = grd.getNumTriangles();
    for (i=0; i<numTri; i++) {
      XcTriangle t = grd.getTriangle(i);
      float x,y,z;
      o << "  Coordinate3{point[";
      grd.getXYZ(t.vertices[0], x,y,z);
      o << x <<" "<<y<<" "<<z+0.1<<",";
      grd.getXYZ(t.vertices[1], x,y,z);
      o << x <<" "<<y<<" "<<z+0.1<<",";
      grd.getXYZ(t.vertices[2], x,y,z);
      o << x <<" "<<y<<" "<<z+0.1<<",";
      o << "]}\n";

      o << "  IndexedLineSet { coordIndex [ 0 1 2 0 ] }\n";
    } // for i
  } // Coord indexes



  ////////////////////////////////////////
  // NORMAL SPIKES
  ////////////////////////////////////////
  if (normalSpikes) {
    o << "\n\n"
      << "  Separator { # Normal Spikes - mostly for debugging\n"
      << "    BaseColor { rgb [ 1 .5 0 ] }\n";
    size_t x,y;
    for (y=0;y<grd.getHeight();y++) {
      for (x=0;x<grd.getWidth();x++) {
	float x1,y1,z1,x2,y2,z2,nx,ny,nz;
	grd.getXYZ(x,y,x1,y1,z1);
	grd.getNormal(x,y,nx,ny,nz);
	x2 = x1 + nx * grd.getDX();
	y2 = y1 + ny * grd.getDX();
	z2 = z1 + nz * grd.getDX();
	if (grd.isNaN(x,y)) continue;
	// FIX: Try using SoLineSet or something more compact
	o << "    Coordinate3{point["
	  << x1 << " " << y1 << " " << z1 << ","
	  << x2 << " " << y2 << " " << z2 << "]}\n";
	// FIX:  make this a def and run reuse!
	o << "    IndexedLineSet{coordIndex[0,1]}\n";
	  
      }
    } // for y

    o << "  } # Separator Normal Spikes\n";
  }

  ////////////////////////////////////////
  // AXES FOR DEBUGGING
  ////////////////////////////////////////
#if 0
  {
    o << "  Separator { # X\n"
      << "    BaseColor { rgb [ 1 0 0] }\n"
      << "    Coordinate3{point[ 0 0 -0.1, 1 0 -0.1]}\n"
      << "    IndexedLineSet{coordIndex[0,1]}\n"
      << "  } # Separator \n\n";
 
    o << "  Separator { # Y\n"
      << "    BaseColor { rgb [ 0 1 0] }\n"
      << "    Coordinate3{point[ 0 0 -0.1, 0 1 -0.1]}\n"
      << "    IndexedLineSet{coordIndex[0,1]}\n"
      << "  } # Separator \n\n";
 
    o << "  Separator { # Z\n"
      << "    BaseColor { rgb [ 0 0 1] }\n"
      << "    Coordinate3{point[ 0 0 -0.1, 0 0 1]}\n"
      << "    IndexedLineSet{coordIndex[0,1]}\n"
      << "  } # Separator \n";
  }
#endif

  o << "\n\n} # End of Start Separator for " << filename << "\n\n"
    << "# Have a nominal day.  Good bye.\n";

  return(true);
}


//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////

#include "XcGridDrv.H"
using namespace XcG;

std::string XcGridDrvIv::getName() { return("IV");}
//bool XcGridDrvIv::canReadType(FileType t) {return (false);}
//bool XcGridDrvIv::canWriteType(FileType t) {return(false);}
std::vector<FileType> XcGridDrvIv::getReadTypes() {
  std::vector<FileType> types;
  //types.push_back(IV);
  return (types);
}
std::vector<FileType> XcGridDrvIv::getWriteTypes() {
  std::vector<FileType> types;
  types.push_back(IV);
  return (types);
}

bool XcGridDrvIv::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  assert (IV==t);
  return (false); // Can't write nuthin.
}
bool XcGridDrvIv::WriteGrid(std::string filename, FileType t, XcGrid &grd, gengetopt_args_info &opts) {
  assert (IV==t);
  std::cerr << "FIX: handle texture\n";

  grd.genCompressedIndex();
  grd.genTriangles();
  if (opts.normals_flag || opts.spikes_flag) {
    grd.genNormals();
  }
  grd.genUV();

  bool useTexture=false;
  if (opts.tex_given) {
    useTexture=true;
    if (grd.getTextureFile().size() > 0) {
      std::cout << "Warning: texture file name is over writing previous value: " << std::endl
		<< "  Was: " << grd.getTextureFile() << std::endl
		<< "  Now: " << opts.tex_arg << std::endl;
    }
    grd.setTextureFile(opts.tex_arg); // May be overwriting a desired value.
  }

  std::cout << "  Writing inventor file " << filename << ":" << std::endl
	    << "    useTexture:       " << (useTexture?"true":"false") << std::endl
	    << "    textureFile:      " << grd.getTextureFile() << std::endl
	    << "    expliciteNormals: " << (opts.normals_flag?"true":"false") << std::endl
	    << "    colorTable:       " << (opts.cpt_given?"true":"false") << std::endl
	    << "    normalsSpikes:    " << (opts.spikes_flag?"true":"false") << std::endl;


  bool r = XcGridWriteIV(filename, grd, 
			 useTexture, 
			 opts.normals_flag,
			 opts.cpt_given,
			 opts.spikes_flag);
  return (r);
}


//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST

#include "XcGridSamples.H"

// Have to provide out own register command so we don't depend on
// other drivers... Only register ourself.
XcG::VecDriverPtr XcG::XcGridDrvRegister() {
  VecDriverPtr drivers;
  drivers.push_back(new XcGridDrvIv);	// JPL VICAR planetary data format
  return(drivers);
}

int main (int argc, char *argv[]) {
  bool ok=true;

  {
    XcGrid *g=XcGridSampleOK();
    g->genCompressedIndex();
    g->genTriangles();
    g->genNormals();
    g->genUV();
    bool r;
    r=XcGridWriteIV("test-SampleOK.iv",*g,false,true,false,true);
    if (!r) {
      ok = false;
    }
    std::cout << "  SampleOK: " << (r?"ok.":"FAILED!") << std::endl;
    delete g;
  }


  // SIMPLE TEST 
  if (true){
    std::cout << "  Simple test 1:";
    XcGrid g(2,2);
    g.setDelXDelY(1.0,1.0);
    g.setZ(0,0,0.01);
    g.setZ(1,0,0);
    g.setZ(0,1,0);
    g.setZ(1,1,0);
    //g.setZToNaN(0,0);
    g.genCompressedIndex();
    g.genTriangles();
    g.genNormals();
    g.genUV();
    if (!XcGridWriteIV("test1.iv",g,false,true,false,true)) {
      std::cout << "  FAILED." << std::endl;
      ok=false;
    } else {
      std::cout << "  ok." << std::endl;
    }
  }



  // SIMPLE TEST 2
  if (true) {
    XcGrid g(2,2);
    g.setDelXDelY(1.0,1.0);
    g.setZ(0,0,0.3);
    g.setZ(1,0,0);
    g.setZ(0,1,0);
    g.setZ(1,1,0);
    g.setZToNaN(1,1);
    g.genCompressedIndex();
    g.genTriangles();
    g.genNormals();
    g.genUV();
    if (!XcGridWriteIV("test2.iv",g,false,true,false,true)) {
      std::cout << "  test2: FAILED." << std::endl;
      ok=false;
    } else {
      std::cout << "  test2: ok." << std::endl;
    }
  }


  // SIMPLE TEST 3
  if (true) {
    XcGrid g(3,4);
    g.setDelXDelY(1.0,-1.0);
    g.setOrigin(0,0);
    g.setZ(0,0,100);
    g.setZ(1,0,110);
    g.setZ(2,0,90);

    g.setZ(0,1,0); //g.setZToNaN(0,1);
    g.setZ(1,1,0);
    g.setZ(2,1,0);

    g.setZ(0,2,-50);
    g.setZ(1,2,-80);
    g.setZ(2,2,-90);

    g.setZ(0,3,-150);
    g.setZ(1,3,-180);
    g.setZ(2,3,-190);

    g.scaleZ(0.5,0);
    //g.setZToNaN(1,2);
    //g.setZToNaN(2,2);

    //g.setZToNaN(1,1);
    g.genCompressedIndex();
    g.genTriangles();
    g.genNormals();
    g.genUV();
    if (!XcGridWriteIV("test3.iv",g,false,false,false,true)) {
      std::cout << "  test3: FAILED." << std::endl;
      ok=false;
    } else {
      std::cout << "  test3: ok." << std::endl;
    }
  }

  // SIMPLE TEST 4
  if (true) {
    XcGrid g(3,3);
    g.setOrigin(-124.593, 41.);
    g.setDelXDelY(0.00724638,-0.00434783);
    g.setZ(0,0,-455.88);
    g.setZ(1,0,-443);
    g.setZ(2,0,-440);

    g.setZ(0,1,-401);
    g.setZ(1,1,-380);
    g.setZ(2,1,-390);

    g.setZ(0,2,-350);
    g.setZ(1,2,-323);
    g.setZ(2,2,-301);

    //g.setZToNaN(1,1);
    g.genCompressedIndex();
    g.genTriangles();
    g.genUV();
    g.scaleZ(1/10000.,0);
    g.genNormals();
    if (!XcGridWriteIV("test4.iv",g,false,true,false,true)) {
      std::cout << "  FAILED." << std::endl;
      ok=false;
    } else {
      std::cout << "  Simple test 4: ok." << std::endl;
    }
  } // Simple test 4

  // Drv Test
  {
    VecDriverPtr drivers = XcGridDrvRegister();  // vector of drivers
    bool r = (1==drivers.size());
    if (!r) ok=false;
    std::cout << "  XcGridDrvRegister test: " << (r?"ok.":": FAILED!") << std::endl;

    // FIX: what is the not found for vector, it's NOT npos??
    if (r) {
      size_t drv;
      
      drv = XcGridDrvGetWriter(drivers,IV);
      //XcGridDrvPrint(drivers);  // FIX: remove print/cout
      if (std::string::npos==drv || !drivers[drv]->canWriteType(IV)) {
	ok=r=false;
      }
      std::cout << "  XcGridDrvGetWriter test: " << (r?"ok.":": FAILED!") << std::endl;
    }
  } // Drv Test


  std::cout << argv[0] << (ok?": ok.":": FAILED!")<<std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
} // main


#endif // REGRESSION_TEST
