/* $Revision: 1.17 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    xc-lightwave
 *
 * DESCRIPTION: Try to handle lightwave objects (lwo)
 *
 * EXPORTS:  
 *
 * NOTES:        
 *     LightWave 3D is by NewTek: http://www.newtek.com/products/lightwave/index.php
 *     File format: http://www.sandbox.de/osg/lightwave.htm
 *       LightWave 3D Object File Format  Oct 16, 1996
 *
 *     lwo's are in EA IFF 85
 *	http://www.newtek.com/products/lightwave/developer/75lwsdk/docs/filefmts/eaiff85.html
 *
 *     LWO2 format (2001)
 *	http://www.newtek.com/products/lightwave/developer/75lwsdk/docs/filefmts/lwo2.html
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

// FIX:  Handle byte order for non Mac OSX machines!!!
// FIX:  Make sure that a float really is 4 bytes


/***************************************************************************
 * INCLUDES
 ***************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 	// open
#include <sys/mman.h>	// mmap

#include <cmath> // Why does it undefine isnan?
extern "C" int isnan (double);  // FIX: what is the right way to do this???

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace std;

#include "xc-lightwave.H"

/***************************************************************************
 * TYPES
 ***************************************************************************/



layer::layer() {
  number=0; flags=0;
  pivot.X=pivot.Y=pivot.Z=0.;
  name = string("DEFAULT_LAYR");
  parent=0;
}

void layer::print () {
  cout << "LAYR: " << number << " " << flags 
       << " '"  << name << "' " << parent << endl;
}


// constructor
BlockHeader::BlockHeader () {
  CHAN_unknown = 0;
  CHAN_id = string ("not set");
  enabled = true;  // default to on in spec
  OPAC_type = UNKNOWN_OPACITY;
  opacity = 0.;
  envelope = 666;
  displacement_axis = UNKNOWN_AXIS;
}

static char *OpacityStrings[] = {
  "NORMAL", "SUBTRACTIVE",
  "DIFFERENCE",
  "MULTIPLY",
  "DIVIDE",
  "ALPHA",
  "TEXTURE_DISPLACEMENT",
  "ADDITIVE",
  "UNKNOWN_OPACITY"
};

void BlockHeader::print () {
  cout << "BlockHeader:" << endl
       << "  CHAN_unknown " << CHAN_unknown << endl
       << "  CHAN_id      " << CHAN_id << endl
       << "  enabled      " << (enabled?"on":"off") << endl
       << "  OPAC_type    " << OpacityStrings[OPAC_type] << endl
       << "  opacity      " << opacity << endl
       << "  envelope     " << envelope << endl
       << "  axis         " << displacement_axis << endl
    ;
  
}
/***************************************************************************
 * LOCAL FUNCTIONS
 ***************************************************************************/

// FIX: handle byte order!!!!
U4 getU4 (const char *buf, const size_t len) {assert(len>=4); U4 *tmp=(U4 *)buf;
  cout << "U4: " << *tmp << endl;
  printBuf (buf,4);
  cout << "end U4" << endl;
  return(*tmp);}
U2 getU2 (const char *buf, const size_t len) {
  assert(len>=2); 
  U2 *tmp=(U2 *)buf;
  cout << "  getU2: " << *tmp << endl;
  return(*tmp);
}
U1 getU1 (const char *buf, const size_t len) {assert(len>=1); U1 *tmp=(U1 *)buf;return(*tmp);}

I4 getI4 (const char *buf, const size_t len) {assert(len>=4); I4 *tmp=(I4 *)buf;return(*tmp);}
I2 getI2 (const char *buf, const size_t len) {assert(len>=2); I2 *tmp=(I2 *)buf;return(*tmp);}
I1 getI1 (const char *buf, const size_t len) {assert(len>=1); I1 *tmp=(I1 *)buf;return(*tmp);}

F4 getF4 (const char *buf, const size_t len) {assert(len>=4); F4 *tmp=(F4 *)buf;return(*tmp);}

string getID4 (const char *buf, const size_t len) {
  assert(buf);
  assert(len>=4);
  return (string(buf,4));
}

// Returns the index and the size (either 2 for U2 or 4 for U4)
U4 getVX (const char *buf, const size_t len, size_t &size) {
  assert (buf);
  assert (len>=0);
  unsigned char *mask = (unsigned char *)buf;
  if (0xFF == *mask) {
    assert (false); // need to mask out the 0xFF
    cout << "Found 4 byte index!!" << endl;
    assert (len>=4);
    size=4;
    return (getU4(buf,len));
  }
  size=2;
  return (getU2(buf,len));
}


// FIX: make S0 a class.
string getS0(const char *buf, const size_t len) {
  size_t offset;
  //printBuf (buf,(30<len)?30:len);
  for (offset=0;offset < len && buf[offset]!='\0'; offset++);
  if ('\0'!=buf[offset]) {
    cerr << "getS0 ERROR: failed to find end of string in buffer" << endl;
    return (string ("getS0 ERROR"));
  }
  string s(buf,offset);
  cout << "  getS0() size: " << s.size() << endl; // FIX: remove
  return (s);
}

// Figure out how many bytes this string takes up.
size_t getS0Len(string s) {
  size_t l = s.size();
  if (s[s.size()-1] != '\0') l++;
  if (1==l%2) l++; // nul pads to make it fall on an even boundary;
  return (l);
}


//FNAM0 ::= name[S0]
//Filenames are written as strings in a platform-neutral format. For
//absolute (fully qualified) paths, the first node represents a disk
//or similar storage device, and its name is separated from the rest
//of the path by a colon. Other nodes in the path are separated by
//forward slashes. disk:path/file is an absolute path, and
//path/subpath/file is a relative path.

string getFNAM0 (const char *buf, const size_t len) {
  assert(buf);
  assert(len>0);
  assert(len<30000);

  const U2 strLen = getU2 (buf,len);
  string s(buf+2,strLen);
  // Chop off trailing nuls
  while (0==s[s.size()-1]) {
    s = s.substr(0,s.size()-2); // FIX: use erase
  }
  return (s);
}

void
coordinate::getFromBuf(const char *buf,size_t bufLen) {
  assert (bufLen>=getSizeOf());
  assert(buf);
  char *cur=const_cast<char *>(buf);
  X = getF4(cur,bufLen); cur+=4;
  Y = getF4(cur,bufLen); cur+=4;
  Z = getF4(cur,bufLen); //cur+=4;
}

void
coordinate::print() {
  cout << "coord:  " << X << " " << Y << " " << Z << endl;
}

void
printBuf (const char *buf, const size_t len) {
  size_t i;
  for (i=0;i<len;i++)
    cout << i << ": " << (isprint(buf[i])?buf[i]:'*')
	 << " Int: " << int((unsigned char)(buf[i])) << endl;
}



//////////////////////////////////////////////////////////////////////
// PARSE EACH TYPE OF TAG
//////////////////////////////////////////////////////////////////////

// TAGS usually give the name of the object.
// This is a string.
string
GetTagTAGS(const char *buf,size_t bufLen, const U4 tagLength) {
  cerr << bufLen << " " << tagLength << endl;
  assert (tagLength <= bufLen-8);
  //return (string(buf+8,tagLength-1));  // Last byte is a nul, so we don't need to include it.

  string s = getS0(buf+8,tagLength);
  return (s);  // Last byte is a nul, so we don't need to include it.
}

// Start of a new LAYR
// LAYR { number[U2], flags[U2], pivot[VEC12], name[S0], parent[U2] ?  }

layer
GetTagLAYR(const char *buf,size_t bufLen, const U4 tagLength) {
  cerr << bufLen << " taglength:" << tagLength << endl;
  assert (tagLength <= bufLen-8);
  char *cur=const_cast<char *>(buf)+8;
  layer l;
  l.number = getU2(cur,bufLen); cur+=2;
  l.flags =  getU2(cur,bufLen); cur+=2;
  l.pivot.X = getF4(cur,bufLen); cur+=4;
  l.pivot.Y = getF4(cur,bufLen); cur+=4;
  l.pivot.Z = getF4(cur,bufLen); cur+=4;
  size_t stringlen=tagLength - 2 - (cur-(buf+8)); // +1; Why the +1?  Seems to get a NUL
  cout << "stringlen = " << stringlen << endl;
  l.name = string (cur,stringlen); cur += stringlen;
  //cout << "=> " << (isprint(cur[0])?cur[0]:'*')<<" Int: " << int(cur[0]) << endl;
  //cout << "=> " << (isprint(cur[1])?cur[1]:'*')<<" Int: " << int(cur[1]) << endl;

  // FIX: parent seems to be U1?!?!
  //l.parent = getU2(cur,bufLen); cur+=2;  // Spec says "U2 ?"
  l.parent = getU1(cur,bufLen); cur+=2;
  return (l);
}


// PNTS { point-location[VEC12] * }
// VEC12 ::= X[F4], Y[F4], Z[F4]


bool
GetTagPNTS(const char *buf,size_t bufLen, const U4 tagLength,vector<coordinate> &coords) {
  assert(buf);
  coords.clear();
  assert (tagLength%4==0);
  const size_t numPoints=tagLength/(4*3); // 3 floats of 4 bytes each: 4*3
  cout << "numPoints: " << numPoints << "  tagLength = " << tagLength << "\n";
  char *cur=const_cast<char *>(buf)+8;
  size_t i;
  for (i=0;i<numPoints;i++) {
    coordinate c;
    char *base = cur+(i*c.getSizeOf());

    cout << i<< " offset = "<< (int)(base-buf) << " "; //<< endl;

    c.getFromBuf(base, tagLength - (cur - base));
    coords.push_back(c);
    c.print();
  }
  cout << "Points Read: " << coords.size() << endl; 
  return (true);
}

// BBOX { min[VEC12], max[VEC12]  }

bool
GetTagBBOX(const char *buf,size_t bufLen, const U4 tagLength,bbox &b) {
  assert(buf);
  assert (tagLength== 2*b.min.getSizeOf());
  b.min.getFromBuf(buf+8,bufLen-8);
  b.max.getFromBuf(buf+8+b.min.getSizeOf(),bufLen-8-b.min.getSizeOf());
  b.min.print();
  b.max.print();
  return (true);
}




//////////////////////////////////////////////////////////////////////
//  VX ::= index[U2] | (index + 0xFF000000)[U4]
// length 2- or 4-byte element. If the index value is less than 65,280
// (0xFF00), then the index is written as an unsigned two-byte
// integer. Otherwise the index is written as an unsigned four byte
// integer with bits 24-31 set. When reading an index, if the first
// byte encountered is 255 (0xFF), then the four-byte form is being
// used and the first byte

// POLS { type[ID4], ( numvert+flags[U2],  vert[VX] # numvert )* }

bool
GetTagPOLS(const char *buf,size_t bufLen, const U4 tagLength,vector<Polygon> &polys) {
  polys.clear();
  assert (buf);
  char *cur=const_cast<char *>(buf)+8;
  size_t remain = size_t(tagLength);

  PolyEnum pe = GetPolyEnum (cur,tagLength);
  if (FACE != pe) {
    cerr << "WARNING: probably only know how to handle FACE polys." << endl;
  }
  cur +=4; remain -= 4;

  while (remain>0 && cur < buf+bufLen) {
    U2 numvert_flags = getU2(cur,remain);
    const size_t numvert=(numvert_flags & 0x3FF);

    // FIX: deal with flags.

    cout << "numvert = " << numvert_flags << " " << numvert << endl;
    cur +=2; remain -=2;
    size_t i;
    Polygon p;
    for (i=0;i<numvert;i++) {
      U2 index2 = getU2(cur,remain);
      assert(index2< 0xFF00);  // FIX: need to handle 4 byte indexes
      p.indices.push_back(size_t(index2));
      cur += 2; remain -= 2;
    }
    cout << polys.size() << ": ";
    for (i=0;i<p.indices.size();i++) cout << p.indices[i] << " ";
    cout << endl;
    polys.push_back(p);
  }

  cout << "Num of polys read: " << polys.size() << endl;

  return (true);
}

// SURF { name[S0], source[S0], attributes[SUB-CHUNK] * }

// Describes the shading attributes of a surface. The name uniquely
// identifies the surface. This is the string that's stored in TAGS
// and referenced by tag index in PTAG. If the source name is
// non-null, then this surface is derived from, or composed with, the
// source surface. The base attributes of the source surface can be
// overridden by this surface, and texture blocks can be added to the
// source surface. The material attributes follow as a variable list
// of subchunks documented below in the Surface Subchunks section.

// Does length include the 4 byte header?



// SUB-CHUNK ::= tag[ID4], length[U2],  data[...], pad[U1] ?
// Buf points to beginning of tag.
// bufLen includes data + 4 byte tag + 2 byte size field
SurfaceSubChunk::SurfaceSubChunk (const char *buf, const size_t buflen) {
  assert (buf);
  assert (buflen>4);
  //printBuf (buf,(10<len)?10:buflen);

  type = getSurfEnum (buf,buflen);
  start = const_cast<char *>(buf);
  {
    string s(buf,4);
    cout << "SurfaceSubChunk start: " << s << endl;
  }
  len = getU2 (buf+4, buflen-4);
  cout << "SurfaceSubChunk::SurfaceSubChunk len: "  << len << endl;
}

void SurfaceSubChunk::print () {
  cout << "  SurfaceSubChunk: " << endl
       << "    type " << type << endl
       << "    start " << start << endl
       << "    len " << len << endl
    ;
}


// BLOK { header[SUB-CHUNK],  attributes[SUB-CHUNK] * }
// block-header { ordinal[S0], block-attributes[SUB-CHUNK] * }
// SUB-CHUNK ::= tag[ID4], length[U2],  data[...], pad[U1] ?
bool
GetTagBLOK (const char *buf, const size_t bufSize) {
  assert (buf);
  assert (bufSize>8);


  // Skip the BLOK (4 bytes) and BLOKsize (2 bytes)
  char *cur = const_cast<char *>(buf) + 6;
  size_t curLen = bufSize - 6;

  printBuf (buf,15);
  //
  // BLOCK HEADER
  //
  BlockEnum blockType = GetBlockEnum(cur,curLen);
  cout << "found block type of: " << BlockEnumString(blockType) << endl;
#if 1
  cur+=4; curLen -=4;
  // Is this the ordinal?
  U4 blockSize = getU4 (cur, curLen);
  cout << "block size: " << blockSize << endl;
  string ordinal = getS0(cur,curLen);
  cout << "AKA?? ordinal " << ordinal << endl;
  cur+=4; curLen -=4;
#else
  string ordinal = getS0 (cur,curLen);
  {
    const size_t jump = getS0Len(ordinal);
    cur+=jump; curLen +=jump;
  }
#endif

  BlockHeader header;

  bool done = false;
  while (!done && curLen>0) {
    cout << "Parsing next Blok header" << endl;
    printBuf (cur,4);
    const BlockEnum b = GetBlockEnum(cur,curLen);
    size_t jump;
    switch (b) {
    case CHAN:
      // CHAN { texture-channel[ID4] }
      {
	U2 unknown = getU2(cur+4,curLen-4); // FIX: PROBABLY SIZE
	string id = getID4 (cur+6,curLen-6);
	cout << "  CHAN... unknown U2: " << unknown << "  id: " << id << endl;
	header.CHAN_unknown = unknown;
	header.CHAN_id = id;
      }
      jump = 4 + 2 + 4;
      break;
    case ENAB:
      // ENAB { enable[U2] }
      // default = enabled
      {
	U2 enable = getU2 (cur+4,curLen-4);
	cout << "  ENAB: " << (enable?"yes":"no") << endl;
	printBuf (cur+4,4);
	header.enabled = (enable?true:false);
	//jump = 4 + 2; should be U2???
	jump = 4 + 4; // FIX: but the spec says U2???
      }
      break;
    case OPAC:
      // OPAC { type[U2], opacity[FP4],  envelope[VX] }
      // FP4 is justs a 4 byte float in range [0..4]
      //cur
      {
	//printBuf(cur+4,8);
	U2 type = getU2(cur+4,curLen-4);
	// FIX: why does TriangleWithTexture.lwo have a type of 8?!?! out of range
	F4 opacity = getF4(cur+4+2,curLen-4-2);
	size_t vx_size=666;
	U4 envelope = getVX(cur+4+2+4, curLen-4-2-4, vx_size);
	cout << "  OPAC:  type="<<type
	     << "  opacity=" << opacity
	     << "  envelope(" << vx_size << ")=" << envelope << endl;
	cout << "FIX: why does the spec lie... vx_size should be 2 but is 4" << endl;
	vx_size=4; // FIX: Why??!?!?!

	header.OPAC_type = OpacityEnum(type);
	header.opacity = opacity;
	header.envelope = envelope;
	
	jump = 4 + 2 /*U2*/ + 4 /*FP4*/ + vx_size; 
      }
      break;
    case AXIS:
      {
	U2 axis = getU2(cur+4,curLen-4);
	assert (2>=axis);
	header.displacement_axis=AxisEnum(axis);
	jump = 4 + 4; // FIX: put in correct jump... why 4 for a U2????
      }
      break;

    default:
      cerr << "ERROR: Un-recognized block enum!!!" << endl;
      //jump += 4;
      done = true;
    }
    if (!done) {
      assert (jump <= curLen);
      cur+=jump;curLen -= jump;
    }
  }
  header.print();
  // Done with header.

  cout << "Done with header............." << endl;

  done=false;
  while (!done && curLen>0) {
    cout << "Parsing next Blok header" << endl;
    printBuf (cur,4);
    const BlockEnum b = GetBlockEnum(cur,curLen);
    cout << " found " << BlockEnumString(b) << endl;
    size_t jump;
    switch (b) {
    case NEGA:
#if 0
      {
	bool nega = getU2Bool(cur+4,curLen-4);
	cout << " NEGA: " << (nega?"yes":"no") << endl;
	jump=4+4;
      }
#endif
      break;
    default:
      assert(false);
    }
    if (!done) {
      assert (jump <= curLen);
      cur+=jump;curLen -= jump;
    }

  }  


  return (false);
}

// SUB-CHUNK ::= tag[ID4], length[U2],  data[...], pad[U1] ?
Surface
GetTagSURF(const char *buf,size_t bufLen, const U4 tagLength) {
  cerr << "FIX: GetTagSURF not yet implemented" << endl;

  Surface s;
  s.start = const_cast<char *>(buf);
  s.len = tagLength;
  char *cur=s.start+8;   size_t curLen=tagLength;

  s.name = getS0 (cur,curLen);
  cur += getS0Len(s.name); curLen -= getS0Len(s.name);
  cout << "SURF name: " << s.name << " s.name.size():" << s.name.size()
       << "  size: " << getS0Len(s.name) << endl;

  s.source = getS0 (cur,curLen);
  cur += getS0Len(s.source); curLen -= getS0Len(s.source);
  //cout << "SURF source: '" << s.source << "'  size: " << getS0Len(s.source) << endl;

  // FIX: how better to detect overrun?
  while (curLen>0 && curLen < 9999999) {
    cout << "  --> Working on next SUB-CHUNK  curLen=" << curLen  << endl;
    // SUB-CHUNK ::= tag[ID4], length[U2],  data[...], pad[U1] ?
    SurfaceSubChunk ssc(cur,curLen);
    //cout << "SubChunk size: " << ssc.len << endl;
    size_t jump = ssc.len + 4 +2; // 4 for TAG, 2 for size (U2)
    //cout << "JUMP INFO: " << int(cur) << " " << curLen << " " << jump << endl;
    cur += jump; curLen -= jump;
    //cout << "JUMP INFO: " << int(cur) << " " << curLen << " " << jump << endl;
    s.subchunks.push_back(ssc);
    //cout << "Adding chunk: " << s.subchunks.size()-1 << endl;
    //ssc.print();

    if (BLOK==ssc.type) {
      GetTagBLOK (ssc.start, ssc.len);
      // FIX: do something with the blok
    }

  }
  cout << "  --> DONE with SUB-CHUNK  curLen=" << curLen  << endl;


  return (s);
}




// ----------------------------------------------------------------------
bool WriteInventor(const string filename, vector<coordinate> &pts, vector<Polygon> polys) {
  bool ok=true;

  std::ofstream o(filename.c_str(), std::ios::out);
  if (!o.is_open()) {
    std::cerr << "ERROR: output file open failed for " << filename << endl;
    return (false);
  }

  o << "#Inventor V2.1 ascii\n"
    << "\n"
    << "# xc-lightwave to OpenInventor\n"
    << "# \n"
    << "# http://schwehr.org/xcore\n"
    << "\n";

  o << "  ShapeHints {\n"
    << "     vertexOrdering COUNTERCLOCKWISE\n"
    << "     creaseAngle 1.57\n"
    << "  }\n";

  //
  // Setup output stream floating point writing.  Want best precision.
  // 
  o.setf (std::ios_base::fixed,std::ios_base::floatfield);
  const int precision=12;  // FIX: what is really a good number for this?
  o.precision (precision);

  ////////////////////////////////////////
  // COORDINATES
  ////////////////////////////////////////
  {
    //o.precision (precision); // Reset to higher precission
    o << "  Coordinate3 { point [\n";
    //float x,y,z;
    size_t i;
    for (i=0;i<pts.size();i++) {
      if (isnan(pts[i].X) || isnan(pts[i].Y) || isnan(pts[i].Z)) {
	pts[i].X = pts[i].Y = pts[i].Z = 0.0;
	static bool warned=false;
	if (!warned) {
	  warned=true;
	  cout << "WARNING: found coordinates with NAN values! Shutting up now" << endl;
	}
      }
      o << "      " << pts[i].X << " " << pts[i].Y << " " << pts[i].Z << ",\n";
    }
    o << "    ] # point\n"
      << "  } # Coord3\n\n";
  } // Coords

  ////////////////////////////////////////
  // INDEX TO COORDS
  ////////////////////////////////////////
  {
    o << "  IndexedFaceSet { coordIndex [\n";
    const size_t numPoly = polys.size();
    o << "    # Number of polys: "<< polys.size()<<"\n";

    size_t p;
    for (p=0;p<numPoly;p++) {
      o << "    ";
      size_t i;
      for (i=0;i<polys[p].indices.size();i++) {
	o << polys[p].indices[i] << ",";
	  
      }
      o << "-1,\n";
    }

    o << "    ] # coordIndex\n"
      << "  } # IndexedFaceSet\n";
  } // Coord indexes


  o.close();
  // FIX: error check close
  return (ok);
}



//////////////////////////////////////////////////////////////////////
// FIRST TRY AT READING LIGHTWAVE
//////////////////////////////////////////////////////////////////////

bool
readLightWave(const string filename) {
  bool ok=true;
  cout << "Opening " << filename << endl;

  struct stat sb;
  int r = stat (filename.c_str(), &sb);
  if (0 != r) {
    perror("stat to get file size FAILED");
    return (false);
  }
  cout << "File size: " << sb.st_size << endl;

  int fd = open (filename.c_str(), O_RDONLY, 0);
  if (-1==fd) {
    perror ("open failed");
    return (false);
  }
  char *file = (char *)mmap (0, sb.st_size, PROT_READ,  MAP_FILE, fd, 0);
  if ((char *)(-1)==file) {
    perror ("mmap FAILED");
    return (false);
  }
  close(fd); // Close does not munmap
  
  if (0!=strncasecmp("FORM",file,4)) {
    cerr << "File does not start with 'form'" << endl;
    ok=false;
    munmap(file,sb.st_size);
    return (false);
  }

  {
    //const U4 *length = (U4 *)(file+4);
    //cout << *length << endl;
    const U4 length = getU4(file+4,sb.st_size-4);
    cout << length << endl;
    if (length + 8 != sb.st_size) {
      cerr << "ERROR: Lightwave file size incorrect!" << endl
	   << "  " << length + 8 << " does not match file size: " << sb.st_size << endl;
      munmap(file,sb.st_size);
      return (false);
    }
  }


  string formType;
  {
    char formTypeChar[5];
    strncpy (formTypeChar,(file+8),4);
    formTypeChar[4]='\0';
    //string formType(formTypeChar);
    formType = string(formTypeChar);
    cout << "Form Type: " << formType << endl;
    if (string::npos == formType.find("LWO2")) {
      cerr << "ERROR: Expected LWO2 type" << endl;
      munmap(file,sb.st_size);
      return (false);
    }
  }

  //
  // Parse the chunks
  //

  string curName; // name of the current object
  layer curLayer;
  vector<coordinate> curPoints;
  bbox curBBOX;
  vector<Polygon> curPolygons; 

  char *cur = file + 12;
  while (cur-file < sb.st_size) {
    cout << "-----------------------------------------" << endl;
    U4 len;
    const TagsEnum t = GetTagInfo (cur, cur-file, len);
    cout << TagString (t) << " " << len << endl;
    if (UNKNOWN_TAG==t) {
      cerr << "ERROR: Unknown tag!" << endl;
      ok=false;
      break;
    }
    switch (t) {
    case LAYR: 
      //cerr << "FIX:  Implement " << "LAYR" << endl;
      curLayer = GetTagLAYR(cur,sb.st_size - (cur-file),len);
      curLayer.print();
      break;
    case PNTS:
      // Point List
      GetTagPNTS(cur,sb.st_size - (cur-file),len, curPoints);
      break;
    case VMAP: cerr << "FIX:  Implement " << "VMAP" << endl; break;
    case POLS:
      {
	GetTagPOLS(cur,sb.st_size - (cur-file),len, curPolygons);
      }
      break;
    case TAGS: 
      curName = GetTagTAGS(cur,sb.st_size - (cur-file),len);
      cout << "CURRENT OBJECT: '" << curName << "'" << endl;
      break;
    case PTAG: 
      // Polygon Tag Mapping
      cout << "FIX:  Implement " << "PTAG" << endl; 
      break;
    case VMAD: cout << "FIX:  Implement " << "VMAD" << endl; break;
    case ENVL: cout << "FIX:  Implement " << "ENVL" << endl; break;
    case CLIP:
      cout << "FIX:  Implement " << "CLIP" << endl;
      //The CLIP chunk contains a series of subchunks describing a
      //single, possibly time-varying image. The first subchunk has to
      //be one of the source chunks: STIL, ISEQ, ANIM, XREF or STCC.
      {
	// There is a U4 after the block size which must be the number of items.
	const U4 numClips = getU4 (cur+8,sb.st_size - (cur-file) - 8);
	cout << "Num of CLIPs??: " << numClips << endl;
	if (1!=numClips) {
	  cerr << "ERROR: Expected numClips to be 1.  FIX: handle more clips" << endl;
	  exit(EXIT_FAILURE);
	}

	ClipType type = GetClipType(cur,sb.st_size - (cur-file),len);
	cout << "Found clip type: " << ClipTypeString(type) << endl;
	if (UNKNOWN_CLIP_TYPE==type) {
	  cout << "Found unknown clip type! Skipping." << endl;
	  break;
	}

	switch (type) {
	case STIL:
	  {
	    string name = getFNAM0(cur+8+8,len-8); // +4 skips STIL
	    cout << "Got name: '" << name << "'" << "  FIX: save the name"<< endl;
	  }
	  break;
	case ISEQ: cout << "FIX: implement ISEQ handler" << endl; break;
	case ANIM: cout << "FIX: implement ANIM handler" << endl; break;
	case XREF: cout << "FIX: implement XREF handler" << endl; break;
	case STCC: cout << "FIX: implement STCC handler" << endl; break;
	case UNKNOWN_CLIP_TYPE: cout << "FIX: Handle UNKNOWN_CLIP_TYPE" << endl; break;
	default:
	  cerr << "ERROR: bad clip type in ClipTypeString!!!" << endl;
	}

      }
      break;
    case SURF: 
      cout << "FIX:  Implement " << "SURF" << endl; 
      GetTagSURF(cur,sb.st_size - (cur-file),len);
      break;
    case BBOX:

      GetTagBBOX(cur,sb.st_size - (cur-file),len,curBBOX);
      break;
    case DESC: cout << "FIX:  Implement " << "DESC" << endl; break;
    case TEXT: cout << "FIX:  Implement " << "TEXT" << endl; break;
    case UNKNOWN_TAG: cerr << "FIX:  Implement " << "UNKNOWN_TAG" << endl; break;
    default:
      cerr << "TagString ERROR: Bad TAG!" << endl;
      ok=false;
    }
    if (!ok) break;
    cur += 4 + 4 + len;  // 4 byte TAG + 4 byte tag size + tag size
  }
  cout << "Done: prepairing to unmap"  << endl;

// bomb: // Jump here when something went wrong.  Unmap and exit
  r = munmap(file,sb.st_size);
  if (-1==r) {
    perror ("munmap FAILED");
    return(false);
  }

  if(!WriteInventor("test.iv",curPoints, curPolygons)) {
    cerr << "ERROR: WriteInventor FAILED!" << endl;
    ok = false;
  }

  return (ok);
}


//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////


#ifdef REGRESSION_TEST
int main (int argc, char *argv[]) {
  bool ok=true;

#if 0
  if (2!=argc) {
    cerr << "Need to specify a file to read" << endl;
    exit (EXIT_FAILURE);
  }
  string filename(argv[1]);
  if (!readLightWave (filename)) ok=false;
#else
  cerr << "Add real test case for lightwave" << endl;
#endif

  std::cout << argv[0] << (ok?": ok.":": FAILED!") << std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif // REGRESSION_TEST
