/* $Revision: 1.5 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGridWriteObj.C
 *
 * DESCRIPTION: Use XcGrid to write an Obj formatted geometry file.
 *
 * EXPORTS:  
 *
 * NOTES:        
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

#include <iostream>
#include <fstream>
using namespace std;

#include "XcGrid.H"
#include "xc-obj.H"


/***************************************************************************
 * FUNCTION: FilenameToObjName
 * DESCRIPTION: Make a name compliant with object_name
 * INPUTS:  string to rectify
 *	    
 * OUTPUTS: 
 * RETURNS: corrected string
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   I am guessing that the name is just like group_name in that it
 * 	can only be letters or numbers.
 ***************************************************************************/

const string stripObjExt(const string filename) {
  string s(filename);
  const string suf1=".obj";  // suffix
  const string suf2=".OBJ";  // suffix
  const string suf3=".Obj";  // suffix
  size_t p;
  p=s.rfind(suf1); if ( (s.length()-suf1.length()) == p) s.erase(p); return(s);
  p=s.rfind(suf2); if ( (s.length()-suf2.length()) == p) s.erase(p); return(s);
  p=s.rfind(suf3); if ( (s.length()-suf3.length()) == p) s.erase(p);
  return (s);
}

// FIX: need a better translation function!
const string FilenameToObjName(const string filename) {
  string s(filename);
  size_t p;

  s = stripObjExt(s);

  for (p=s.find(".");p!=string::npos; p=s.find(".")) {
    s.erase(p,1);
  }
  return (s);
}

/***************************************************************************
 * FUNCTION: XcGridWriteObj
 * DESCRIPTION: Create a basic Obj Ascii file
 * INPUTS:  
 *	    
 * OUTPUTS: 
 * RETURNS: true on success, false on failure
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   http://astronomy.swin.edu.au/~pbourke/geomformats/obj/
 ***************************************************************************/

// This specifies the texture file for the object
bool XcGridWriteObjMtl (const std::string filename, 
			const std::string mtlname,
			const std::string texturefilename)
{
  //cout << " mtl: " << filename << " " << texturefilename << endl;
  std::ofstream o(filename.c_str(),std::ios::out);
  if (!o.is_open()) {
    std::cerr << " output file open failed for "<< filename << std::endl;
    return (false);
  }
  o << "#" << endl
    << "# Wavefront OBJ file" << endl
    << "# Created by xcore terrain " << __FILE__ << " $Id: xc-obj.C,v 1.5 2004/08/26 18:37:40 schwehr Exp $" << endl
    << "# http://schwehr.org" << endl
    << "#" << endl
    << endl << endl;

  o << "newmtl " << mtlname << endl
    << "Ka 0 0 0" << endl
    << "Kd 1 1 1" << endl
    << "Ks 0.2 0.2 0.2" << endl
    << "illum 2" << endl
    << "Ns 25.6" << endl
    << "map_Kd " << texturefilename << endl;

  o.close();
  return (true);
}

bool
XcGridWriteObj (const std::string filename, XcGrid &grd, bool texture) {
  assert(texture);  // no way to turn of right now
  std::ofstream o(filename.c_str(),std::ios::out);
  if (!o.is_open()) {
    std::cerr << " output file open failed for "<< filename << std::endl;
    return (false);
  }
  const size_t nelements=grd.getWidth()*grd.getHeight();
  size_t i;
  
  // FIX: is there a magic number?
  o << "#" << endl
    << "# Wavefront OBJ file" << endl
    << "# Created by xcore terrain " << __FILE__ << " $Id: xc-obj.C,v 1.5 2004/08/26 18:37:40 schwehr Exp $" << endl
    << "# http://schwehr.org" << endl
    << "#" << endl
    << endl << endl;

  // FIX: add date, time, username, pwd, and hostname to the comments

  const string mtl(FilenameToObjName(filename)+"mtl");
  const string mtlFilename (stripObjExt(filename)+".mtl");

  if (!XcGridWriteObjMtl(mtlFilename, mtl, grd.getTextureFile())) {
    cerr << "ERROR:  mtl failed to write" << endl;
    return (false);
  }

  // object_name
  o << "o " << FilenameToObjName(filename) << "o" << endl;  // Does not seem to be used
  o << "g " << FilenameToObjName(filename) << endl;  // Maya uses this as the obj name

  o << "mtllib " << mtlFilename << endl;
  o << "usemtl " << mtl << endl;

  for (i=0;i<nelements;i++) {
    float x,y,z; grd.getXYZ(i,x,y,z);
    o << "v " << x << " " << y << " " << z << endl;
  }
  for (i=0;i<nelements;i++) {
    size_t xoff,yoff;
    grd.getXYfromIndex(i,xoff,yoff);
    float x,y,z; grd.getNormal(xoff,yoff,x,y,z);
    o << "vn " << x << " " << y << " " << z << endl;
  }

  for (i=0;i<nelements;i++) {
    size_t xoff,yoff;
    grd.getXYfromIndex(i,xoff,yoff);
    float u,v; grd.getUV(xoff,yoff,u,v);
    o << "vt " << u << " " << v << endl;
  }
  {
    size_t i;
    const size_t numTri = grd.getNumTriangles();
    for (i=0; i<numTri; i++) {
      XcTriangle t = grd.getTriangle(i);
      size_t v0 = t.vertices[0] + 1;
      size_t v1 = t.vertices[1] + 1;
      size_t v2 = t.vertices[2] + 1;
      o << "f " 
	<< v0 << "/" << v0 << "/" << v0 << " "
	<< v1 << "/" << v1 << "/" << v1 << " "
	<< v2 << "/" << v2 << "/" << v2 << endl;
    }
  } // Coord indexes

  return (true);
}

//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////

#include "XcGridDrv.H"
using namespace XcG;

std::string XcGridDrvObj::getName() { return("OBJ");}
std::vector<FileType> XcGridDrvObj::getReadTypes() {
  std::vector<FileType> types;
  //types.push_back(OBJ);
  return (types);
}
std::vector<FileType> XcGridDrvObj::getWriteTypes() {
  std::vector<FileType> types;
  types.push_back(OBJ);
  return (types);
}

bool XcGridDrvObj::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  assert (OBJ==t);
  return (false); // Can't write nuthin.
}
bool XcGridDrvObj::WriteGrid(std::string filename, FileType t, XcGrid &grd, gengetopt_args_info &opts) {
  assert (OBJ==t);
  std::cerr << "FIX: handle texture\n";

  grd.genCompressedIndex();
  grd.genTriangles();
  if (opts.normals_flag || opts.spikes_flag) {
    grd.genNormals();
  }
  grd.genUV();

  bool useTexture=false;
  if (opts.tex_given) {
    useTexture=true;
    if (grd.getTextureFile().size() > 0) {
      std::cout << "Warning: texture file name is over writing previous value: " << std::endl
		<< "  Was: " << grd.getTextureFile() << std::endl
		<< "  Now: " << opts.tex_arg << std::endl;
    }
    grd.setTextureFile(opts.tex_arg); // May be overwriting a desired value.
  }

  std::cout << "  Writing inventor file " << filename << ":" << std::endl
	    << "    useTexture:       " << (useTexture?"true":"false") << std::endl
	    << "    textureFile:      " << grd.getTextureFile() << std::endl
	    << "    expliciteNormals: " << (opts.normals_flag?"true":"false") << std::endl
	    << "    colorTable:       " << (opts.cpt_given?"true":"false") << std::endl
	    << "    normalsSpikes:    " << (opts.spikes_flag?"true":"false") << std::endl;


#if 0
  bool r = XcGridWriteObj(filename, grd);
#else
  cerr << "Write the damn driver!!!" << endl;
  bool r=false;
#endif
  return (r);
}


//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST

#include "XcGridSamples.H"

// Have to provide out own register command so we don't depend on
// other drivers... Only register ourself.
XcG::VecDriverPtr XcG::XcGridDrvRegister() {
  VecDriverPtr drivers;
  drivers.push_back(new XcGridDrvObj);	// Obj Alias Wavefront ASCII
  return(drivers);
}


int main (int argc, char *argv[]) {
  bool ok=true;

  // Drv Test
  {
    VecDriverPtr drivers = XcGridDrvRegister();  // vector of drivers
    bool r = (1==drivers.size());
    if (!r) ok=false;
    std::cout << "  XcGridDrvRegister test: " << (r?"ok.":": FAILED!") << std::endl;

    // FIX: what is the not found for vector, it's NOT npos??
    if (r) {
      size_t drv;
      
      drv = XcGridDrvGetWriter(drivers,OBJ);
      if (std::string::npos==drv) { cerr << "  ERROR: driver not found!" << endl; r=ok=false;}
      else if (std::string::npos==drv || drivers[drv]->canReadType(OBJ)) {
	ok=r=false; cerr << "  ERROR: Driver can NOT read files yet, but reports that it can." << endl;
      }
      else if (std::string::npos==drv || !drivers[drv]->canWriteType(OBJ)) {
	ok=r=false; cerr << "  Driver is supposed to be able to write" << endl;
      }
      std::cout << "  XcGridDrvGetWriter test: " << (r?"ok.":": FAILED!") << std::endl;
    }
  } // Drv Test

  {
    XcGrid *g = XcGridSample3x3(true);
    g->setTextureFile("sample1.jpg");
    XcGridWriteObj(string("sample1.obj"),*g);
    delete g;
  }


  std::cout << argv[0] << (ok?": ok.":": FAILED!")<<std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
} // main


#endif // REGRESSION_TEST

