%{  /* -*- c++ -*-  */  // Emacs thinks this is a lisp file otherwise

/* $Revision: 1.13 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    PdsLabel.l
 *
 * DESCRIPTION: Flex parser of vicar labels.
 *
 * EXPORTS:  
 *
 * NOTES:        
 *
 *   - Derived from Terrain/PdsLabel.l
 *   - NOT thread safe!!
 *   - FIX: what to do if different objects have same keywords/labels?
 * 
 * SEE ALSO:
 * 
 *	Specification: http://pds.jpl.nasa.gov/documents/sr/index.html
 *
 *     PDSLabel.H
 *     PdsLabel.H
 *     XcPdsGrid.C which uses this parser
 *
 *     STL tutorial including maps: http://www.yrl.co.uk/~phil/stl/stl.htmlx
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

/***************************************************************************
 * INCLUDES
 ***************************************************************************/

#include <cassert>

#include <iostream>
#include "xc-pdsLabel.H"

/***************************************************************************
 * GLOBALS
 ***************************************************************************/

  // These get cleared in LoadPdsLabelsFromFile
  static PdsLabelMap Labels;
  std::vector<std::string> Objects;  // All objects in order of appearance



  extern "C" int pdswrap (void ){
    //std::cout << "pdswrap capped\n";
    // Do nothing
    return (1);
  }

  extern int pdslex(); // Generated by flex>

  int yylex_pds (FILE *inFile)
  {
    int result;
    yyin = inFile;
    result=pdslex();
    return result;  // 0 on success
  }

  void PrintLabels (PdsLabelMap &Labels) {
    std::cout << "   **** Printing label map ****\n"
	      <<"Num entries: " << Labels.size() << "\n";
    PdsLabelMap::iterator i;
    for (i=Labels.begin(); i != Labels.end();i++) {
      std::cout << "\nNext keyword is: >" << (*i).first << "<";
      PdsLabel l = (*i).second;
      l.print();
    }

  }


  PdsLabel::PdsLabel () {
    //std::cout << "default pdslabel constructor\n";
    label="Unset";
    labelType = TYPE_UNKNOWN;
    intVal = -666;
    realVal = -66.6;
    stringVal = "no string";
    
    inObject=false;
    objNum=-1;
    objName = NOT_IN_OBJECT;
  }


  // Fill in the data structure with a string "name=value" pair
  // objName is for if we are inside some PDS object block
  PdsLabel::PdsLabel (type t, std::string labelString,
		      bool _inObj, int _objNum, std::string _objName) {
    
    inObject = _inObj;
    objNum = _objNum;
    objName = _objName;

    label = labelString;
    labelType = t;

    labelType = TYPE_UNKNOWN;
    intVal = -666;
    realVal = -66.6;
    // Leave string empty

    switch (t) {
    case POINTER:
      labelType = POINTER;
      intVal = atoi(getLabelValStr().c_str());
      break;
    case INT:
      labelType = INT;
      intVal = atoi(getLabelValStr().c_str());
      break;
    case REAL:
      labelType = REAL;
      realVal = atof(getLabelValStr().c_str());
      break;
    case STRING: 
      {
	labelType = STRING;
	std::string s = getLabelValStr();

	if (-1 == int(s.find('"'))) {
	  // String without quotes
	  // FIX: this maybe should be some sort of enum type?
	  //std::cout << "found an enum like thingy\n";
	  int start = s.find_first_not_of(" \t");
	  stringVal = s.substr(start,s.length());
	  //std::cout << " Setting enumy thing to " << stringVal << std::endl;
	} else {
	  // String with quotes
	  size_t start = s.find('"');
	  size_t end = s.rfind('"');
	  if (start>s.size() || end>s.size()) {
	    stringVal = "ERROR READING STRING\n";
	    std::cerr << "Error reading string: '"<<s<<"'\n";
	  }
	  stringVal = s.substr(start+1,end-2); // Drop the "" around the label value
	  //stringVal = s.substr(2,s.length()-3); // Drop the '' around the label value
	}
	break;
      }
    case INTARRAY:
      {
	labelType = INTARRAY;
	std::string s = getLabelValStr();
	size_t start = s.find('(');
	size_t end = s.find(')');
	s = s.substr(start+1,end-2);
	//std::cout << "INTARRAY: " << s << "\n";
	while (s.length()>0) {
	  int v = atoi(s.c_str());
	  intArray.push_back(v);
	  //std::cout << s<< "  ---- " << v << "\n";
	  int comma = s.find(',');
	  if (-1==comma) break;
	  //std::cout << "comma " << comma << "\n";
	  s = s.substr(comma+1,end);
	}
	//assert(false);
      }
      break;
    case REALARRAY:
      {
	labelType = REALARRAY;
	std::string s = getLabelValStr();
	size_t start = s.find('(');
	size_t end = s.find(')');
	s = s.substr(start+1,end-2);
	//std::cout << "REALARRAY: " << s << "\n";
	while (s.length()>0) {
	  float v = atof(s.c_str());
	  //std::cout << v <<"\n";
	  realArray.push_back(v);
	  int comma = s.find(',');
	  if (-1==comma) break;
	  s = s.substr(comma+1,end);
	}
      }
      break;
    case STRARRAY:
      {
	labelType = STRARRAY;
	std::string s = getLabelValStr();
	size_t start = s.find('(');
	size_t end = s.find(')');
	s = s.substr(start+1,end-2);
	//std::cout << "STRINGARRAY: " << s << "\n";
	while (s.length()>0) {
	  size_t startQuote = s.find('"');
	  size_t endQuote = s.find('"',startQuote+1);
	  std::string nextStr = s.substr(startQuote+1,endQuote-1);
	  //std::cout << "next str: " << nextStr << "\n";
	  stringArray.push_back(nextStr);
	  int comma = s.find(',');
	  if (-1==comma) break;
	  s = s.substr(comma+1,end);
	}
      }


      break;
    default:
      std::cerr << "ERROR: Unknown label type!!\n";
      assert(false);
    }
    assert (TYPE_UNKNOWN != labelType);
  } // PdsLabel::PdsLabel


  // Copy constructor
  PdsLabel::PdsLabel(const PdsLabel &oldpc) {
    //std::cout << "Copy constr\n";
    inObject   = oldpc.inObject;
    objNum  = oldpc.objNum;
    objName = oldpc.objName;

    label = oldpc.label;
    labelType = oldpc.labelType;
    intVal = oldpc.intVal;
    realVal = oldpc.realVal;
    stringVal = oldpc.stringVal;
    intArray = oldpc.intArray;
    realArray = oldpc.realArray;
    stringArray = oldpc.stringArray;
  }

  // Overload = so it works in a map
  void PdsLabel::operator= (const PdsLabel &oldpc) {
    //std::cout << "operator=\n";
    inObject   = oldpc.inObject;
    objNum  = oldpc.objNum;
    objName = oldpc.objName;

    label = oldpc.label;
    labelType = oldpc.labelType;
    intVal = oldpc.intVal;
    realVal = oldpc.realVal;
    stringVal = oldpc.stringVal;
    intArray = oldpc.intArray;
    realArray = oldpc.realArray;
    stringArray = oldpc.stringArray;
  }

  // Return the label name before the '='
  // Adds the object name in front if in object.

  // NOTE: '-' breaks an object name from the label name

  std::string PdsLabel::getLabelName() {
    size_t eq;
    eq=label.find("=");
    std::string s = label.substr(0,eq-1); // FIX:  deal with white space in a safer way

    size_t end = s.find_last_not_of(" \t");
    s = s.substr(0,end+1);
    //std::cout << ">" << s << "<\n";
    if (inObject) return (objName+"-"+s);
    return (s);
  }

  // return what is after the '=' in a label
  std::string PdsLabel::getLabelValStr() {
    size_t eq;
    eq=label.find("=");
    std::string s = label.substr(eq+1,label.length());
    return (s);
  }

  int PdsLabel::getIntVal() {
    if(INT!=labelType && POINTER!=labelType) {
      std::cerr << "bad access.  Not INT! "<< label << "\n";
      print();
      assert(false);
    }
    return(intVal);
  }
  float PdsLabel::getRealVal() {
    if(REAL!=labelType) {
      std::cerr << "ERROR: bad access.  Not REAL!\n";
      print();
      assert(false);
    }
    return(realVal);
  }
  std::string PdsLabel::getStringVal() {
    if(STRING!=labelType) {
      std::cerr << "ERROR: bad access.  Not STRING!\n";
      print();
      assert(false);
    }
    return(stringVal);
  }

  std::string PdsLabel::getWholeLabelStr() {
    return(label);
  }


  // debugging
  void PdsLabel::print() {
    std::cout << "\nLABEL: '" << label << "' ...\n";
    if (inObject) {
      std::cout << "  In PDS Object ("<<objNum<<"): "<<objName<<"\n"; 
    }

    //std::cout << " labelType ";
    switch (labelType) {
    case POINTER:
      std::cout << "  POINTER: " << intVal << "\n";
      break;
    case INT:
      std::cout << "  INT: " << intVal << "\n";
      break;
    case REAL:
      std::cout << "  REAL: " << realVal << "\n";
      break;
    case STRING: 
      std::cout << "  STRING: " << stringVal << "\n";
     break;
    case INTARRAY:
      {
	std::cout << "  INTARRAY: ";
	size_t i;
	for (i=0;i<intArray.size();i++) {
	  std::cout << intArray[i] << " ";
	}
	std::cout << "\n";
      }
      break;
    case REALARRAY:
      {
	std::cout << "  REALARRAY: ";
	size_t i;
	for (i=0;i<realArray.size();i++) {
	  std::cout << realArray[i] << " ";
	}
	std::cout << "\n";
      }
      break;
    case STRARRAY:
      {
	std::cout << "  STRINGARRAY: ";
	size_t i;
	for (i=0;i<stringArray.size();i++) {
	  std::cout << stringArray[i] << " ";
	}
	std::cout << "\n";
      }
      break;
    default:
      std::cout << "????\n";
      std::cerr << "ERROR: Unknown label type!!\n";
    }
  }





%}

NUL		\0
WS		([ \t])
NL		\n
LETTER		([a-zA-Z])
DIGIT		([0-9])
SCALEFACTOR	(("e"|"E"|"D")("+"|"-")?{DIGIT}+)

/* Variables */
KEYLETTER	([A-Z_])
KEYWORD		(\^?{KEYLETTER}({KEYLETTER}|{DIGIT}){0,31})

STRLETTER	([^\'[:cntrl:]])

EQL		({WS}*"="{WS}*)
PDSLABEL	({KEYWORD}{EQL})

INTEGER		(("+"|"-")?{DIGIT}+)
REAL		(("+"|"-")?{DIGIT}+"."{DIGIT}*{SCALEFACTOR}?)
STRING		(\"{STRLETTER}*\")

LEFT		\(
RIGHT		\)
INTARRAY	({LEFT}({INTEGER},)+{INTEGER}{RIGHT})
REALARRAY	({LEFT}({REAL},)+{REAL}{RIGHT})
STRARRAY	({LEFT}({STRING},)+{STRING}{RIGHT})

/* The above is basically the same as vicar. */
/* Below are PDS specific items */

UNITS		("<"{STRLETTER}*">")
END		("END")
OBJECT		("OBJECT"{EQL}{KEYWORD})
END_OBJECT	("END_OBJECT"{EQL}{KEYWORD})

COMMENT		("/*"{STRLETTER}*"*/")

%%

%{ /* Can place C code here */

  // FIX: need to support the bit width numbers like 
  // SAMPLE_BIT_MASK = 2#11111111#
  // CHECKSUM = 16#FFB3DB1C#

  // FIX: can not handle objects inside objects
  int objNum=-1;  // Keep track of object numbers
  bool inObj=false;  // True when we are inside an object
  std::string objName=NOT_IN_OBJECT;

%}

{COMMENT} {
  // FIX: Do we want to keep comments in the log?
  //std::cout << "*** COMMENT: " << yytext << "\n";
}

{OBJECT} {
  // FIX: How to avoid name collisions between objs in the map?
  //std::cout << "*** OBJECT: " << yytext << "\n";
  objNum++;
  inObj=true;
  std::string s(yytext);
  int name = s.find('=');
  name+=2;  // Skip = and space
  objName = s.substr(name,s.length()-1);
  //std::cout << "Setting OBJECT name: >" << objName << "<\n";
  Objects.push_back(objName);
}

{END_OBJECT} {
  //std::cout << "*** END OBJECT: " << yytext << "\n";
  inObj=false;
  objName=NOT_IN_OBJECT;
}

"^"{PDSLABEL}{INTEGER}	{
  //std::cout << "*** POINTER: " << yytext << "\n";
  PdsLabel l(PdsLabel::POINTER,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
}


{PDSLABEL}{INTEGER}{UNITS}?	{
  //std::cout << "INT: " << yytext << "\n";
  PdsLabel l(PdsLabel::INT,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
}

{PDSLABEL}{REAL}{UNITS}?	{
  //std::cout << "REAL: " << yytext << "\n";
  PdsLabel l(PdsLabel::REAL,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
}

{PDSLABEL}{STRING}	{
  //std::cout << "STRING: " << yytext << "\n";
  PdsLabel l(PdsLabel::STRING,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
			}

{PDSLABEL}{KEYWORD}	{
  // Seems that quotes are not required for strings in PDS
  // FIX:  This really should be an enum.
  //std::cout << "STRING: " << yytext << "\n";
  PdsLabel l(PdsLabel::STRING,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
			}

{PDSLABEL}{INTARRAY}	{
  PdsLabel l(PdsLabel::INTARRAY,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
			}

{PDSLABEL}{REALARRAY}	{
  PdsLabel l(PdsLabel::REALARRAY,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
			}

{PDSLABEL}{STRARRAY}	{
  PdsLabel l(PdsLabel::STRARRAY,yytext,inObj,objNum,objName);
  Labels[l.getLabelName()] = l;
			}

{WS}*			{ //std::cout << "ws: '" << yytext << "'\n";
}
{NL}* 			{ //std::cout << "NL\n";
}
.			{ //std::cout << "what was this?: '" << yytext << "'\n";
} 

^{END}$	         	{
  //std::cout << "\n 1)END found: '" << yytext << "'\n";
  yyterminate();
}

^{END}.$	         	{ // FIX... Need a better solution!  This is for DOS Files
  //std::cout << "\n 2)END found: \n" << yytext << "\n";
  //std::cout << "Looks like we found a DOS formatted Pds file\n";
  yyterminate();
}

<<EOF>>         	{ std::cerr << "ERROR: unexpected EOF\n";yyterminate(); }

%%  /* C code section */


#include <fstream>
using namespace std; // be lazy

PdsLabelMap
LoadPdsLabelsFromFile(const std::string &filename, std::vector<std::string> &_objs) {
  Labels.clear();  // Nuke all entries currently in the global table
  Objects.clear();
  _objs.clear();

  FILE *inFile = fopen (filename.c_str(),"r");
  if (!inFile) {
    std::cerr << "Unable to open PDS file: " << filename << "\n";
    return (Labels);
  }

  int r;
  // Must use C IO functions to work with flex...
  r = yylex_pds(inFile);
  fclose (inFile);
  if (0!=r) {
    cerr << "ERROR: labels were not correctly read.\n"
	 << "  Continuing, but unpredictable results may follow.\n";
    return(Labels);
  }

  _objs = Objects;
  return (Labels);
}

//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////


#ifdef REGRESSION_TEST

int main (int argc, char *argv[]) {
  bool ok=true;

  {
    PdsLabelMap m;
    // FIX: make these actual tests.
    PdsLabel l(PdsLabel::INT, "FOO = 10",false,0,"None");
    PdsLabel ll;
    ll=l;
    m[l.getLabelName()] = l;

    //PrintLabels(m);
    PdsLabel l2;
    l2 = m["FOO"];
  }

  std::vector<std::string> files;
  files.push_back("TestData/2N126543636ETH0100P1505L0M1.IMG");
  // FIX: add more test files

  size_t i=0;
  for (i=0;i < files.size();i++) {
    int r; // result from flex

    // Must use C IO functions to work with flex...
    FILE *inFile = fopen(files[i].c_str(),"r");
    if (!inFile) {
      std::cerr << "Failed to open file: " << files[i] << "\n";
      r = -1;
    } else {
      r = yylex_pds(inFile);
      fclose (inFile);
    }

    std::cout << "  " << files[i] << ((0==r)?": ok.":": FAILED!") << endl;
    if (0!=r) ok=false;
    // FIX: put in some good tests

    PdsLabel l;
    l = Labels["^IMAGE"];
  } // for files

  std::cout << argv[0] << (ok?": ok.":": FAILED!")<<std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}

#endif // REGRESSION_TEST
