/* $Revision: 1.14 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    XcGridVicar.C
 *
 * DESCRIPTION: Fill in an XcGrid from a VICAR/PDS file.  This is a JPL/MIPL
 * 	image format
 *
 * EXPORTS:  
 *
 * NOTES:        
 *      http://www-mipl.jpl.nasa.gov/vicar/vic_file_fmt.html
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

/* 
   Reader development was supported by sample files from Steve Levoe.
   These files were created using JPL's vicar:

# FORMAT    Data format. Valid:                 "BYTE"
#          BYTE,HALF,FULL,                     
#          REAL,DOUB,COMP. 
# byte is 8 bit, half is 16 bit, full is 32
# real is a 32 bit ieee float 
#
# LINC      Line increment.                                      1.0       (1)
# SINC      Sample increment.                                    1.0       (1)
# BINC      Band increment.                                      1.0       (1)
# IVAL      Initial value.                                       0.0       (1)
# since I am using defaults the first value will be zero and increment by 1
# HOST='MAC-OSX'  INTFMT='HIGH'  REALFMT='IEEE'

#  MIPL linux
# $P2LIB /usr/local/vicar/ops/p2/lib/x86-linux 
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4x4byteLinux.vic nl=4 ns=4 nb=1 format=BYTE
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4x4byteLinux.vic nl=4 ns=4 nb=1 format=BYTE
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4x4halfLinux.vic nl=4 ns=4 nb=1 format=HALF
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4x4halfLinux.vic nl=4 ns=4 nb=1 format=HALF
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4xfullLinux.vic nl=4 ns=4 nb=1 format=FULL
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4x4fullLinux.vic nl=4 ns=4 nb=1 format=FULL
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4x4realLinux.vic nl=4 ns=4 nb=1 format=REAL
/usr/local/vicar/ops/p2/lib/x86-linux/gen out=gen4x4realLinux.vic nl=4 ns=4 nb=1 format=REAL

/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4x4byteMac.vic nl=4 ns=4 nb=1 format=BYTE
/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4x4byteMac.vic nl=4 ns=4 nb=1 format=BYTE
/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4x4halfMac.vic nl=4 ns=4 nb=1 format=HALF
/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4x4halfMac.vic nl=4 ns=4 nb=1 format=HALF
/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4xfullMac.vic nl=4 ns=4 nb=1 format=FULL
/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4x4fullMac.vic nl=4 ns=4 nb=1 format=FULL
/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4x4realMac.vic nl=4 ns=4 nb=1 format=REAL
/vicar/vicar3103/p2/lib/mac-osx/gen out=gen4x4realMac.vic nl=4 ns=4 nb=1 format=REAL

*/


/***************************************************************************
 * INCLUDES
 ***************************************************************************/

//
// C headers
//
#include <cassert>
#include <cstdlib>

//
// C++ headers
//
#include <iostream>
#include <fstream>
#include <string>

//
// Local headers
//
//#include "XcGrid.H" // Already included from XcVicarGrid.H
#include "xc-vicar.H"
#include "xc-vicarLabel.H"  // Handles the Vicar Labels


/***************************************************************************
 * FUNCTIONS AND DEFINITIONS
 ***************************************************************************/

enum dataType {BYTE, HALF, FULL, REAL, DOUB, COMP, UNKNOWN_DATA_TYPE};
enum endian {LITTLE,BIG,UNKNOWN_ENDIAN}; // aka LOW, HIGH

static dataType getDataType (VicLabelMap &labels) {
  std::string f = labels["FORMAT"].getStringVal();
  if ("BYTE" == f) return BYTE;
  if ("HALF" == f) return HALF;
  if ("FULL" == f) return FULL;
  if ("REAL" == f) return REAL;
  if ("DOUB" == f) return DOUB;
  if ("COMP" == f) return COMP;
  std::cerr << "ERROR: getDataType does not recognize this type..."
	    << f << "\n";
  return (UNKNOWN_DATA_TYPE);
}

static endian getDataEndian(VicLabelMap &labels) {
  std::string intfmt = labels["INTFMT"].getStringVal(); // FIX: what about BINTFMT??
  if ("LOW"==intfmt) return (LITTLE);
  if ("HIGH"==intfmt) return (BIG);
  std::cerr << "Encountered unknown endian type: " << intfmt << "\n";
  return (UNKNOWN_ENDIAN);
}

static endian getHostEndian(void) {
  int short s=1;
  char *b = (char *)&s;
  if (b[0]) return (LITTLE);
  return (BIG);
}


// Number of bytes per sample
static int getDataValueSize(const dataType t) {
  switch(t) {
  case BYTE:
    return(1);
  case HALF:
    return(2);
  case FULL:
    return(4);
  case REAL:
    return(4);
  case DOUB:
    return(8);
  case COMP:
    return(8); // FIX: is this right??
  case UNKNOWN_DATA_TYPE:
    std::cerr << "ERROR: asking for size on UNKNOWN_DATA_TYPE.  0 is what you get.\n";
    return(0);
  default:
    std::cerr << "ERROR: how the heck did we get here?\n";
  }
  return(-1);
}

static void fixEndian(short &s,const endian e, const endian he) {
  if (e == he) return;
  char *b = (char *)&s;
  char tmp;
  tmp=b[0];  b[0]=b[1]; b[1]=tmp;
}

static void fixEndian(int &i,const endian e, const endian he) {
  if (e == he) return;
  char *b = (char *)&i;
  char tmp[4];
  int x;
  for (x=0;x<4;x++) tmp[x]=b[x];
  for (x=0;x<4;x++) b[x]=tmp[3-x];
}

static void fixEndian(float &f,const endian e, const endian he) {
  if (e == he) return;
  char *b = (char *)&f;
  char tmp[4];
  int x;
  for (x=0;x<4;x++) tmp[x]=b[x];
  for (x=0;x<4;x++) b[x]=tmp[3-x];
}

// Number of bytes per sample
static float getFloatVal(char *ptr, const dataType t, const endian e) {
  static endian he = getHostEndian();
  switch(t) {
  case BYTE:
    {
      // No endianness to worry about.  Assum
      unsigned char *v = (unsigned char *)ptr;
      return(*v);
    }
  case HALF:
    {
      short s = *(short *)ptr;
      fixEndian(s,e,he);
      return(s);
    }
  case FULL:
    {
      int i = *(int *)ptr;
      fixEndian(i,e,he);
      return(i);
    }
  case REAL:
    {
      float f = *(float *)ptr;
      fixEndian(f,e,he);
      return(f);
    }
  case DOUB:
    std::cerr << "FIX: need to implement reading DOUB format data\n";
    assert(false);
  case COMP:
    std::cerr << "FIX: need to implement reading COMP format data\n";
    // FIX: what the heck do I do with complex?
    assert(false);
  case UNKNOWN_DATA_TYPE:
    std::cerr << "ERROR: asking for size on UNKNOWN_DATA_TYPE.  0 is what you get.\n";
    return(0);
  default:
    std::cerr << "ERROR: how the heck did we get here?\n";
  }
  return(-1);
}


using namespace std; // be lazy.  FIX:  just include ios?

// FIX: what to do about multiband data?
bool
LoadVicarDataFromFile(std::string filename,VicLabelMap &labels, XcGrid &grd) {
  const int width  = labels["NS"].getIntVal();
  const int height = labels["NL"].getIntVal();
  const dataType type = getDataType(labels);
  const int valueSize = getDataValueSize(type); // Size of one sample

  const int startOffset = labels["LBLSIZE"].getIntVal();
  const int dataSize = width * height * valueSize;

  //cout << "dataSize="<<dataSize<<" "<<width<<" "<<height<<" "<<valueSize<<"\n";

  std::ifstream in(filename.c_str(), ios::in|ios::binary|ios::ate);
  const size_t fileSize=in.tellg();
  //cout << "FileSize = " << fileSize << "\n";

#if 0
  if (int(fileSize) != (startOffset+dataSize)) {
    std::cerr << "Warning: looks like there are labels at the end.\n";
    // FIX: Handle labels at the end!
  }
#endif
  if (int(fileSize) <(startOffset+dataSize)) {
    std::cerr << "ERROR: file appears truncated!\n"
	      << "  " << fileSize << " " << startOffset << " " << dataSize<<"\n";
    return (false);
  }

  in.seekg(startOffset);
  if (in.fail()) {
    std::cerr << "Failed to seek to start of datablock\n";
    return(false);
  }

  char *buf = new char[dataSize+1];
  if (!buf) {
    std::cerr << "Unable to load buffer of "<<dataSize<<"byte.  Out of memory?\n";
    return(false);
  }
  in.read(buf,dataSize);
  if (in.fail()) {
    std::cerr << "ERROR: unable to read the data block\n";
    delete [] buf;
    return (false);
  }

  endian e = getDataEndian(labels);
  int x,y;
  for (y=0;y<height;y++) {
      for (x=0;x<width;x++) {
	const int i = y * width + x;
	char *ptr = buf + i * valueSize;
	float f;
	f = getFloatVal(ptr, type,  e);
	grd.setZ(x,y,f);
      }
  }
  
  delete [] buf;


  return(true);
}
/***************************************************************************
 * FUNCTION: AddPdsLabelsToGridLog
 * DESCRIPTION: Put all the labels in the grid log array of strings
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   Only handles straight mxn DEM files.  NO RANGE MAPS!
 ***************************************************************************/
static void AddVicarLabelsToGridLog(XcGrid &grd,VicLabelMap &labels) {
  VicLabelMap::iterator i;
  //std::cout << "ADDING LOG ENTRIES...\n";
  for (i=labels.begin(); i != labels.end();i++) {
    //std::cout << "Next keyword is: " << (*i).first << "\n";
    VicLabel l = (*i).second;
    grd.addLogEntry(l.getWholeLabelStr());
    //std::cout << "Added log entry: " << l.getWholeLabelStr() << "\n";
  }
}

/***************************************************************************
 * FUNCTION: XcGridVicarLoad
 * DESCRIPTION: 
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   Only handles straight mxn DEM files.  NO RANGE MAPS!
 ***************************************************************************/
bool
XcGridVicarLoad (std::string filename,XcGrid &grd) {
  VicLabelMap labels = LoadVicarLabelsFromFile(filename);

  const int width  = labels["NS"].getIntVal();
  const int height = labels["NL"].getIntVal();
  if (1 != labels["NB"].getIntVal()) {
    std::cerr << "ERROR: XcVicarGridLoad can only handle 1 Band files.\n";
    return (false);
  }

  AddVicarLabelsToGridLog(grd,labels);

  grd.setRegular(true); // FIX: this will not be the case for range maps
  grd.setSize(width,height);
  grd.setDelXDelY(1.,1.);  	// FIX: what label would these be in?
  grd.setOrigin(0.,0.);		// FIX: what label for origin?

  if (!LoadVicarDataFromFile(filename,labels,grd)) {
    std::cerr << "XcVicarGridLoad: unable to load data\n";
    return(false);
  }

  return (true);
}

/***************************************************************************
 * FUNCTION: XcGridVicarWrite
 * DESCRIPTION: The most minimal vicar writer
 * INPUTS:  
 * OUTPUTS: 
 * RETURNS: 
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   Ignores what is in the grid log
 ***************************************************************************/

bool XcGridVicarWrite (std::string filename, XcGrid &grd) {
  std::vector<std::string> labels;
  std::string space("  ");

  //assert (XC_FLOAT == grd.getDataType()); // FIX: not ready to handle other yet.

  //labels.push_back("FORMAT="+getVicFormat(grd.getDataType())+"  ");
  #if 0
  labels.push_back("FORMAT='"+"REAL"+"'"+space);
  labels.push_back("TYPE='IMAGE'"+space);
  labels.push_back("BUFSIZE=20480"+space); // Set same as RECSIZE.  BUFSIZE is obsolete.
  labels.push_back("DIM=3"+space); // Strange thing that is always 3???
  labels.push_back("EOL=0"+space); // No additional stuff after the image data.
  labels.push_back("RECSIZE"+space); // NBB + N1 *pixel_size
  labels.push_back("ORG='BSQ'"+space); // Organisation of the file.  Band Sequential
  labels.push_back("NL="+space);
  labels.push_back("NS="+space);
  labels.push_back("NB="+space);
  // FOR BSQ - N1=Samples, N2=Lines, N3=Bands
  labels.push_back("N1="+space); // Fastest varying dim
  labels.push_back("N2="+space);
  labels.push_back("N3="+space); // Slowest varying dim
  labels.push_back("N4="+space);
  labels.push_back("NBB="+space);
  labels.push_back("NLB="+space);
  labels.push_back("HOST="+space);
  labels.push_back("INTFMT="+space);
  labels.push_back("REALFMT="+space);
  labels.push_back("BHOST="+space);
  labels.push_back("TASK="+space);
  labels.push_back("USER="+space);
  labels.push_back("DAT_TIM="+space);
#endif

  size_t bytesToHere=0;
  size_t i;
  for (i=0;i<labels.size();i++) {
    bytesToHere += labels[i].length();
  }

  // Now at add a 25 byte string of LBLSIZE= with padding
  return (true);
}

//////////////////////////////////////////////////////////////////////
// DRIVER IMPLEMENTATION - XcGridDrv
//////////////////////////////////////////////////////////////////////

std::string XcGridDrvVicar::getName() { return("VICAR");}
//bool XcGridDrvVicar::canReadType(FileType t) {return (false);}
//bool XcGridDrvVicar::canWriteType(FileType t) {return(false);}
std::vector<FileType> XcGridDrvVicar::getReadTypes() {
  std::vector<FileType> types;
  types.push_back(VICAR);
  return (types);
}
std::vector<FileType> XcGridDrvVicar::getWriteTypes() {
  std::vector<FileType> types;
  types.push_back(VICAR);
  return (types);
}

bool XcGridDrvVicar::ReadGrid(std::string filename, FileType t, XcGrid &grd) {
  assert (t == VICAR);
  bool r = XcGridVicarLoad(filename, grd);
  return (r);
}
bool XcGridDrvVicar::WriteGrid(std::string filename, FileType t, XcGrid &grd, gengetopt_args_info &opts) {
  assert (t == VICAR);
  // Nothing needed from opts just yet.
  bool r = XcGridVicarWrite (filename, grd);
  return (r);
}


//////////////////////////////////////////////////////////////////////
// TEST CODE
//////////////////////////////////////////////////////////////////////

#ifdef REGRESSION_TEST

// Have to provide out own register command so we don't depend on
// other drivers... Only register ourself.
XcG::VecDriverPtr XcG::XcGridDrvRegister() {
  VecDriverPtr drivers;
  drivers.push_back(new XcGridDrvVicar);	// JPL VICAR planetary data format
  return(drivers);
}

bool
isequal (const float a, const float b, const float del) {
  if ( a<b+del && a > b-del) return (true);
  return (false);
}

int main (int argc, char *argv[]) {
  bool ok=true;

  if (2!=sizeof(short)) {
    cerr << "  ERROR: short is not 2 bytes.  On a 64bit machine?"
	 << argv[0] << ": FAILED.";
    exit(EXIT_FAILURE);
  }
  if (4!=sizeof(int)) {
    cerr << "  ERROR: int is not 4 bytes.  On a 64bit machine?"
	 << argv[0] << ": FAILED.";
    exit(EXIT_FAILURE);
  }
  if (4!=sizeof(float)) {
    cerr << "  ERROR: float is not 4 bytes.  On a 64bit machine?"
	 << argv[0] << ": FAILED.";
    exit(EXIT_FAILURE);
  }
  //  if (8!=sizeof(float)) {
  //    cerr << "  ERROR: double is not 8 bytes.  On a 64bit machine?"
  //	 << argv[0] << ": FAILED.";
  //    exit(EXIT_FAILURE);
  //  }

  // gen4x4{byte,half,full,real}{Linux,Mac}.vic tests
  {
    float expected[4][4] = 
      {
	{0, 1, 2, 3},
	{1, 2, 3, 4},
	{2, 3, 4, 5},
	{3, 4, 5, 6}
      };

    size_t i=0;
    vector<std::string> files;
    files.push_back("TestData/gen4x4byteMac.vic");
    files.push_back("TestData/gen4x4byteLinux.vic");
    files.push_back("TestData/gen4x4halfMac.vic");
    files.push_back("TestData/gen4x4halfLinux.vic");
    files.push_back("TestData/gen4x4fullMac.vic");
    files.push_back("TestData/gen4x4fullLinux.vic");
    files.push_back("TestData/gen4x4realMac.vic");
    files.push_back("TestData/gen4x4realLinux.vic");

    for (i=0;i<files.size();i++) {
      bool loadOK=true;
      std::string filename(files[i]);
      XcGrid grd;
      if (!XcGridVicarLoad (filename,grd)) {
	std::cerr << "ERROR: XcVicGridLoad failed!\n";
	return(EXIT_FAILURE);
      }
      size_t x,y;
      for (y=0;y<4;y++) {
	for (x=0;x<4;x++) {
	  float _x,_y,z;
	  float e = expected[x][y];
	  grd.getXYZ(x,y,_x,_y,z);
	  if (!isequal(z,e,0.01)) {
	    ok=loadOK=false;
	    std::cerr << "ERROR: 4x4 failed for: ("
		      << x << "," << y << ")\n"
		      << " Wanted: " << e << "\n"
		      << "  Found: " << z << "\n\n";
	  }
	}
      }
      std::cout << "  Load " << filename << (loadOK?": ok.":": FAILED!")<<std::endl;
    }
  } // Load file test 4x4 HALF

  ////////////////////////////////////////
  // TEST REGISTRATION
  ////////////////////////////////////////

  {
    VecDriverPtr drivers = XcGridDrvRegister();  // vector of drivers
    bool r = (1==drivers.size()); // Should get at least 3 drivers (vicar, pds, pgm)
    if (!r) ok=false;
    std::cout << "  XcGridDrvRegister test: " << (ok?"ok.":": FAILED!") << std::endl;

    // FIX: what is the not found for vector, it's NOT npos??
    if (r) {
      size_t drv;
      drv = XcGridDrvGetReader(drivers,VICAR);
      if (string::npos==drv || !drivers[drv]->canReadType(VICAR)) {
	std::cout << "  ERROR:  vicar driver failed to respond correctly" << std::endl;
	ok=false;
      }
      drv = XcGridDrvGetWriter(drivers,VICAR);
      if (string::npos==drv || !drivers[drv]->canWriteType(VICAR)) {
	std::cout << "  ERROR:  vicar driver failed to respond correctly" << std::endl;
	ok=false;
      }
    }

    // Test reading.  Btw, we know we have the driver at position 0.
    {
      //XcGridOpts opts;
      XcGrid g;
      std::string filename("TestData/gen4x4byteMac.vic");
      r = drivers[0]->ReadGrid(filename,VICAR,g);
      if (!r) ok=false;
      std::cout << "  XcGridDrvRead  test: " << (r?"ok.":"FAILED!") << std::endl;
    }

    // Test write
    {
      gengetopt_args_info opts;
      XcGrid g(2,2);
      g.setDelXDelY(1.0,1.0);
      g.setZ(0,0,0);
      g.setZ(1,0,1.1);
      g.setZ(0,1,2.2);
      g.setZ(1,1,3.3);

      std::string filename("test-vicar-write.vic");
      r = drivers[0]->WriteGrid(filename,VICAR,g,opts);
      if (!r) ok=false;
      std::cout << "  XcGridDrvWrite test: " << (r?"ok.":"FAILED!") << std::endl;
    }

    XcGridDrvDelete(drivers);
    r = (0==drivers.size());
    if (!r) ok=false;
    std::cout << "  XcGridDrvDelete test: " << (r?"ok.":"FAILED!") << std::endl;

  } // Registry test

  std::cout << argv[0] << (ok?": ok.":": FAILED!")<<std::endl;
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}
#endif // REGRESSION_TEST
