/* $Revision: 1.11 $ */

/****************************************************************************
 *
 * PROJECT: X-Core / My Thesis
 *
 * MODULE:  Terrain Generators
 *
 * FILE:    xcgridconvert.C
 *
 * DESCRIPTION: Command line program to convert from one type to another
 *	using XcGrid infrastructure.  This is the general interface using
 *
 * EXPORTS:  
 *
 * NOTES:        
 *	Uses gengetops struct to pass args now.
 *
 * (c) Copyright 2004 Kurt Schwehr. All rights reserved.
 ***************************************************************************/

// FIX: list does not work.


/***************************************************************************
 * INCLUDES
 ***************************************************************************/
#include <cassert>

#include <iostream>
#include <string>

#include "cmdline.h"

#include "XcGrid.H"
#include "XcGridFile.H" // Contains the type names that we will use
#include "XcGridDrv.H" // Parent for all drivers

using namespace XcG;
using namespace std;

/***************************************************************************
 *
 * FUNCTION: print
 * DESCRIPTION: Print out the command line args
 * INPUTS:  
 * OUTPUTS: The gengetopt struct
 * RETURNS: Nasa
 * EXCEPTIONS: 
 * DESIGN:  
 * NOTES:   
 *
 ***************************************************************************/

void
print(const gengetopt_args_info &a) {
  cout << "help_given: " << (a.help_given?"yes":"no") << endl; // should never be true
  cout << "version_given: " << (a.version_given?"yes":"no") << endl; // should never be true
  assert (!a.help_given);
  assert (!a.version_given);
  cout << "in:      " << (a.in_given?a.in_arg:"none given") << endl
       << "out:     " << (a.out_given?a.out_arg:"none given") << endl
       << "format:  " << (a.format_given?a.format_arg:"none given") << endl
       << "cpt:     " << (a.cpt_given?a.cpt_arg:"none given") << endl
       << "tex:     " << (a.tex_given?a.tex_arg:"none given") << endl
       << "list:    " << (a.list_given?"Yes list":"Do not list") << endl
       << "normals: " << (a.normals_given?"specified - ":"none given - ") << a.normals_flag << endl
       << "spikes:  " << (a.spikes_given?"specified - ":"none given - ") << a.spikes_flag << endl
       << "xoffset: " << (a.xoffset_given?"specified - ":"none given - ") << a.xoffset_arg << endl
       << "yoffset: " << (a.yoffset_given?"specified - ":"none given - ") << a.yoffset_arg << endl
       << "zoffset: " << (a.zoffset_given?"specified - ":"none given - ") << a.zoffset_arg << endl
       << "zscale:  " << (a.zscale_given?"specified - ":"none given - ") << a.zscale_arg << endl
       << "verbose: " << (a.verbose_given?"specified - ":"none given - ") << a.verbose_arg << endl
       << endl;
  return;
} // print gengetopt_args_info

/***************************************************************************
 * MAIN
 ***************************************************************************/

int main (int argc, char *argv[]) {
  std::string infile;
  std::string outfile;
  std::string outFormatName;
  FileType inFileType = UNKNOWN_FILETYPE;
  FileType outFileType = UNKNOWN_FILETYPE;

  gengetopt_args_info a;
  if (0!=cmdline_parser(argc,argv,&a)) {
    cerr << "FIX: should never get here" << endl;
    cerr << "Early exit" << endl;
    return (EXIT_FAILURE);
  }

  if (a.verbose_arg>10) {
    print(a);
  }

  if (a.list_given) {
    VecDriverPtr drivers = XcGridDrvRegister();
    cout << endl;
    XcGridDrvPrint(drivers);
    return (EXIT_SUCCESS); // FIX: what should it return???
  }

  // FIX: also identify all files listed after the arg.
  if (a.identify_given) {
    VecDriverPtr drivers = XcGridDrvRegister();
    infile = std::string(a.identify_arg);
    inFileType = XcGridGetFileType (infile);
    if (UNKNOWN_FILETYPE==inFileType || BAD_FILE==inFileType) {
      cerr << "ERROR: unable to determine filetype for " << infile << endl;
      return (EXIT_FAILURE);
    }
    cout << XcGridFileTypeString(inFileType) << endl;
    return (EXIT_SUCCESS);
  }

  // Have to check the --in, --out, and --format args here since we don't want 
  // to force them to be required for --list  (make sure *.ggo does not require these)
  {
    bool missing_required_options = false;
    if (!a.in_given) {
      cerr << CMDLINE_PARSER_PACKAGE << ": '--in' ('-i') option required" << endl;
      missing_required_options = true;
    }
    if (!a.out_given) {
      cerr <<  CMDLINE_PARSER_PACKAGE << ": '--out' ('-o') option required" << endl;
      missing_required_options = true;
    }
    if (!a.format_given) {
      cerr <<  CMDLINE_PARSER_PACKAGE << ": '--format' ('-f') option required" << endl;
      missing_required_options = true;
    }
    if (missing_required_options) {
      cerr << endl << "You are missing some important options.  Here is some help:" << endl << endl;
      cmdline_parser_print_help();
      return (EXIT_FAILURE);
    }
  }

  assert (a.format_given);
  outFormatName = std::string(a.format_arg);
  outFileType = XcGridFileType(outFormatName);
  if (UNKNOWN_FILETYPE==outFileType || BAD_FILE==outFileType) {
    std::cerr << "ERROR! Bad file type: " << outFormatName << std::endl;
    // FIX: print list of ok output file types.
    exit(EXIT_FAILURE);
  }

  infile = std::string(a.in_arg);
  inFileType = XcGridGetFileType (infile);
  if (UNKNOWN_FILETYPE==inFileType || BAD_FILE==inFileType) {
    cerr << "ERROR: unable to determine filetype for " << infile << endl;
    return (EXIT_FAILURE);
  }

  outfile = std::string(a.out_arg);
  // FIX: bring this back?
  //if (a.tex_given) opts.textureFile = std::string(a.in_tex);

  VecDriverPtr drivers = XcGridDrvRegister();
  size_t inDrv = XcGridDrvGetReader(drivers,inFileType);
  if (string::npos==inDrv || !drivers[inDrv]->canReadType(inFileType)) {
    cerr << "ERROR: no working driver to read your file" << endl;
    if (!string::npos==inDrv) {
      cerr << "  Your file type: " << XcGridFileTypeString(inFileType) << endl
	   << "  Tried driver: " << drivers[inDrv]->getName() << endl
	   << "  Internal programming error!" << endl;
    }
    return (EXIT_FAILURE);
  }

  size_t outDrv = XcGridDrvGetWriter(drivers,outFileType);
  if (string::npos==outDrv || !drivers[outDrv]->canWriteType(outFileType)) {
    cerr << "ERROR: no working driver to write your file" << endl;
    if (!string::npos==inDrv) {
      cerr << "  Your file type: " << XcGridFileTypeString(inFileType) << endl
	   << "  Tried driver: " << drivers[outDrv]->getName() << endl
	   << "  Internal programming error!" << endl;
    }
    return (EXIT_FAILURE);
  }

  XcGrid grd;
  if (!drivers[inDrv]->ReadGrid(infile,inFileType,grd)) {
    cerr << "ERROR: Read failed!" << endl;
    return (EXIT_FAILURE);
  }

  // Sanity check the input file.
  bool ok=true;
  if (grd.getWidth()<=0) {
    cerr << "ERROR: input grid width too small: " << grd.getWidth() << endl;
    ok=false;
  }
  if (grd.getHeight()<=0) {
    cerr << "ERROR: input grid height too small: " << grd.getHeight() << endl;
    ok=false;
  }
  if (!ok) {
    cerr << "Bailing out without writing anything" << endl;
    return (EXIT_FAILURE);
  }

  if (a.cpt_given) {
    string cpt_filename (a.cpt_arg);
    if (!grd.setColorPalette(cpt_filename)) ok=false;
  }
  if (!ok) {
    cerr << "Bailing out without writing anything.  Color Palette failed to load." << endl;
    return (EXIT_FAILURE);
  }

  if (a.zscale_given) {
    cout << "Scaling... by " << a.zscale_arg << endl;
    grd.scaleZ(a.zscale_arg,0.);
  }

  if (a.xoffset_given) {cerr << "FIX: xoffset not yet supported" << endl;}
  if (a.yoffset_given) {cerr << "FIX: xoffset not yet supported" << endl;}


  if (!drivers[outDrv]->WriteGrid(outfile,outFileType,grd,a)) {
    cerr << "\nERROR: write failed!" << endl;
  }
  
  return (ok?EXIT_SUCCESS:EXIT_FAILURE);
}

